@@ -124,8 +124,10 @@ BAN_Release(void)
 	assert(ban_holds > 0);
 	ban_holds--;
 	Lck_Unlock(&ban_mtx);
-	if (ban_holds == 0)
+	if (ban_holds == 0) {
+		BANIDX_fini();
 		WRK_BgThread(&ban_thread, "ban-lurker", ban_lurker, NULL);
+	}
 }
 
 /*--------------------------------------------------------------------
@@ -220,6 +222,7 @@ ban_get_lump(const uint8_t **bs)
 static void
 ban_iter(const uint8_t **bs, struct ban_test *bt)
 {
+	const void *lump;
 	uint64_t dtmp;
 
 	memset(bt, 0, sizeof *bt);
@@ -229,15 +232,14 @@ ban_iter(const uint8_t **bs, struct ban_test *bt)
 		bt->arg1_spec = (const char *)*bs;
 		(*bs) += (*bs)[0] + 2;
 	}
+	lump = ban_get_lump(bs);
+	bt->oper = *(*bs)++;
 	if (BANS_HAS_ARG2_DOUBLE(bt->arg1)) {
-		dtmp = vbe64dec(ban_get_lump(bs));
-		bt->oper = *(*bs)++;
-
+		dtmp = vbe64dec(lump);
 		memcpy(&bt->arg2_double, &dtmp, sizeof dtmp);
 		return;
 	}
-	bt->arg2 = ban_get_lump(bs);
-	bt->oper = *(*bs)++;
+	bt->arg2 = lump;
 	if (BANS_HAS_ARG2_SPEC(bt->oper))
 		bt->arg2_spec = ban_get_lump(bs);
 }
@@ -294,7 +296,8 @@ BAN_FindBan(vtim_real t0)
 	vtim_real t1;
 
 	assert(ban_holds > 0);
-	VTAILQ_FOREACH(b, &ban_head, list) {
+	b = BANIDX_lookup(t0);
+	VTAILQ_FOREACH_FROM(b, &ban_head, list) {
 		t1 = ban_time(b->spec);
 		if (t1 == t0)
 			return (b);
@@ -392,11 +395,13 @@ ban_reload(const uint8_t *ban, unsigned len)
 	vtim_real t0, t1, t2 = 9e99;
 	ASSERT_CLI();
 	Lck_AssertHeld(&ban_mtx);
+	assert(ban_holds > 0);
 
 	t0 = ban_time(ban);
 	assert(len == ban_len(ban));
 
-	VTAILQ_FOREACH(b, &ban_head, list) {
+	b = BANIDX_lookup(t0);
+	VTAILQ_FOREACH_FROM(b, &ban_head, list) {
 		t1 = ban_time(b->spec);
 		assert(t1 < t2);
 		t2 = t1;
@@ -495,11 +500,12 @@ ban_evaluate(struct worker *wrk, const uint8_t *bsarg, struct objcore *oc,
 	const char *p;
 	const char *arg1;
 	double darg1, darg2;
+	hdr_t hdr;
 	int rv;
 
 	/*
-	 * for ttl and age, fix the point in time such that banning refers to
-	 * the same point in time when the ban is evaluated
+	 * for ttl, age and last_hit, fix the point in time such that banning
+	 * refers to the same point in time when the ban is evaluated
 	 *
 	 * for grace/keep, we assume that the absolute values are pola and that
 	 * users will most likely also specify a ttl criterion if they want to
@@ -521,11 +527,13 @@ ban_evaluate(struct worker *wrk, const uint8_t *bsarg, struct objcore *oc,
 			break;
 		case BANS_ARG_REQHTTP:
 			AN(reqhttp);
-			(void)http_GetHdr(reqhttp, bt.arg1_spec, &p);
+			CAST_HDR(hdr, bt.arg1_spec);
+			(void)http_GetHdr(reqhttp, hdr, &p);
 			arg1 = p;
 			break;
 		case BANS_ARG_OBJHTTP:
-			arg1 = HTTP_GetHdrPack(wrk, oc, bt.arg1_spec);
+			CAST_HDR(hdr, bt.arg1_spec);
+			arg1 = HTTP_GetHdrPack(wrk, oc, hdr);
 			break;
 		case BANS_ARG_OBJSTATUS:
 			arg1 = HTTP_GetHdrPack(wrk, oc, H__Status);
@@ -546,6 +554,12 @@ ban_evaluate(struct worker *wrk, const uint8_t *bsarg, struct objcore *oc,
 			darg1 = oc->keep;
 			darg2 = bt.arg2_double;
 			break;
+		case BANS_ARG_OBJLASTHIT:
+			if (isnan(oc->last_lru))
+				return (0);
+			darg1 = 0.0 - oc->last_lru;
+			darg2 = 0.0 - (ban_time(bsarg) - bt.arg2_double);
+			break;
 		default:
 			WRONG("Wrong BAN_ARG code");
 		}
