@@ -71,18 +71,35 @@ VDP_Fini(const struct vdp_ctx *vdc)
 
 void
 VDP_Init(struct vdp_ctx *vdc, struct worker *wrk, struct vsl_log *vsl,
-    struct req *req)
+    const struct req *req, const struct busyobj *bo, intmax_t *clen)
 {
 	AN(vdc);
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	AN(vsl);
-	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+
+	AN(clen);
+
+	assert((req ? 1 : 0) ^ (bo ? 1 : 0));
+
+	AN(clen);
+	assert(*clen >= -1);
 
 	INIT_OBJ(vdc, VDP_CTX_MAGIC);
 	VTAILQ_INIT(&vdc->vdp);
 	vdc->wrk = wrk;
 	vdc->vsl = vsl;
-	vdc->req = req;
+	vdc->clen = clen;
+
+	if (req != NULL) {
+		CHECK_OBJ(req, REQ_MAGIC);
+		vdc->oc = req->objcore;
+		vdc->hp = req->resp;
+	}
+	else {
+		CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+		vdc->oc = bo->bereq_body;
+		vdc->hp = bo->bereq;
+	}
 }
 
 /* VDP_bytes
@@ -152,10 +169,13 @@ VDP_Push(VRT_CTX, struct vdp_ctx *vdc, struct ws *ws, const struct vdp *vdp,
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	CHECK_OBJ_ORNULL(vdc->oc, OBJCORE_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc->hp, HTTP_MAGIC);
+	AN(vdc->clen);
+	assert(*vdc->clen >= -1);
 	AN(ws);
 	AN(vdp);
 	AN(vdp->name);
-	AN(vdp->bytes);
 
 	if (vdc->retval)
 		return (vdc->retval);
@@ -176,17 +196,17 @@ VDP_Push(VRT_CTX, struct vdp_ctx *vdc, struct ws *ws, const struct vdp *vdp,
 	vdc->nxt = VTAILQ_FIRST(&vdc->vdp);
 
 	AZ(vdc->retval);
-	if (vdpe->vdp->init == NULL)
-		return (vdc->retval);
-
-	vdc->retval = vdpe->vdp->init(ctx, vdc, &vdpe->priv,
-	    vdpe == vdc->nxt ? vdc->req->objcore : NULL);
+	if (vdpe->vdp->init != NULL)
+		vdc->retval = vdpe->vdp->init(ctx, vdc, &vdpe->priv);
+	vdc->oc = NULL;
 
 	if (vdc->retval > 0) {
 		VTAILQ_REMOVE(&vdc->vdp, vdpe, list);
 		vdc->nxt = VTAILQ_FIRST(&vdc->vdp);
 		vdc->retval = 0;
 	}
+	else if (vdc->retval == 0)
+		AN(vdp->bytes);
 	return (vdc->retval);
 }
 
@@ -201,20 +221,15 @@ VDP_Close(struct vdp_ctx *vdc, struct objcore *oc, struct boc *boc)
 	CHECK_OBJ_ORNULL(oc, OBJCORE_MAGIC);
 	CHECK_OBJ_ORNULL(boc, BOC_MAGIC);
 
-	while (!VTAILQ_EMPTY(&vdc->vdp)) {
-		vdpe = VTAILQ_FIRST(&vdc->vdp);
+	while ((vdpe = VTAILQ_FIRST(&vdc->vdp)) != NULL) {
+		CHECK_OBJ(vdpe, VDP_ENTRY_MAGIC);
 		rv = vdpe->bytes_in;
 		VSLb(vdc->vsl, SLT_VdpAcct, "%s %ju %ju", vdpe->vdp->name,
 		    (uintmax_t)vdpe->calls, (uintmax_t)rv);
-		if (vdc->retval >= 0)
-			AN(vdpe);
-		if (vdpe != NULL) {
-			CHECK_OBJ(vdpe, VDP_ENTRY_MAGIC);
-			if (vdpe->vdp->fini != NULL)
-				AZ(vdpe->vdp->fini(vdc, &vdpe->priv));
-			AZ(vdpe->priv);
-			VTAILQ_REMOVE(&vdc->vdp, vdpe, list);
-		}
+		if (vdpe->vdp->fini != NULL)
+			AZ(vdpe->vdp->fini(vdc, &vdpe->priv));
+		AZ(vdpe->priv);
+		VTAILQ_REMOVE(&vdc->vdp, vdpe, list);
 		vdc->nxt = VTAILQ_FIRST(&vdc->vdp);
 #ifdef VDP_PEDANTIC_ARMED
 		// enable when we are confident to get VDP_END right
@@ -229,8 +244,195 @@ VDP_Close(struct vdp_ctx *vdc, struct objcore *oc, struct boc *boc)
 
 /*--------------------------------------------------------------------*/
 
-static int v_matchproto_(objiterate_f)
-vdp_objiterate(void *priv, unsigned flush, const void *ptr, ssize_t len)
+/*
+ * Push a VDPIO vdp. This can only be used with only vdpio-enabled VDPs or
+ * after a successful upgrade
+ */
+int
+VDPIO_Push(VRT_CTX, struct vdp_ctx *vdc, struct ws *ws, const struct vdp *vdp,
+    void *priv)
+{
+	struct vdp_entry *vdpe;
+	int r;
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	CHECK_OBJ_ORNULL(vdc->oc, OBJCORE_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc->hp, HTTP_MAGIC);
+	AN(vdc->clen);
+	assert(*vdc->clen >= -1);
+	AN(ws);
+	AN(vdp);
+	AN(vdp->name);
+
+	if (vdc->retval < 0)
+		return (vdc->retval);
+
+	AN(vdp->io_init);
+
+	// the first VDP (which leases from storage) only gets the minimum
+	// capacity requirement of 1
+	if (vdc->retval == 0) {
+		assert(VTAILQ_EMPTY(&vdc->vdp));
+		vdc->retval = 1;
+	}
+
+	if (DO_DEBUG(DBG_PROCESSORS))
+		VSLb(vdc->vsl, SLT_Debug, "VDPIO_push(%s)", vdp->name);
+
+	vdpe = WS_Alloc(ws, sizeof *vdpe);
+	if (vdpe == NULL) {
+		vdc->retval = -ENOMEM;
+		return (vdc->retval);
+	}
+	INIT_OBJ(vdpe, VDP_ENTRY_MAGIC);
+	vdpe->vdp = vdp;
+	vdpe->priv = priv;
+	VTAILQ_INSERT_TAIL(&vdc->vdp, vdpe, list);
+	vdc->nxt = VTAILQ_FIRST(&vdc->vdp);
+
+	assert(vdc->retval > 0);
+	if (vdpe->vdp->io_init != NULL) {
+		r = vdpe->vdp->io_init(ctx, vdc, &vdpe->priv, vdc->retval);
+		if (r <= 0) {
+			VTAILQ_REMOVE(&vdc->vdp, vdpe, list);
+			vdc->nxt = VTAILQ_FIRST(&vdc->vdp);
+		}
+		else
+			AN(vdp->io_lease);
+		if (r != 0)
+			vdc->retval = r;
+	}
+	vdc->oc = NULL;
+	return (vdc->retval);
+}
+
+/*
+ * upgrade an already initialized VDP filter chain to VDPIO, if possible
+ * returns:
+ * > 0 cap
+ * -ENOTSUP io_upgrade missing for at least one filter
+ * vdc->retval if < 0
+ */
+int
+VDPIO_Upgrade(VRT_CTX, struct vdp_ctx *vdc)
+{
+	struct vdp_entry *vdpe;
+	int cap, r;
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+
+	VTAILQ_FOREACH(vdpe, &vdc->vdp, list)
+		if (vdpe->vdp->io_upgrade == NULL)
+			return (-ENOTSUP);
+
+	if (vdc->retval < 0)
+		return (vdc->retval);
+
+	// minimum capacity requirement for the first filter (after storage)
+	r = cap = 1;
+	VTAILQ_FOREACH(vdpe, &vdc->vdp, list) {
+		r = vdpe->vdp->io_upgrade(ctx, vdc, &vdpe->priv, cap);
+		if (DO_DEBUG(DBG_PROCESSORS)) {
+			VSLb(vdc->vsl, SLT_Debug, "VDPIO_Upgrade "
+			    "%d = %s(cap = %d)",
+			    r, vdpe->vdp->name, cap);
+		}
+		if (r < 0)
+			return ((vdc->retval = r));
+		// XXX remove if filter does not want to be pushed?
+		assert(r != 0);
+		cap = r;
+	}
+	return ((vdc->retval = r));
+}
+
+uint64_t
+VDPIO_Close1(struct vdp_ctx *vdc, struct vdp_entry *vdpe)
+{
+	uint64_t rv;
+
+	CHECK_OBJ_NOTNULL(vdpe, VDP_ENTRY_MAGIC);
+	rv = vdpe->bytes_in;
+	VSLb(vdc->vsl, SLT_VdpAcct, "%s %ju %ju", vdpe->vdp->name,
+	    (uintmax_t)vdpe->calls, (uintmax_t)rv);
+	if (vdpe->vdp->io_fini != NULL)
+		vdpe->vdp->io_fini(vdc, &vdpe->priv);
+	AZ(vdpe->priv);
+	VTAILQ_REMOVE(&vdc->vdp, vdpe, list);
+	vdc->nxt = VTAILQ_FIRST(&vdc->vdp);
+	return (rv);
+}
+
+uint64_t
+VDPIO_Close(struct vdp_ctx *vdc, struct objcore *oc, struct boc *boc)
+{
+	struct vdp_entry *vdpe;
+	uint64_t rv = 0;
+
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc->wrk, WORKER_MAGIC);
+	CHECK_OBJ_ORNULL(oc, OBJCORE_MAGIC);
+	CHECK_OBJ_ORNULL(boc, BOC_MAGIC);
+
+	while ((vdpe = VTAILQ_FIRST(&vdc->vdp)) != NULL)
+		rv = VDPIO_Close1(vdc, vdpe);
+
+	if (oc != NULL)
+		HSH_Cancel(vdc->wrk, oc, boc);
+	return (rv);
+}
+
+/*
+ * ============================================================
+ * VDPIO helpers: VAI management
+ *
+ * Transports should not need to talk to the VAI Object interface directly,
+ * because its state is kept in the vdp_ctx
+ *
+ * So we wrap init, return and fini
+ */
+
+// return true if error
+int
+VDPIO_Init(struct vdp_ctx *vdc, struct objcore *oc, struct ws *ws,
+    vai_notify_cb *notify_cb, void *notify_priv, struct vscaret *scaret)
+{
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	VSCARET_CHECK_NOTNULL(scaret);
+	AN(scaret->capacity);
+	AZ(scaret->used);
+
+	AZ(vdc->vai_hdl);
+	vdc->vai_hdl = ObjVAIinit(vdc->wrk, oc, ws, notify_cb, notify_priv);
+	if (vdc->vai_hdl == NULL)
+		return (1);
+	vdc->scaret = scaret;
+	return (0);
+}
+
+// return leases stashed in scaret
+void
+VDPIO_Return(const struct vdp_ctx *vdc)
+{
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+
+	ObjVAIreturn(vdc->wrk, vdc->vai_hdl, vdc->scaret);
+}
+
+void
+VDPIO_Fini(struct vdp_ctx *vdc)
+{
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+
+	VDPIO_Return(vdc);
+	ObjVAIfini(vdc->wrk, &vdc->vai_hdl);
+}
+
+/*--------------------------------------------------------------------*/
+int v_matchproto_(objiterate_f)
+VDP_ObjIterate(void *priv, unsigned flush, const void *ptr, ssize_t len)
 {
 	enum vdp_action act;
 
@@ -254,9 +456,11 @@ VDP_DeliverObj(struct vdp_ctx *vdc, struct objcore *oc)
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 	CHECK_OBJ_NOTNULL(vdc->wrk, WORKER_MAGIC);
 	AN(vdc->vsl);
-	vdc->req = NULL;
-	final = oc->flags & (OC_F_PRIVATE | OC_F_HFM | OC_F_HFP) ? 1 : 0;
-	r = ObjIterate(vdc->wrk, oc, vdc, vdp_objiterate, final);
+	AZ(vdc->oc);
+	vdc->hp = NULL;
+	vdc->clen = NULL;
+	final = oc->flags & OC_F_TRANSIENT ? 1 : 0;
+	r = ObjIterate(vdc->wrk, oc, vdc, VDP_ObjIterate, final);
 	if (r < 0)
 		return (r);
 	return (0);
