@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -30,8 +32,10 @@
 
 #include "config.h"
 
-#include "cache.h"
+#include "cache_varnishd.h"
+#include "cache_filter.h"
 
+#include "cache_vgz.h"
 #include "cache_esi.h"
 #include "vct.h"
 #include "vend.h"
@@ -104,12 +108,13 @@ struct vep_state {
 	struct vep_match	*match;
 	struct vep_match	*match_hit;
 
-	char			tag[10];
+	char			tag[8];
 	int			tag_i;
 
 	dostuff_f		*dostuff;
 
 	struct vsb		*include_src;
+	unsigned		include_continue;
 
 	unsigned		nm_skip;
 	unsigned		nm_verbatim;
@@ -128,9 +133,11 @@ static const char * const VEP_NEXTTAG =		"[NxtTag]";
 static const char * const VEP_NOTMYTAG =	"[NotMyTag]";
 
 static const char * const VEP_STARTTAG =	"[StartTag]";
+static const char * const VEP_COMMENTESI =	"[CommentESI]";
 static const char * const VEP_COMMENT =		"[Comment]";
 static const char * const VEP_CDATA =		"[CDATA]";
 static const char * const VEP_ESITAG =		"[ESITag]";
+static const char * const VEP_ESIENDTAG =	"[/ESITag]";
 
 static const char * const VEP_ESIREMOVE =	"[ESI:Remove]";
 static const char * const VEP_ESIINCLUDE =	"[ESI:Include]";
@@ -153,7 +160,10 @@ static const char * const VEP_MATCH =		"[Match]";
 /*---------------------------------------------------------------------*/
 
 static struct vep_match vep_match_starttag[] = {
+	{ "!--esi",	&VEP_COMMENTESI },
+	{ "!---->",	&VEP_NEXTTAG },
 	{ "!--",	&VEP_COMMENT },
+	{ "/esi:",	&VEP_ESIENDTAG },
 	{ "esi:",	&VEP_ESITAG },
 	{ "![CDATA[",	&VEP_CDATA },
 	{ NULL,		&VEP_NOTMYTAG }
@@ -172,6 +182,7 @@ static struct vep_match vep_match_esi[] = {
 
 static struct vep_match vep_match_attr_include[] = {
 	{ "src=",	&VEP_ATTRGETVAL },
+	{ "onerror=",	&VEP_ATTRGETVAL },
 	{ NULL,		&VEP_SKIPATTR }
 };
 
@@ -189,12 +200,9 @@ static struct vep_match vep_match_bom[] = {
 static void
 vep_error(const struct vep_state *vep, const char *p)
 {
-	intmax_t l;
-
 	VSC_C_main->esi_errors++;
-	l = (intmax_t)(vep->ver_p - vep->hack_p);
-	VSLb(vep->vc->wrk->vsl, SLT_ESI_xmlerror, "ERR at %jd %s", l, p);
-
+	VSLb(vep->vc->wrk->vsl, SLT_ESI_xmlerror, "ERR: after %zd %s",
+	     vep->o_last, p);
 }
 
 /*--------------------------------------------------------------------
@@ -204,12 +212,9 @@ vep_error(const struct vep_state *vep, const char *p)
 static void
 vep_warn(const struct vep_state *vep, const char *p)
 {
-	intmax_t l;
-
 	VSC_C_main->esi_warnings++;
-	l = (intmax_t)(vep->ver_p - vep->hack_p);
-	VSLb(vep->vc->wrk->vsl, SLT_ESI_xmlerror, "WARN at %jd %s", l, p);
-
+	VSLb(vep->vc->wrk->vsl, SLT_ESI_xmlerror, "WARN: after %zd %s",
+	     vep->o_last, p);
 }
 
 /*---------------------------------------------------------------------
@@ -217,28 +222,22 @@ vep_warn(const struct vep_state *vep, const char *p)
  */
 
 static struct vep_match *
-vep_match(struct vep_state *vep, const char *b, const char *e)
+vep_match(const struct vep_state *vep, const char *b, const char *e)
 {
 	struct vep_match *vm;
 	const char *q, *r;
-	ssize_t l;
 
-	for (vm = vep->match; vm->match; vm++) {
+	AN(vep->match);
+	for (vm = vep->match; vm->match != NULL; vm++) {
+		assert(strlen(vm->match) <= sizeof (vep->tag));
 		r = b;
-		for (q = vm->match; *q && r < e; q++, r++)
+		for (q = vm->match; *q != '\0' && r < e; q++, r++)
 			if (*q != *r)
 				break;
-		if (*q != '\0' && r == e) {
-			if (b != vep->tag) {
-				l = e - b;
-				assert(l < sizeof vep->tag);
-				memmove(vep->tag, b, l);
-				vep->tag_i = l;
-			}
-			return (NULL);
-		}
 		if (*q == '\0')
-			return (vm);
+			break;
+		if (r == e)
+			return (NULL);
 	}
 	return (vm);
 }
@@ -391,7 +390,6 @@ vep_mark_pending(struct vep_state *vep, const char *p)
 	AN(vep->ver_p);
 	l = p - vep->ver_p;
 	assert(l > 0);
-	assert(l >= 0);
 	vep->crcp = crc32(vep->crcp, (const void *)vep->ver_p, l);
 	vep->ver_p = p;
 
@@ -402,7 +400,7 @@ vep_mark_pending(struct vep_state *vep, const char *p)
 /*---------------------------------------------------------------------
  */
 
-static void __match_proto__()
+static void v_matchproto_()
 vep_do_comment(struct vep_state *vep, enum dowhat what)
 {
 	Debug("DO_COMMENT(%d)\n", what);
@@ -414,60 +412,90 @@ vep_do_comment(struct vep_state *vep, enum dowhat what)
 /*---------------------------------------------------------------------
  */
 
-static void __match_proto__()
+static void v_matchproto_()
 vep_do_remove(struct vep_state *vep, enum dowhat what)
 {
 	Debug("DO_REMOVE(%d, end %d empty %d remove %d)\n",
 	    what, vep->endtag, vep->emptytag, vep->remove);
 	assert(what == DO_TAG);
-	if (vep->emptytag) {
-		vep_error(vep,
-		    "ESI 1.0 <esi:remove/> not legal");
-	} else {
-		if (vep->remove && !vep->endtag)
-			vep_error(vep,
-			    "ESI 1.0 <esi:remove> already open");
-		else if (!vep->remove && vep->endtag)
-			vep_error(vep,
-			    "ESI 1.0 <esi:remove> not open");
-		else
-			vep->remove = !vep->endtag;
-	}
+	if (vep->emptytag)
+		vep_error(vep, "ESI 1.0 <esi:remove/> not legal");
+	else if (vep->remove && !vep->endtag)
+		vep_error(vep, "ESI 1.0 <esi:remove> already open");
+	else if (!vep->remove && vep->endtag)
+		vep_error(vep, "ESI 1.0 <esi:remove> not open");
+	else
+		vep->remove = !vep->endtag;
 }
 
 /*---------------------------------------------------------------------
  */
 
-static void __match_proto__()
+static void
+include_attr_src(struct vep_state *vep)
+{
+	const char *p;
+
+	if (vep->include_src != NULL) {
+		vep_error(vep,
+		    "ESI 1.0 <esi:include> "
+		    "has multiple src= attributes");
+		vep->state = VEP_TAGERROR;
+		VSB_destroy(&vep->attr_vsb);
+		VSB_destroy(&vep->include_src);
+		return;
+	}
+	for (p = VSB_data(vep->attr_vsb); *p != '\0'; p++)
+		if (vct_islws(*p))
+			break;
+	if (*p != '\0') {
+		vep_error(vep,
+		    "ESI 1.0 <esi:include> "
+		    "has whitespace in src= attribute");
+		vep->state = VEP_TAGERROR;
+		VSB_destroy(&vep->attr_vsb);
+		if (vep->include_src != NULL)
+			VSB_destroy(&vep->include_src);
+		return;
+	}
+	vep->include_src = vep->attr_vsb;
+	vep->attr_vsb = NULL;
+}
+
+static void
+include_attr_onerror(struct vep_state *vep)
+{
+
+	vep->include_continue = !strcmp("continue", VSB_data(vep->attr_vsb));
+	VSB_destroy(&vep->attr_vsb);
+}
+
+static void v_matchproto_()
 vep_do_include(struct vep_state *vep, enum dowhat what)
 {
 	const char *p, *q, *h;
 	ssize_t l;
+	char incl;
 
 	Debug("DO_INCLUDE(%d)\n", what);
 	if (what == DO_ATTR) {
 		Debug("ATTR (%s) (%s)\n", vep->match_hit->match,
 			VSB_data(vep->attr_vsb));
-		if (vep->include_src != NULL) {
-			vep_error(vep,
-			    "ESI 1.0 <esi:include> "
-			    "has multiple src= attributes");
-			vep->state = VEP_TAGERROR;
-			VSB_destroy(&vep->attr_vsb);
-			VSB_destroy(&vep->include_src);
+		if (!strcmp("src=", vep->match_hit->match)) {
+			include_attr_src(vep);
 			return;
 		}
-		vep->include_src = vep->attr_vsb;
-		vep->attr_vsb = NULL;
-		return;
+		if (!strcmp("onerror=", vep->match_hit->match)) {
+			include_attr_onerror(vep);
+			return;
+		}
+		WRONG("Unhandled <esi:include> attribute");
 	}
 	assert(what == DO_TAG);
 	if (!vep->emptytag)
-		vep_warn(vep,
-		    "ESI 1.0 <esi:include> lacks final '/'");
+		vep_warn(vep, "ESI 1.0 <esi:include> lacks final '/'");
 	if (vep->include_src == NULL) {
-		vep_error(vep,
-		    "ESI 1.0 <esi:include> lacks src attr");
+		vep_error(vep, "ESI 1.0 <esi:include> lacks src attr");
 		return;
 	}
 
@@ -486,6 +514,8 @@ vep_do_include(struct vep_state *vep, enum dowhat what)
 	l = VSB_len(vep->include_src);
 	h = 0;
 
+	incl = vep->include_continue ? VEC_IC : VEC_IA;
+
 	if (l > 7 && !memcmp(p, "http://", 7)) {
 		h = p + 7;
 		p = strchr(h, '/');
@@ -498,7 +528,7 @@ vep_do_include(struct vep_state *vep, enum dowhat what)
 			return;
 		}
 		Debug("HOST <%.*s> PATH <%s>\n", (int)(p-h),h, p);
-		VSB_printf(vep->vsb, "%c", VEC_INCL);
+		VSB_printf(vep->vsb, "%c", incl);
 		VSB_printf(vep->vsb, "Host: %.*s%c", (int)(p-h), h, 0);
 	} else if (l > 8 && !memcmp(p, "https://", 8)) {
 		if (!FEATURE(FEATURE_ESI_IGNORE_HTTPS)) {
@@ -521,13 +551,13 @@ vep_do_include(struct vep_state *vep, enum dowhat what)
 			VSB_destroy(&vep->include_src);
 			return;
 		}
-		VSB_printf(vep->vsb, "%c", VEC_INCL);
+		VSB_printf(vep->vsb, "%c", incl);
 		VSB_printf(vep->vsb, "Host: %.*s%c", (int)(p-h), h, 0);
 	} else if (*p == '/') {
-		VSB_printf(vep->vsb, "%c", VEC_INCL);
+		VSB_printf(vep->vsb, "%c", incl);
 		VSB_printf(vep->vsb, "%c", 0);
 	} else {
-		VSB_printf(vep->vsb, "%c", VEC_INCL);
+		VSB_printf(vep->vsb, "%c", incl);
 		VSB_printf(vep->vsb, "%c", 0);
 		/* Look for the last / before a '?' */
 		h = NULL;
@@ -560,15 +590,14 @@ vep_do_include(struct vep_state *vep, enum dowhat what)
 	}
 #undef R
 	VSB_printf(vep->vsb, "%c", 0);
-
-	VSB_delete(vep->include_src);
-	vep->include_src = NULL;
+	VSB_destroy(&vep->include_src);
+	vep->include_continue = 0;
 }
 
 /*---------------------------------------------------------------------
  * Lex/Parse object for ESI instructions
  *
- * This function is called with the input object piecemal so do not
+ * This function is called with the input object piecemeal so do not
  * assume that we have more than one char available at at time, but
  * optimize for getting huge chunks.
  *
@@ -607,14 +636,11 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 
 	while (p < e) {
 		AN(vep->state);
-		i = e - p;
-		if (i > 10)
-			i = 10;
 		Debug("EP %s %d (%.*s) [%.*s]\n",
 		    vep->state,
 		    vep->remove,
 		    vep->tag_i, vep->tag,
-		    i, p);
+		    (e - p) > 10 ? 10 : (int)(e-p), p);
 		assert(p >= vep->ver_p);
 
 		/******************************************************
@@ -622,7 +648,8 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 		 */
 
 		if (vep->state == VEP_START) {
-			if (FEATURE(FEATURE_ESI_REMOVE_BOM) && *p == '\xeb') {
+			if (FEATURE(FEATURE_ESI_REMOVE_BOM) &&
+			    *p == (char)0xeb) {
 				vep->match = vep_match_bom;
 				vep->state = VEP_MATCH;
 			} else
@@ -644,16 +671,16 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 			if (p < e && *p == '<') {
 				p++;
 				vep->state = VEP_STARTTAG;
-			} else if (p < e && *p == '\xeb') {
+			} else if (p < e && *p == (char)0xeb) {
 				VSLb(vep->vc->wrk->vsl, SLT_ESI_xmlerror,
-				    "No ESI processing, "
+				    "WARN: No ESI processing, "
 				    "first char not '<' but BOM."
 				    " (See feature esi_remove_bom)"
 				);
 				vep->state = VEP_NOTXML;
 			} else if (p < e) {
 				VSLb(vep->vc->wrk->vsl, SLT_ESI_xmlerror,
-				    "No ESI processing, "
+				    "WARN: No ESI processing, "
 				    "first char not '<'."
 				    " (See feature esi_disable_xml_check)"
 				);
@@ -692,7 +719,6 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 			 * out for end of EsiCmt if armed.
 			 */
 			vep->emptytag = 0;
-			vep->endtag = 0;
 			vep->attr = NULL;
 			vep->dostuff = NULL;
 			while (p < e && *p != '<') {
@@ -705,8 +731,7 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 					vep->esicmt_p = vep->esicmt;
 					continue;
 				}
-				if (!vep->remove &&
-				    vep->esicmt_p == vep->esicmt)
+				if (!vep->remove && vep->esicmt_p == vep->esicmt)
 					vep_mark_verbatim(vep, p);
 				p++;
 				if (*++vep->esicmt_p == '\0') {
@@ -735,44 +760,23 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 		 */
 
 		} else if (vep->state == VEP_STARTTAG) {
-			/*
-			 * Start of tag, set up match table
-			 */
-			if (p < e) {
-				if (*p == '/') {
-					vep->endtag = 1;
-					p++;
-				}
-				vep->match = vep_match_starttag;
-				vep->state = VEP_MATCH;
-			}
+			/* Start of tag, set up match table */
+			vep->endtag = 0;
+			vep->match = vep_match_starttag;
+			vep->state = VEP_MATCH;
 		} else if (vep->state == VEP_COMMENT) {
-			/*
-			 * We are in a comment, find out if it is an
-			 * ESI comment or a regular comment
-			 */
-			if (vep->esicmt == NULL)
-				vep->esicmt_p = vep->esicmt = "esi";
-			while (p < e) {
-				if (*p != *vep->esicmt_p) {
-					vep->esicmt_p = vep->esicmt = NULL;
-					vep->until_p = vep->until = "-->";
-					vep->until_s = VEP_NEXTTAG;
-					vep->state = VEP_UNTIL;
-					break;
-				}
-				p++;
-				if (*++vep->esicmt_p != '\0')
-					continue;
-				if (vep->remove)
-					vep_error(vep,
-					    "ESI 1.0 Nested <!--esi"
-					    " element in <esi:remove>");
-				vep->esicmt_p = vep->esicmt = "-->";
-				vep->state = VEP_NEXTTAG;
-				vep_mark_skip(vep, p);
-				break;
-			}
+			vep->esicmt_p = vep->esicmt = NULL;
+			vep->until_p = vep->until = "-->";
+			vep->until_s = VEP_NEXTTAG;
+			vep->state = VEP_UNTIL;
+		} else if (vep->state == VEP_COMMENTESI) {
+			if (vep->remove)
+				vep_error(vep,
+				    "ESI 1.0 Nested <!--esi"
+				    " element in <esi:remove>");
+			vep->esicmt_p = vep->esicmt = "-->";
+			vep->state = VEP_NEXTTAG;
+			vep_mark_skip(vep, p);
 		} else if (vep->state == VEP_CDATA) {
 			/*
 			 * Easy: just look for the end of CDATA
@@ -780,6 +784,9 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 			vep->until_p = vep->until = "]]>";
 			vep->until_s = VEP_NEXTTAG;
 			vep->state = VEP_UNTIL;
+		} else if (vep->state == VEP_ESIENDTAG) {
+			vep->endtag = 1;
+			vep->state = VEP_ESITAG;
 		} else if (vep->state == VEP_ESITAG) {
 			vep->in_esi_tag = 1;
 			vep->esi_found = 1;
@@ -932,8 +939,7 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 				vep->attr_delim = 0;
 				if (vep->attr_vsb != NULL) {
 					AZ(VSB_finish(vep->attr_vsb));
-					VSB_delete(vep->attr_vsb);
-					vep->attr_vsb = NULL;
+					VSB_destroy(&vep->attr_vsb);
 				}
 			} else if (p < e) {
 				vep->attr_delim = 0;
@@ -964,6 +970,7 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 				vep->match = NULL;
 				vep->tag_i = 0;
 			} else {
+				assert(p + sizeof(vep->tag) >= e);
 				memcpy(vep->tag, p, e - p);
 				vep->tag_i = e - p;
 				vep->state = VEP_MATCHBUF;
@@ -975,28 +982,33 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 			 * sections.
 			 */
 			AN(vep->match);
-			do {
-				if (*p == '>') {
-					for (vm = vep->match;
-					    vm->match != NULL; vm++)
-						continue;
-					AZ(vm->match);
-				} else {
-					vep->tag[vep->tag_i++] = *p++;
-					vm = vep_match(vep,
-					    vep->tag, vep->tag + vep->tag_i);
-					if (vm && vm->match == NULL) {
-						vep->tag_i--;
-						p--;
-					}
-				}
-			} while (vm == NULL && p < e);
-			vep->match_hit = vm;
+			i = sizeof(vep->tag) - vep->tag_i;
+			if (i > e - p)
+				i = e - p;
+			memcpy(vep->tag + vep->tag_i, p, i);
+			vm = vep_match(vep, vep->tag,
+			    vep->tag + vep->tag_i + i);
+			Debug("MB (%.*s) tag_i %d i %d = vm %p match %s\n",
+			    vep->tag_i + i, vep->tag,
+			    vep->tag_i,
+			    i,
+			    vm,
+			    vm ? vm->match : "(nil)");
+
 			if (vm == NULL) {
+				vep->tag_i += i;
+				p += i;
 				assert(p == e);
 			} else {
+				vep->match_hit = vm;
 				vep->state = *vm->state;
+				if (vm->match != NULL) {
+					i = strlen(vm->match);
+					if (i > vep->tag_i)
+						p += i - vep->tag_i;
+				}
 				vep->match = NULL;
+				vep->tag_i = 0;
 			}
 		} else if (vep->state == VEP_UNTIL) {
 			/*
@@ -1014,7 +1026,7 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 				vep_mark_verbatim(vep, p);
 		} else {
 			Debug("*** Unknown state %s\n", vep->state);
-			INCOMPL();
+			WRONG("WRONG ESI PARSER STATE");
 		}
 	}
 	/*
@@ -1035,7 +1047,7 @@ VEP_Parse(struct vep_state *vep, const char *p, size_t l)
 /*---------------------------------------------------------------------
  */
 
-static ssize_t __match_proto__()
+static ssize_t v_matchproto_(vep_callback_t)
 vep_default_cb(struct vfp_ctx *vc, void *priv, ssize_t l, enum vgz_flag flg)
 {
 	ssize_t *s;
@@ -1059,8 +1071,12 @@ VEP_Init(struct vfp_ctx *vc, const struct http *req, vep_callback_t *cb,
 
 	CHECK_OBJ_NOTNULL(vc, VFP_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(req, HTTP_MAGIC);
-	vep = WS_Alloc(vc->http->ws, sizeof *vep);
-	AN(vep);
+	vep = WS_Alloc(vc->resp->ws, sizeof *vep);
+	if (vep == NULL) {
+		VSLb(vc->wrk->vsl, SLT_VCL_Error,
+		     "VEP_Init() workspace overflow");
+		return (NULL);
+	}
 
 	INIT_OBJ(vep, VEP_MAGIC);
 	vep->url = req->hd[HTTP_HDR_URL].b;
@@ -1116,14 +1132,14 @@ VEP_Finish(struct vep_state *vep)
 		lcb = vep->cb(vep->vc, vep->cb_priv, 0, VGZ_ALIGN);
 		vep_emit_common(vep, lcb - vep->o_last, vep->last_mark);
 	}
-	// NB: We don't account for PAD+SUM+LEN in gzip'ed objects
+	// NB: We don't account for PAD+SUM+LEN in gzipped objects
 	(void)vep->cb(vep->vc, vep->cb_priv, 0, VGZ_FINISH);
 
 	AZ(VSB_finish(vep->vsb));
 	l = VSB_len(vep->vsb);
 	if (vep->esi_found && l > 0)
 		return (vep->vsb);
-	VSB_delete(vep->vsb);
+	VSB_destroy(&vep->vsb);
 	return (NULL);
 }
 
