@@ -145,7 +145,7 @@ EXP_RefNewObjcore(struct objcore *oc)
 
 
 /*--------------------------------------------------------------------
- * Call EXP's attention to a an oc
+ * Call EXP's attention to an oc
  */
 
 void
@@ -191,7 +191,7 @@ EXP_Insert(struct worker *wrk, struct objcore *oc)
 	if (!(oc->exp_flags & OC_EF_REFD))
 		return;
 
-	/* One ref held by the caller, and one that wil be owned by
+	/* One ref held by the caller, and one that will be owned by
 	 * expiry. */
 	assert(oc->refcnt >= 2);
 
@@ -213,12 +213,34 @@ EXP_Insert(struct worker *wrk, struct objcore *oc)
 		ObjSendEvent(wrk, oc, OEV_EXPIRE);
 		tmpoc = oc;
 		assert(oc->refcnt >= 2); /* Silence coverity */
-		(void)HSH_DerefObjCore(wrk, &oc, 0);
+		(void)HSH_DerefObjCore(wrk, &oc);
 		AZ(oc);
 		assert(tmpoc->refcnt >= 1); /* Silence coverity */
 	}
 }
 
+/*--------------------------------------------------------------------
+ * Reduce object timers
+ */
+
+void
+EXP_Reduce(struct objcore *oc, vtim_real now,
+    vtim_dur ttl, vtim_dur grace, vtim_dur keep)
+{
+
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	assert(oc->refcnt > 0);
+
+	if (!isnan(ttl) && now + ttl - oc->t_origin >= oc->ttl)
+		ttl = NAN;
+	if (!isnan(grace) && grace >= oc->grace)
+		grace = NAN;
+	if (!isnan(keep) && keep >= oc->keep)
+		keep = NAN;
+
+	EXP_Rearm(oc, now, ttl, grace, keep);
+}
+
 /*--------------------------------------------------------------------
  * We have changed one or more of the object timers, tell the exp_thread
  *
@@ -265,7 +287,7 @@ EXP_Rearm(struct objcore *oc, vtim_real now,
  */
 
 static void
-exp_inbox(struct exp_priv *ep, struct objcore *oc, unsigned flags, double now)
+exp_inbox(struct exp_priv *ep, struct objcore *oc, unsigned flags, vtim_real now)
 {
 
 	CHECK_OBJ_NOTNULL(ep, EXP_PRIV_MAGIC);
@@ -287,7 +309,7 @@ exp_inbox(struct exp_priv *ep, struct objcore *oc, unsigned flags, double now)
 		    VXID(ObjGetXID(ep->wrk, oc)), EXP_Ttl(NULL, oc) - now,
 		    (intmax_t)oc->hits);
 		ObjSendEvent(ep->wrk, oc, OEV_EXPIRE);
-		(void)HSH_DerefObjCore(ep->wrk, &oc, 0);
+		(void)HSH_DerefObjCore(ep->wrk, &oc);
 		return;
 	}
 
@@ -365,7 +387,7 @@ exp_expire(struct exp_priv *ep, vtim_real now)
 		    VXID(ObjGetXID(ep->wrk, oc)), EXP_Ttl(NULL, oc) - now,
 		    (intmax_t)oc->hits);
 		ObjSendEvent(ep->wrk, oc, OEV_EXPIRE);
-		(void)HSH_DerefObjCore(ep->wrk, &oc, 0);
+		(void)HSH_DerefObjCore(ep->wrk, &oc);
 	}
 	return (0);
 }
@@ -407,6 +429,8 @@ exp_thread(struct worker *wrk, void *priv)
 	CAST_OBJ_NOTNULL(ep, priv, EXP_PRIV_MAGIC);
 	ep->wrk = wrk;
 	VSL_Setup(&ep->vsl, NULL, 0);
+	AZ(wrk->vsl);
+	wrk->vsl = &ep->vsl;
 	ep->heap = VBH_new(NULL, object_cmp, object_update);
 	AN(ep->heap);
 	while (exp_shutdown == 0) {
@@ -416,6 +440,7 @@ exp_thread(struct worker *wrk, void *priv)
 		CHECK_OBJ_ORNULL(oc, OBJCORE_MAGIC);
 		if (oc != NULL) {
 			assert(oc->refcnt >= 1);
+			assert(oc->exp_flags & OC_EF_POSTED);
 			VSTAILQ_REMOVE(&ep->inbox, oc, objcore, exp_list);
 			VSC_C_main->exp_received++;
 			tnext = 0;
@@ -438,6 +463,7 @@ exp_thread(struct worker *wrk, void *priv)
 		else
 			tnext = exp_expire(ep, t);
 	}
+	wrk->vsl = NULL;
 	return (NULL);
 }
 
