@@ -86,6 +86,7 @@
 
 #include "cache_varnishd.h"
 #include "cache_obj.h"
+#include "cache_objhead.h"
 #include "vend.h"
 #include "storage/storage.h"
 
@@ -109,7 +110,6 @@ obj_newboc(void)
 	Lck_New(&boc->mtx, lck_busyobj);
 	PTOK(pthread_cond_init(&boc->cond, NULL));
 	boc->refcount = 1;
-	boc->transit_buffer = cache_param->transit_buffer;
 	return (boc);
 }
 
@@ -141,8 +141,6 @@ ObjNew(const struct worker *wrk)
 	AN(oc);
 	wrk->stats->n_objectcore++;
 	oc->last_lru = NAN;
-	oc->flags = OC_F_BUSY;
-
 	oc->boc = obj_newboc();
 
 	return (oc);
@@ -183,6 +181,130 @@ ObjIterate(struct worker *wrk, struct objcore *oc,
 	return (om->objiterator(wrk, oc, priv, func, final));
 }
 
+/*====================================================================
+ * ObjVAI...(): Asynchronous Iteration
+ *
+ *
+ * ObjVAIinit() returns an opaque handle, or NULL if not supported
+ *
+ *	A VAI handle must not be used concurrently
+ *
+ *	the vai_notify_cb(priv) will be called asynchronously by the storage
+ *	engine when a -EAGAIN / -ENOBUFS condition is over and ObjVAIlease()
+ *	can be called again.
+ *
+ *	Note:
+ *	- the callback gets executed by an arbitrary thread
+ *	- WITH the boc mtx held
+ *	so it should never block and only do minimal work
+ *
+ * ObjVAIlease() fills the vscarab with leases. returns:
+ *
+ *	-EAGAIN:  nothing available at the moment, storage will notify, no use to
+ *		  call again until notification
+ *	-ENOBUFS: caller needs to return leases, storage will notify
+ *	-EPIPE:	  BOS_FAILED for busy object
+ *	-(errno): other problem, fatal
+ *
+ *	>= 0:	  number of viovs added (== scarab->capacity - scarab->used)
+ *
+ *	struct vscarab:
+ *
+ *	the leases can be used by the caller until returned with
+ *	ObjVAIreturn(). The storage guarantees that the lease member is a
+ *	multiple of 8 (that is, the lower three bits are zero). These can be
+ *	used by the caller between lease and return, but must be cleared to
+ *	zero before returning.
+ *
+ * ObjVAIbuffer() allocates temporary buffers, returns:
+ *
+ *	-EAGAIN:  allocation can not be fulfilled immediately, storage will notify,
+ *		  no use to call again until notification
+ *	-EINVAL:  size larger than UINT_MAX requested
+ *	-(errno): other problem, fatal
+ *	n:	  n > 0, number of viovs filled
+ *
+ *	The struct vscarab is used on the way in and out: On the way in, the
+ *	iov.iov_len members contain the sizes the caller requests, all other
+ *	members of the struct viovs are expected to be zero initialized.
+ *
+ *	The maximum size to be requested is UINT_MAX.
+ *
+ *	ObjVAIbuffer() may return sizes larger than requested. The returned n
+ *	might be smaller than requested.
+ *
+ * ObjVAIreturn() returns leases collected in a struct vscaret
+ *
+ *	it must be called with a vscaret, which holds an array of lease values
+ *	received via ObjVAIlease() or ObjVAIbuffer() when the caller can
+ *	guarantee that they are no longer accessed.
+ *
+ *	ObjVAIreturn() may retain leases in the vscaret if the implementation
+ *	still requires them, iow, the vscaret might not be empty upon return.
+ *
+ * ObjVAIfini() finalized iteration
+ *
+ *	it must be called when iteration is done, irrespective of error status
+ */
+
+vai_hdl
+ObjVAIinit(struct worker *wrk, struct objcore *oc, struct ws *ws,
+    vai_notify_cb *cb, void *cb_priv)
+{
+	const struct obj_methods *om = obj_getmethods(oc);
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+
+	if (om->vai_init == NULL)
+		return (NULL);
+	return (om->vai_init(wrk, oc, ws, cb, cb_priv));
+}
+
+int
+ObjVAIlease(struct worker *wrk, vai_hdl vhdl, struct vscarab *scarab)
+{
+	struct vai_hdl_preamble *vaip = vhdl;
+
+	AN(vaip);
+	assert(vaip->magic2 == VAI_HDL_PREAMBLE_MAGIC2);
+	AN(vaip->vai_lease);
+	return (vaip->vai_lease(wrk, vhdl, scarab));
+}
+
+int
+ObjVAIbuffer(struct worker *wrk, vai_hdl vhdl, struct vscarab *scarab)
+{
+	struct vai_hdl_preamble *vaip = vhdl;
+
+	AN(vaip);
+	assert(vaip->magic2 == VAI_HDL_PREAMBLE_MAGIC2);
+	AN(vaip->vai_buffer);
+	return (vaip->vai_buffer(wrk, vhdl, scarab));
+}
+
+void
+ObjVAIreturn(struct worker *wrk, vai_hdl vhdl, struct vscaret *scaret)
+{
+	struct vai_hdl_preamble *vaip = vhdl;
+
+	AN(vaip);
+	assert(vaip->magic2 == VAI_HDL_PREAMBLE_MAGIC2);
+	AN(vaip->vai_return);
+	vaip->vai_return(wrk, vhdl, scaret);
+}
+
+void
+ObjVAIfini(struct worker *wrk, vai_hdl *vhdlp)
+{
+	AN(vhdlp);
+	struct vai_hdl_preamble *vaip = *vhdlp;
+
+	AN(vaip);
+	assert(vaip->magic2 == VAI_HDL_PREAMBLE_MAGIC2);
+	AN(vaip->vai_lease);
+	vaip->vai_fini(wrk, vhdlp);
+}
+
 /*====================================================================
  * ObjGetSpace()
  *
@@ -225,18 +347,33 @@ obj_extend_condwait(const struct objcore *oc)
 	if (oc->boc->transit_buffer == 0)
 		return;
 
-	assert(oc->flags & (OC_F_PRIVATE | OC_F_HFM | OC_F_HFP));
-	/* NB: strictly signaling progress both ways would be prone to
-	 * deadlocks, so instead we wait for signals from the client side
-	 * when delivered_so_far so far is updated, but in case the fetch
-	 * thread was not waiting at the time of the signal, we will see
-	 * updates to delivered_so_far after timing out.
-	 */
+	assert(oc->flags & OC_F_TRANSIENT);
 	while (!(oc->flags & OC_F_CANCEL) && oc->boc->fetched_so_far >
 	    oc->boc->delivered_so_far + oc->boc->transit_buffer)
 		(void)Lck_CondWait(&oc->boc->cond, &oc->boc->mtx);
 }
 
+// notify of an extension of the boc or state change
+
+static void
+obj_boc_notify(struct boc *boc)
+{
+	struct vai_qe *qe, *next;
+
+	PTOK(pthread_cond_broadcast(&boc->cond));
+	qe = VSLIST_FIRST(&boc->vai_q_head);
+	VSLIST_FIRST(&boc->vai_q_head) = NULL;
+	while (qe != NULL) {
+		CHECK_OBJ(qe, VAI_Q_MAGIC);
+		AN(qe->flags & VAI_QF_INQUEUE);
+		qe->flags &= ~VAI_QF_INQUEUE;
+		next = VSLIST_NEXT(qe, list);
+		VSLIST_NEXT(qe, list) = NULL;
+		qe->cb(qe->hdl, qe->priv);
+		qe = next;
+	}
+}
+
 void
 ObjExtend(struct worker *wrk, struct objcore *oc, ssize_t l, int final)
 {
@@ -247,16 +384,21 @@ ObjExtend(struct worker *wrk, struct objcore *oc, ssize_t l, int final)
 	AN(om->objextend);
 	assert(l >= 0);
 
-	Lck_Lock(&oc->boc->mtx);
 	if (l > 0) {
+		Lck_Lock(&oc->boc->mtx);
 		obj_extend_condwait(oc);
 		om->objextend(wrk, oc, l);
 		oc->boc->fetched_so_far += l;
-		PTOK(pthread_cond_broadcast(&oc->boc->cond));
-	}
-	Lck_Unlock(&oc->boc->mtx);
+		obj_boc_notify(oc->boc);
+		Lck_Unlock(&oc->boc->mtx);
 
-	assert(oc->boc == NULL || oc->boc->state < BOS_FINISHED);
+		if (oc->boc->transit_buffer > 0)
+			wrk->stats->transit_buffered += l;
+		else if (oc->flags & OC_F_TRANSIENT)
+			wrk->stats->transit_stored += l;
+	}
+
+	assert(oc->boc->state < BOS_FINISHED);
 	if (final && om->objtrimstore != NULL)
 		om->objtrimstore(wrk, oc);
 }
@@ -264,9 +406,22 @@ ObjExtend(struct worker *wrk, struct objcore *oc, ssize_t l, int final)
 /*====================================================================
  */
 
-uint64_t
-ObjWaitExtend(const struct worker *wrk, const struct objcore *oc, uint64_t l)
+static inline void
+objSignalFetchLocked(const struct objcore *oc, uint64_t l)
 {
+	if (oc->boc->transit_buffer > 0) {
+		assert(oc->flags & OC_F_TRANSIENT);
+		/* Signal the new client position */
+		oc->boc->delivered_so_far = l;
+		PTOK(pthread_cond_signal(&oc->boc->cond));
+	}
+}
+
+uint64_t
+ObjWaitExtend(const struct worker *wrk, const struct objcore *oc, uint64_t l,
+    enum boc_state_e *statep)
+{
+	enum boc_state_e state;
 	uint64_t rv;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
@@ -276,27 +431,68 @@ ObjWaitExtend(const struct worker *wrk, const struct objcore *oc, uint64_t l)
 	while (1) {
 		rv = oc->boc->fetched_so_far;
 		assert(l <= rv || oc->boc->state == BOS_FAILED);
-		if (oc->boc->transit_buffer > 0) {
-			assert(oc->flags & (OC_F_PRIVATE | OC_F_HFM | OC_F_HFP));
-			/* Signal the new client position */
-			oc->boc->delivered_so_far = l;
-			PTOK(pthread_cond_signal(&oc->boc->cond));
-		}
-		if (rv > l || oc->boc->state >= BOS_FINISHED)
+		state = oc->boc->state;
+		objSignalFetchLocked(oc, l);
+		if (rv > l || state >= BOS_FINISHED)
 			break;
 		(void)Lck_CondWait(&oc->boc->cond, &oc->boc->mtx);
 	}
-	rv = oc->boc->fetched_so_far;
 	Lck_Unlock(&oc->boc->mtx);
+	if (statep != NULL)
+		*statep = state;
 	return (rv);
 }
 
+// get a new extension _or_ register a notification
+uint64_t
+ObjVAIGetExtend(struct worker *wrk, const struct objcore *oc, uint64_t l,
+    enum boc_state_e *statep, struct vai_qe *qe)
+{
+	enum boc_state_e state;
+	uint64_t rv;
+
+	(void) wrk;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	CHECK_OBJ_NOTNULL(oc->boc, BOC_MAGIC);
+	CHECK_OBJ_NOTNULL(qe, VAI_Q_MAGIC);
+	Lck_Lock(&oc->boc->mtx);
+	rv = oc->boc->fetched_so_far;
+	assert(l <= rv || oc->boc->state == BOS_FAILED);
+	state = oc->boc->state;
+	objSignalFetchLocked(oc, l);
+	if (l == rv && state < BOS_FINISHED &&
+	    (qe->flags & VAI_QF_INQUEUE) == 0) {
+		qe->flags |= VAI_QF_INQUEUE;
+		VSLIST_INSERT_HEAD(&oc->boc->vai_q_head, qe, list);
+	}
+	Lck_Unlock(&oc->boc->mtx);
+	if (statep != NULL)
+		*statep = state;
+	return (rv);
+}
+
+void
+ObjVAICancel(struct worker *wrk, struct boc *boc, struct vai_qe *qe)
+{
+
+	(void) wrk;
+	CHECK_OBJ_NOTNULL(boc, BOC_MAGIC);
+	CHECK_OBJ_NOTNULL(qe, VAI_Q_MAGIC);
+
+	Lck_Lock(&boc->mtx);
+	// inefficient, but should be rare
+	if ((qe->flags & VAI_QF_INQUEUE) != 0)
+		VSLIST_REMOVE(&boc->vai_q_head, qe, vai_qe, list);
+	qe->flags = 0;
+	Lck_Unlock(&boc->mtx);
+}
+
 /*====================================================================
  */
 
 void
-ObjSetState(struct worker *wrk, const struct objcore *oc,
-    enum boc_state_e next)
+ObjSetState(struct worker *wrk, struct objcore *oc, enum boc_state_e next,
+    unsigned broadcast)
 {
 	const struct obj_methods *om;
 
@@ -305,7 +501,6 @@ ObjSetState(struct worker *wrk, const struct objcore *oc,
 	assert(next > oc->boc->state);
 
 	CHECK_OBJ_ORNULL(oc->stobj->stevedore, STEVEDORE_MAGIC);
-	assert(next != BOS_STREAM || oc->boc->state == BOS_PREP_STREAM);
 	assert(next != BOS_FINISHED || (oc->oa_present & (1 << OA_LEN)));
 
 	if (oc->stobj->stevedore != NULL) {
@@ -314,9 +509,15 @@ ObjSetState(struct worker *wrk, const struct objcore *oc,
 			om->objsetstate(wrk, oc, next);
 	}
 
+	if (next == BOS_FAILED)
+		HSH_Fail(wrk, oc);
+	else if (oc->boc->state < BOS_STREAM && next >= BOS_STREAM)
+		HSH_Unbusy(wrk, oc);
+
 	Lck_Lock(&oc->boc->mtx);
 	oc->boc->state = next;
-	PTOK(pthread_cond_broadcast(&oc->boc->cond));
+	if (broadcast)
+		obj_boc_notify(oc->boc);
 	Lck_Unlock(&oc->boc->mtx);
 }
 
@@ -655,7 +856,7 @@ ObjSetFlag(struct worker *wrk, struct objcore *oc, enum obj_flags of, int val)
 }
 
 /*====================================================================
- * Object event subscribtion mechanism.
+ * Object event subscription mechanism.
  *
  * XXX: it is extremely unclear what the locking circumstances are here.
  */
