@@ -66,7 +66,8 @@
   REQ_STEP(deliver,             DELIVER,	static) \
   REQ_STEP(vclfail,             VCLFAIL,	static) \
   REQ_STEP(synth,               SYNTH,		static) \
-  REQ_STEP(transmit,            TRANSMIT,	static)
+  REQ_STEP(transmit,            TRANSMIT,	static) \
+  REQ_STEP(finish,              FINISH,		static)
 
 #define REQ_STEP(l, U, priv) \
     static req_state_f cnt_##l; \
@@ -136,6 +137,7 @@ Resp_Setup_Deliver(struct req *req)
 {
 	struct http *h;
 	struct objcore *oc;
+	const void *hdrs;
 
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 	oc = req->objcore;
@@ -145,7 +147,8 @@ Resp_Setup_Deliver(struct req *req)
 
 	HTTP_Setup(h, req->ws, req->vsl, SLT_RespMethod);
 
-	if (HTTP_Decode(h, ObjGetAttr(req->wrk, oc, OA_HEADERS, NULL)))
+	hdrs = ObjGetAttr(req->wrk, oc, OA_HEADERS, NULL);
+	if (hdrs == NULL || HTTP_Decode(h, hdrs))
 		return (-1);
 
 	http_ForceField(h, HTTP_HDR_PROTO, "HTTP/1.1");
@@ -219,7 +222,7 @@ cnt_deliver(struct worker *wrk, struct req *req)
 	ObjTouch(req->wrk, req->objcore, req->t_prev);
 
 	if (Resp_Setup_Deliver(req)) {
-		(void)HSH_DerefObjCore(wrk, &req->objcore, HSH_RUSH_POLICY);
+		(void)HSH_DerefObjCore(wrk, &req->objcore);
 		req->err_code = 500;
 		req->req_step = R_STP_SYNTH;
 		return (REQ_FSM_MORE);
@@ -233,11 +236,11 @@ cnt_deliver(struct worker *wrk, struct req *req)
 	req->t_resp = W_TIM_real(wrk);
 	VCL_deliver_method(req->vcl, wrk, req, NULL, NULL);
 
-	assert(req->restarts <= cache_param->max_restarts);
+	assert(req->restarts <= req->max_restarts);
 
 	if (wrk->vpi->handling != VCL_RET_DELIVER) {
 		HSH_Cancel(wrk, req->objcore, NULL);
-		(void)HSH_DerefObjCore(wrk, &req->objcore, HSH_RUSH_POLICY);
+		(void)HSH_DerefObjCore(wrk, &req->objcore);
 		http_Teardown(req->resp);
 
 		switch (wrk->vpi->handling) {
@@ -353,8 +356,7 @@ cnt_synth(struct worker *wrk, struct req *req)
 		return (REQ_FSM_DONE);
 	}
 
-	if (wrk->vpi->handling == VCL_RET_RESTART &&
-	    req->restarts > cache_param->max_restarts)
+	if (wrk->vpi->handling == VCL_RET_RESTART && req->restarts > req->max_restarts)
 		wrk->vpi->handling = VCL_RET_DELIVER;
 
 	if (wrk->vpi->handling == VCL_RET_RESTART) {
@@ -374,7 +376,8 @@ cnt_synth(struct worker *wrk, struct req *req)
 	    VSB_len(synth_body));
 
 	if (req->doclose == SC_NULL &&
-	    http_HdrIs(req->resp, H_Connection, "close"))
+	    (http_HdrIs(req->resp, H_Connection, "close") ||
+	    http_HdrIs(req->http, H_Connection, "close")))
 		req->doclose = SC_RESP_CLOSE;
 
 	/* Discard any lingering request body before delivery */
@@ -411,7 +414,7 @@ cnt_synth(struct worker *wrk, struct req *req)
 		VSLb(req->vsl, SLT_Error, "Could not get storage");
 		req->doclose = SC_OVERLOAD;
 		VSLb_ts_req(req, "Resp", W_TIM_real(wrk));
-		(void)HSH_DerefObjCore(wrk, &req->objcore, 1);
+		(void)HSH_DerefObjCore(wrk, &req->objcore);
 		http_Teardown(req->resp);
 		return (REQ_FSM_DONE);
 	}
@@ -427,7 +430,8 @@ cnt_synth(struct worker *wrk, struct req *req)
 static enum req_fsm_nxt v_matchproto_(req_state_f)
 cnt_transmit(struct worker *wrk, struct req *req)
 {
-	struct boc *boc;
+	enum req_fsm_nxt nxt = REQ_FSM_MORE;
+	enum vtr_deliver_e dnxt;
 	uint16_t status;
 	int sendbody, head;
 	intmax_t clval;
@@ -437,18 +441,20 @@ cnt_transmit(struct worker *wrk, struct req *req)
 	CHECK_OBJ_NOTNULL(req->transport, TRANSPORT_MAGIC);
 	CHECK_OBJ_NOTNULL(req->objcore, OBJCORE_MAGIC);
 	AZ(req->stale_oc);
-	AZ(req->res_mode);
+	AZ(req->res_pipe | req->res_esi);
+	AZ(req->boc);
+	req->req_step = R_STP_FINISH;
 
 	/* Grab a ref to the bo if there is one (=streaming) */
-	boc = HSH_RefBoc(req->objcore);
-	if (boc && boc->state < BOS_STREAM)
+	req->boc = HSH_RefBoc(req->objcore);
+	if (req->boc && req->boc->state < BOS_STREAM)
 		ObjWaitState(req->objcore, BOS_STREAM);
 	clval = http_GetContentLength(req->resp);
 	/* RFC 7230, 3.3.3 */
 	status = http_GetStatus(req->resp);
 	head = http_method_eq(req->http0->hd[HTTP_HDR_METHOD].b, HEAD);
 
-	if (boc != NULL || (req->objcore->flags & (OC_F_FAILED)))
+	if (req->boc != NULL || (req->objcore->flags & (OC_F_FAILED)))
 		req->resp_len = clval;
 	else
 		req->resp_len = ObjGetLen(req->wrk, req->objcore);
@@ -460,15 +466,15 @@ cnt_transmit(struct worker *wrk, struct req *req)
 		sendbody = 1;
 	}
 
-	VDP_Init(req->vdc, req->wrk, req->vsl, req);
+	VDP_Init(req->vdc, req->wrk, req->vsl, req, NULL, &req->resp_len);
 	if (req->vdp_filter_list == NULL)
 		req->vdp_filter_list = resp_Get_Filter_List(req);
 	if (req->vdp_filter_list == NULL ||
-	    VCL_StackVDP(req, req->vcl, req->vdp_filter_list)) {
+	    VCL_StackVDP(req->vdc, req->vcl, req->vdp_filter_list, req, NULL)) {
 		VSLb(req->vsl, SLT_Error, "Failure to push processors");
 		req->doclose = SC_OVERLOAD;
 		req->acct.resp_bodybytes +=
-			VDP_Close(req->vdc, req->objcore, boc);
+			VDP_Close(req->vdc, req->objcore, req->boc);
 	} else {
 		if (status < 200 || status == 204) {
 			// rfc7230,l,1691,1695
@@ -478,15 +484,12 @@ cnt_transmit(struct worker *wrk, struct req *req)
 			http_Unset(req->resp, H_Content_Length);
 		} else if (clval >= 0 && clval == req->resp_len) {
 			/* Reuse C-L header */
-		} else if (head && req->objcore->flags & OC_F_HFM) {
-			/*
-			 * Don't touch C-L header (debatable)
-			 *
-			 * The only way to do it correctly would be to GET
-			 * to the backend, and discard the body once the
-			 * filters have had a chance to chew on it, but that
-			 * would negate the "pass for huge objects" use case.
+		} else if (head) {
+			/* rfc9110,l,3226,3227
+			 * "MAY send Content-Length ... [for] HEAD"
+			 * do not touch to support cached HEAD #4245
 			 */
+			req->resp_len = 0;
 		} else {
 			http_Unset(req->resp, H_Content_Length);
 			if (req->resp_len >= 0)
@@ -495,8 +498,21 @@ cnt_transmit(struct worker *wrk, struct req *req)
 		}
 		if (req->resp_len == 0)
 			sendbody = 0;
-		req->transport->deliver(req, boc, sendbody);
+		dnxt = req->transport->deliver(req, sendbody);
+		if (dnxt == VTR_D_DISEMBARK)
+			nxt = REQ_FSM_DISEMBARK;
+		else
+			assert(dnxt == VTR_D_DONE);
 	}
+	return (nxt);
+}
+
+static enum req_fsm_nxt v_matchproto_(req_state_f)
+cnt_finish(struct worker *wrk, struct req *req)
+{
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 
 	VSLb_ts_req(req, "Resp", W_TIM_real(wrk));
 
@@ -506,14 +522,17 @@ cnt_transmit(struct worker *wrk, struct req *req)
 		req->doclose = SC_TX_ERROR;
 	}
 
-	if (boc != NULL)
+	if (req->boc != NULL) {
 		HSH_DerefBoc(wrk, req->objcore);
+		req->boc = NULL;
+	}
 
-	(void)HSH_DerefObjCore(wrk, &req->objcore, HSH_RUSH_POLICY);
+	(void)HSH_DerefObjCore(wrk, &req->objcore);
 	http_Teardown(req->resp);
 
 	req->vdp_filter_list = NULL;
-	req->res_mode = 0;
+	req->res_pipe = 0;
+	req->res_esi = 0;
 	return (REQ_FSM_DONE);
 }
 
@@ -536,7 +555,7 @@ cnt_fetch(struct worker *wrk, struct req *req)
 	if (req->objcore->flags & OC_F_FAILED) {
 		req->err_code = 503;
 		req->req_step = R_STP_SYNTH;
-		(void)HSH_DerefObjCore(wrk, &req->objcore, 1);
+		(void)HSH_DerefObjCore(wrk, &req->objcore);
 		AZ(req->objcore);
 		return (REQ_FSM_MORE);
 	}
@@ -555,20 +574,23 @@ cnt_lookup(struct worker *wrk, struct req *req)
 {
 	struct objcore *oc, *busy;
 	enum lookup_e lr;
-	int had_objhead = 0;
+	int waitinglist_gen;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-	AZ(req->objcore);
 	AZ(req->stale_oc);
 
 	AN(req->vcl);
 
 	VRY_Prep(req);
+	waitinglist_gen = req->waitinglist_gen;
+
+	if (req->waitinglist_gen) {
+		CHECK_OBJ_NOTNULL(req->objcore, OBJCORE_MAGIC);
+		req->waitinglist_gen = 0;
+	} else
+		AZ(req->objcore);
 
-	AZ(req->objcore);
-	if (req->hash_objhead)
-		had_objhead = 1;
 	wrk->strangelove = 0;
 	lr = HSH_Lookup(req, &oc, &busy);
 	if (lr == HSH_BUSY) {
@@ -584,7 +606,7 @@ cnt_lookup(struct worker *wrk, struct req *req)
 	if ((unsigned)wrk->strangelove >= cache_param->vary_notice)
 		VSLb(req->vsl, SLT_Notice, "vsl: High number of variants (%d)",
 		    wrk->strangelove);
-	if (had_objhead)
+	if (waitinglist_gen)
 		VSLb_ts_req(req, "Waitinglist", W_TIM_real(wrk));
 
 	if (req->vcf != NULL) {
@@ -664,10 +686,10 @@ cnt_lookup(struct worker *wrk, struct req *req)
 	}
 
 	/* Drop our object, we won't need it */
-	(void)HSH_DerefObjCore(wrk, &req->objcore, HSH_RUSH_POLICY);
+	(void)HSH_DerefObjCore(wrk, &req->objcore);
 
 	if (busy != NULL) {
-		(void)HSH_DerefObjCore(wrk, &busy, 0);
+		HSH_Withdraw(wrk, &busy);
 		VRY_Clear(req);
 	}
 
@@ -694,7 +716,7 @@ cnt_miss(struct worker *wrk, struct req *req)
 		wrk->stats->cache_miss++;
 		VBF_Fetch(wrk, req, req->objcore, req->stale_oc, VBF_NORMAL);
 		if (req->stale_oc != NULL)
-			(void)HSH_DerefObjCore(wrk, &req->stale_oc, 0);
+			(void)HSH_DerefObjCore(wrk, &req->stale_oc);
 		req->req_step = R_STP_FETCH;
 		return (REQ_FSM_MORE);
 	case VCL_RET_FAIL:
@@ -714,8 +736,8 @@ cnt_miss(struct worker *wrk, struct req *req)
 	}
 	VRY_Clear(req);
 	if (req->stale_oc != NULL)
-		(void)HSH_DerefObjCore(wrk, &req->stale_oc, 0);
-	AZ(HSH_DerefObjCore(wrk, &req->objcore, 1));
+		(void)HSH_DerefObjCore(wrk, &req->stale_oc);
+	HSH_Withdraw(wrk, &req->objcore);
 	return (REQ_FSM_MORE);
 }
 
@@ -849,7 +871,7 @@ cnt_restart(struct worker *wrk, struct req *req)
 	AZ(req->objcore);
 	AZ(req->stale_oc);
 
-	if (++req->restarts > cache_param->max_restarts) {
+	if (++req->restarts > req->max_restarts) {
 		VSLb(req->vsl, SLT_VCL_Error, "Too many restarts");
 		req->err_code = 503;
 		req->req_step = R_STP_SYNTH;
@@ -907,6 +929,8 @@ cnt_recv_prep(struct req *req, const char *ci)
 	req->is_hitpass = 0;
 	req->err_code = 0;
 	req->err_reason = NULL;
+
+	req->vfp_filter_list = NULL;
 }
 
 /*--------------------------------------------------------------------
@@ -939,6 +963,7 @@ cnt_recv(struct worker *wrk, struct req *req)
 	if (http_CountHdr(req->http0, H_Host) > 1) {
 		VSLb(req->vsl, SLT_BogoHeader, "Multiple Host: headers");
 		wrk->stats->client_req_400++;
+		req->doclose = SC_RX_BAD;
 		(void)req->transport->minimal_response(req, 400);
 		return (REQ_FSM_DONE);
 	}
@@ -946,6 +971,7 @@ cnt_recv(struct worker *wrk, struct req *req)
 	if (http_CountHdr(req->http0, H_Content_Length) > 1) {
 		VSLb(req->vsl, SLT_BogoHeader, "Multiple Content-Length: headers");
 		wrk->stats->client_req_400++;
+		req->doclose = SC_RX_BAD;
 		(void)req->transport->minimal_response(req, 400);
 		return (REQ_FSM_DONE);
 	}
@@ -982,6 +1008,7 @@ cnt_recv(struct worker *wrk, struct req *req)
 	/* Attempts to cache req.body may fail */
 	if (req->req_body_status == BS_ERROR) {
 		req->doclose = SC_RX_BODY;
+		(void)req->transport->minimal_response(req, 400);
 		return (REQ_FSM_DONE);
 	}
 
@@ -1085,7 +1112,7 @@ cnt_purge(struct worker *wrk, struct req *req)
 
 	(void)HSH_Purge(wrk, boc->objhead, req->t_req, 0, 0, 0);
 
-	AZ(HSH_DerefObjCore(wrk, &boc, 1));
+	HSH_Withdraw(wrk, &boc);
 
 	VCL_purge_method(req->vcl, wrk, req, NULL, NULL);
 	switch (wrk->vpi->handling) {
@@ -1163,6 +1190,7 @@ CNT_Request(struct req *req)
 	 */
 	assert(
 	    req->req_step == R_STP_LOOKUP ||
+	    req->req_step == R_STP_FINISH ||
 	    req->req_step == R_STP_TRANSPORT);
 
 	AN(VXID_TAG(req->vsl->wid) & VSL_CLIENTMARKER);
