@@ -147,42 +147,6 @@ VRT_HDR_LR(beresp, reason,	HTTP_HDR_REASON)
 VRT_STATUS_L(beresp)
 VRT_STATUS_R(beresp)
 
-/*--------------------------------------------------------------------
- * Pulling things out of the packed object->http
- */
-
-VCL_INT
-VRT_r_obj_status(VRT_CTX)
-{
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
-
-	return (HTTP_GetStatusPack(ctx->req->wrk, ctx->req->objcore));
-}
-
-VCL_STRING
-VRT_r_obj_proto(VRT_CTX)
-{
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
-
-	return (HTTP_GetHdrPack(ctx->req->wrk, ctx->req->objcore,
-	    H__Proto));
-}
-
-VCL_STRING
-VRT_r_obj_reason(VRT_CTX)
-{
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
-
-	return (HTTP_GetHdrPack(ctx->req->wrk, ctx->req->objcore,
-	    H__Reason));
-}
-
 /*--------------------------------------------------------------------
  * beresp bool-fields
  */
@@ -530,28 +494,74 @@ VRT_l_beresp_storage_hint(VRT_CTX, const char *str, VCL_STRANDS s)
 
 /*--------------------------------------------------------------------*/
 
-VCL_STEVEDORE
-VRT_r_obj_storage(VRT_CTX)
-{
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
-	AN(ctx->req->objcore->stobj);
-	CHECK_OBJ_NOTNULL(ctx->req->objcore->stobj->stevedore,
-	    STEVEDORE_MAGIC);
-	return (ctx->req->objcore->stobj->stevedore);
+#define VRT_OC_VAR_R(obj, which, which_magic, field)		\
+VCL_STEVEDORE							\
+VRT_r_##obj##_storage(VRT_CTX)					\
+{								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	CHECK_OBJ_NOTNULL(ctx->which, which_magic);		\
+	CHECK_OBJ_NOTNULL(ctx->which->field, OBJCORE_MAGIC);	\
+	AN(ctx->which->field->stobj);				\
+	CHECK_OBJ_NOTNULL(ctx->which->field->stobj->stevedore,	\
+	    STEVEDORE_MAGIC);					\
+	return (ctx->which->field->stobj->stevedore);		\
+}								\
+								\
+VCL_BOOL							\
+VRT_r_##obj##_can_esi(VRT_CTX)					\
+{								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	CHECK_OBJ_NOTNULL(ctx->which, which_magic);		\
+	CHECK_OBJ_NOTNULL(ctx->which->field, OBJCORE_MAGIC);	\
+	return (ObjHasAttr(ctx->which->wrk, ctx->which->field,	\
+	    OA_ESIDATA));					\
+}								\
+								\
+VCL_BOOL							\
+VRT_r_##obj##_uncacheable(VRT_CTX)				\
+{								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	CHECK_OBJ_NOTNULL(ctx->which, which_magic);		\
+	CHECK_OBJ_NOTNULL(ctx->which->field, OBJCORE_MAGIC);	\
+								\
+	return (ctx->which->field->flags & OC_F_HFM ? 1 : 0);	\
+}								\
+								\
+VCL_INT								\
+VRT_r_##obj##_status(VRT_CTX)					\
+{								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	CHECK_OBJ_NOTNULL(ctx->which, which_magic);		\
+	CHECK_OBJ_NOTNULL(ctx->which->field, OBJCORE_MAGIC);	\
+								\
+	return (HTTP_GetStatusPack(ctx->which->wrk,		\
+	    ctx->which->field));				\
+}								\
+								\
+VCL_STRING							\
+VRT_r_##obj##_proto(VRT_CTX)					\
+{								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	CHECK_OBJ_NOTNULL(ctx->which, which_magic);		\
+	CHECK_OBJ_NOTNULL(ctx->which->field, OBJCORE_MAGIC);	\
+								\
+	return (HTTP_GetHdrPack(ctx->which->wrk,		\
+	    ctx->which->field, H__Proto));			\
+}								\
+								\
+VCL_STRING							\
+VRT_r_##obj##_reason(VRT_CTX)					\
+{								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	CHECK_OBJ_NOTNULL(ctx->which, which_magic);		\
+	CHECK_OBJ_NOTNULL(ctx->which->field, OBJCORE_MAGIC);	\
+								\
+	return (HTTP_GetHdrPack(ctx->which->wrk,		\
+	    ctx->which->field, H__Reason));			\
 }
 
-/*--------------------------------------------------------------------*/
-
-VCL_BOOL
-VRT_r_obj_can_esi(VRT_CTX)
-{
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
-	return (ObjHasAttr(ctx->req->wrk, ctx->req->objcore, OA_ESIDATA));
-}
+VRT_OC_VAR_R(obj, req, REQ_MAGIC, objcore)
+VRT_OC_VAR_R(obj_stale, bo, BUSYOBJ_MAGIC, stale_oc)
 
 /*--------------------------------------------------------------------*/
 
@@ -576,12 +586,24 @@ VRT_r_req_##nm(VRT_CTX)							\
 	return (ctx->req->elem);					\
 }
 
+#define REQ_VAR_U(nm, elem, val)					\
+									\
+VCL_VOID								\
+VRT_u_req_##nm(VRT_CTX)							\
+{									\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);				\
+	ctx->req->elem = val;						\
+}
+
 REQ_VAR_R(backend_hint, director_hint, VCL_BACKEND)
 
 REQ_VAR_L(ttl, d_ttl, VCL_DURATION, if (!(arg>0.0)) arg = 0;)
 REQ_VAR_R(ttl, d_ttl, VCL_DURATION)
+REQ_VAR_U(ttl, d_ttl, -1)
 REQ_VAR_L(grace, d_grace, VCL_DURATION, if (!(arg>0.0)) arg = 0;)
 REQ_VAR_R(grace, d_grace, VCL_DURATION)
+REQ_VAR_U(grace, d_grace, -1)
 
 VCL_VOID
 VRT_l_req_backend_hint(VRT_CTX, VCL_BACKEND be)
@@ -628,17 +650,16 @@ VRT_u_bereq_body(VRT_CTX)
 {
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	if (ctx->bo->bereq_body != NULL) {
-		(void)HSH_DerefObjCore(ctx->bo->wrk, &ctx->bo->bereq_body, 0);
-		http_Unset(ctx->bo->bereq, H_Content_Length);
-	}
+
+	http_Unset(ctx->bo->bereq, H_Content_Length);
+
+	if (ctx->bo->bereq_body != NULL)
+		(void)HSH_DerefObjCore(ctx->bo->wrk, &ctx->bo->bereq_body);
 
 	if (ctx->bo->req != NULL) {
 		CHECK_OBJ(ctx->bo->req, REQ_MAGIC);
-		ctx->bo->req = NULL;
-		ObjSetState(ctx->bo->wrk,
-		    ctx->bo->fetch_objcore, BOS_REQ_DONE);
-		http_Unset(ctx->bo->bereq, H_Content_Length);
+		VBO_SetState(ctx->bo->wrk, ctx->bo, BOS_REQ_DONE);
+		AZ(ctx->bo->req);
 	}
 }
 
@@ -780,9 +801,13 @@ VRT_r_##which##_##fld(VRT_CTX)					\
 
 /*lint -save -e835 */	// Zero right hand arg to '-'
 
+VRT_DO_EXP_R(obj_stale, ctx->bo->stale_oc, ttl,
+    ttl_now(ctx) - ctx->bo->stale_oc->t_origin)
 VRT_DO_EXP_R(obj, ctx->req->objcore, ttl,
     ttl_now(ctx) - ctx->req->objcore->t_origin)
+VRT_DO_EXP_R(obj_stale, ctx->bo->stale_oc, grace, 0)
 VRT_DO_EXP_R(obj, ctx->req->objcore, grace, 0)
+VRT_DO_EXP_R(obj_stale, ctx->bo->stale_oc, keep, 0)
 VRT_DO_EXP_R(obj, ctx->req->objcore, keep, 0)
 VRT_DO_EXP_L(beresp, ctx->bo->fetch_objcore, ttl,
     ttl_now(ctx) - ctx->bo->fetch_objcore->t_origin)
@@ -814,6 +839,7 @@ VRT_DO_TIME_R(resp, req, t_resp)
 VRT_DO_TIME_R(bereq, bo, t_first)
 VRT_DO_TIME_R(beresp, bo, t_resp)
 VRT_DO_TIME_R(obj, req->objcore, t_origin)
+VRT_DO_TIME_R(obj_stale, bo->stale_oc, t_origin)
 
 /*--------------------------------------------------------------------
  */
@@ -828,6 +854,7 @@ VRT_r_##which##_##age(VRT_CTX)					\
 	return (ttl_now(ctx) - oc->t_origin);			\
 }
 
+VRT_DO_AGE_R(obj_stale, ctx->bo->stale_oc)
 VRT_DO_AGE_R(obj, ctx->req->objcore)
 VRT_DO_AGE_R(beresp, ctx->bo->fetch_objcore)
 
@@ -989,14 +1016,26 @@ VRT_r_obj_hits(VRT_CTX)
 	return (ctx->req->is_hit ? ctx->req->objcore->hits : 0);
 }
 
-VCL_BOOL
-VRT_r_obj_uncacheable(VRT_CTX)
-{
+/*--------------------------------------------------------------------*/
 
+VCL_INT
+VRT_r_obj_stale_hits(VRT_CTX)
+{
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
-	return (ctx->req->objcore->flags & OC_F_HFM ? 1 : 0);
+	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->bo->stale_oc, OBJCORE_MAGIC);
+
+	return (ctx->bo->stale_oc->hits);
+}
+
+VCL_BOOL
+VRT_r_obj_stale_is_valid(VRT_CTX)
+{
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->bo->stale_oc, OBJCORE_MAGIC);
+
+	return (!(ctx->bo->stale_oc->flags & OC_F_DYING));
 }
 
 /*--------------------------------------------------------------------*/
@@ -1067,6 +1106,7 @@ VRT_l_##which##_body(VRT_CTX, enum lbody_e type,		\
 	if (type == LBODY_SET_BLOB || type == LBODY_ADD_BLOB) {	\
 		AZ(str);					\
 		b = body;					\
+		CHECK_OBJ_NOTNULL(b, VRT_BLOB_MAGIC);		\
 		VSB_bcat(vsb, b->blob, b->len);			\
 		return;						\
 	}							\
@@ -1075,6 +1115,7 @@ VRT_l_##which##_body(VRT_CTX, enum lbody_e type,		\
 	assert(type == LBODY_SET_STRING ||			\
 	    type == LBODY_ADD_STRING);				\
 	s = body;						\
+	CHECK_OBJ_NOTNULL(s, STRANDS_MAGIC);			\
 	for (n = 0; s != NULL && n < s->n; n++)			\
 		if (s->p[n] != NULL)				\
 			VSB_cat(vsb, s->p[n]);			\
@@ -1166,3 +1207,35 @@ SESS_VAR_DUR(timeout_idle, )
 SESS_VAR_DUR(timeout_linger, )
 SESS_VAR_DUR(send_timeout, )
 SESS_VAR_DUR(idle_send_timeout, set_idle_send_timeout(ctx->sp, d))
+
+/*--------------------------------------------------------------------*/
+
+#define PARAM_VAR(x, type)				\
+VCL_##type						\
+VRT_r_param_##x(VRT_CTX)				\
+{							\
+							\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);		\
+	return (cache_param->x);			\
+}							\
+
+PARAM_VAR(backend_idle_timeout, DURATION)
+PARAM_VAR(backend_wait_limit, INT)
+PARAM_VAR(backend_wait_timeout, DURATION)
+PARAM_VAR(between_bytes_timeout, DURATION)
+PARAM_VAR(connect_timeout, DURATION)
+PARAM_VAR(default_grace, DURATION)
+PARAM_VAR(default_keep, DURATION)
+PARAM_VAR(default_ttl, DURATION)
+PARAM_VAR(first_byte_timeout, DURATION)
+PARAM_VAR(idle_send_timeout, DURATION)
+PARAM_VAR(max_esi_depth, INT)
+PARAM_VAR(max_restarts, INT)
+PARAM_VAR(max_retries, INT)
+PARAM_VAR(pipe_task_deadline, DURATION)
+PARAM_VAR(pipe_timeout, DURATION)
+PARAM_VAR(send_timeout, DURATION)
+PARAM_VAR(shortlived, DURATION)
+PARAM_VAR(timeout_idle, DURATION)
+PARAM_VAR(transit_buffer, BYTES)
+PARAM_VAR(uncacheable_ttl, DURATION)
