@@ -43,6 +43,7 @@
 
 #include <sys/uio.h>
 #include "cache/cache_varnishd.h"
+#include "cache/cache_filter.h"
 
 #include <stdio.h>
 
@@ -55,18 +56,19 @@ struct v1l {
 	unsigned		magic;
 #define V1L_MAGIC		0x2f2142e5
 	int			*wfd;
-	enum sess_close		werr;	/* valid after V1L_Flush() */
+	stream_close_t		werr;	/* valid after V1L_Flush() */
 	struct iovec		*iov;
-	unsigned		siov;
-	unsigned		niov;
-	ssize_t			liov;
-	ssize_t			cliov;
-	unsigned		ciov;	/* Chunked header marker */
+	int			siov;
+	int			niov;
+	size_t			liov;
+	size_t			cliov;
+	int			ciov;	/* Chunked header marker */
 	vtim_real		deadline;
 	struct vsl_log		*vsl;
-	ssize_t			cnt;	/* Flushed byte count */
+	uint64_t		cnt;	/* Flushed byte count */
 	struct ws		*ws;
-	uintptr_t		res;
+	uintptr_t		ws_snap;
+	void			**vdp_priv;
 };
 
 /*--------------------------------------------------------------------
@@ -74,81 +76,97 @@ struct v1l {
  * otherwise, up to niov
  */
 
-void
-V1L_Open(struct worker *wrk, struct ws *ws, int *fd, struct vsl_log *vsl,
+struct v1l *
+V1L_Open(struct ws *ws, int *fd, struct vsl_log *vsl,
     vtim_real deadline, unsigned niov)
 {
 	struct v1l *v1l;
 	unsigned u;
-	uintptr_t res;
-
-	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	AZ(wrk->v1l);
+	uintptr_t ws_snap;
+	size_t sz;
 
 	if (WS_Overflowed(ws))
-		return;
+		return (NULL);
 
 	if (niov != 0)
 		assert(niov >= 3);
 
-	res = WS_Snapshot(ws);
+	ws_snap = WS_Snapshot(ws);
 
 	v1l = WS_Alloc(ws, sizeof *v1l);
 	if (v1l == NULL)
-		return;
+		return (NULL);
 	INIT_OBJ(v1l, V1L_MAGIC);
 
 	v1l->ws = ws;
-	v1l->res = res;
+	v1l->ws_snap = ws_snap;
 
 	u = WS_ReserveLumps(ws, sizeof(struct iovec));
 	if (u < 3) {
 		/* Must have at least 3 in case of chunked encoding */
 		WS_Release(ws, 0);
 		WS_MarkOverflow(ws);
-		return;
+		return (NULL);
 	}
 	if (u > IOV_MAX)
 		u = IOV_MAX;
 	if (niov != 0 && u > niov)
 		u = niov;
-	v1l->iov = (void*)ws->f;
-	v1l->siov = u;
-	v1l->ciov = u;
+	v1l->iov = WS_Reservation(ws);
+	v1l->siov = (int)u;
+	v1l->ciov = (int)u;
 	v1l->wfd = fd;
 	v1l->deadline = deadline;
 	v1l->vsl = vsl;
-	wrk->v1l = v1l;
+	v1l->werr = SC_NULL;
 
-	if (niov != 0)
-		WS_Release(ws, u * sizeof(struct iovec));
+	sz = u * sizeof(struct iovec);
+	assert(sz < UINT_MAX);
+	WS_Release(ws, (unsigned)sz);
+	return (v1l);
 }
 
-enum sess_close
-V1L_Close(struct worker *wrk, uint64_t *cnt)
+void
+V1L_NoRollback(struct v1l *v1l)
+{
+
+	CHECK_OBJ_NOTNULL(v1l, V1L_MAGIC);
+	v1l->ws_snap = 0;
+}
+
+stream_close_t
+V1L_Close(struct v1l **v1lp, uint64_t *cnt)
 {
 	struct v1l *v1l;
-	enum sess_close sc;
+	struct ws *ws;
+	uintptr_t ws_snap;
+	stream_close_t sc;
 
-	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	AN(cnt);
-	sc = V1L_Flush(wrk);
-	v1l = wrk->v1l;
-	wrk->v1l = NULL;
-	CHECK_OBJ_NOTNULL(v1l, V1L_MAGIC);
+	TAKE_OBJ_NOTNULL(v1l, v1lp, V1L_MAGIC);
+	if (v1l->vdp_priv != NULL) {
+		assert(*v1l->vdp_priv == v1l);
+		*v1l->vdp_priv = NULL;
+	}
+	sc = V1L_Flush(v1l);
 	*cnt = v1l->cnt;
-	if (v1l->ws->r)
-		WS_Release(v1l->ws, 0);
-	WS_Rollback(v1l->ws, v1l->res);
+	ws = v1l->ws;
+	ws_snap = v1l->ws_snap;
 	ZERO_OBJ(v1l, sizeof *v1l);
+	if (ws_snap != 0)
+		WS_Rollback(ws, ws_snap);
 	return (sc);
 }
 
 static void
-v1l_prune(struct v1l *v1l, size_t bytes)
+v1l_prune(struct v1l *v1l, ssize_t abytes)
 {
-	ssize_t used = 0;
-	ssize_t j, used_here;
+	size_t used = 0;
+	size_t sz, bytes, used_here;
+	int j;
+
+	assert(abytes > 0);
+	bytes = (size_t)abytes;
 
 	for (j = 0; j < v1l->niov; j++) {
 		if (used + v1l->iov[j].iov_len > bytes) {
@@ -157,9 +175,11 @@ v1l_prune(struct v1l *v1l, size_t bytes)
 			v1l->iov[j].iov_len -= used_here;
 			v1l->iov[j].iov_base =
 			    (char*)v1l->iov[j].iov_base + used_here;
-			memmove(v1l->iov, &v1l->iov[j],
-			    (v1l->niov - j) * sizeof(struct iovec));
+			sz = (unsigned)v1l->niov - (unsigned)j;
+			sz *= sizeof(struct iovec);
+			memmove(v1l->iov, &v1l->iov[j], sz);
 			v1l->niov -= j;
+			assert(v1l->liov >= bytes);
 			v1l->liov -= bytes;
 			return;
 		}
@@ -168,16 +188,16 @@ v1l_prune(struct v1l *v1l, size_t bytes)
 	AZ(v1l->liov);
 }
 
-enum sess_close
-V1L_Flush(const struct worker *wrk)
+stream_close_t
+V1L_Flush(struct v1l *v1l)
 {
 	ssize_t i;
-	struct v1l *v1l;
+	size_t sz;
+	int err;
 	char cbuf[32];
 
-	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	v1l = wrk->v1l;
 	CHECK_OBJ_NOTNULL(v1l, V1L_MAGIC);
+	CHECK_OBJ_NOTNULL(v1l->werr, STREAM_CLOSE_MAGIC);
 	AN(v1l->wfd);
 
 	assert(v1l->niov <= v1l->siov);
@@ -186,13 +206,13 @@ V1L_Flush(const struct worker *wrk)
 		if (v1l->ciov < v1l->siov && v1l->cliov > 0) {
 			/* Add chunk head & tail */
 			bprintf(cbuf, "00%zx\r\n", v1l->cliov);
-			i = strlen(cbuf);
+			sz = strlen(cbuf);
 			v1l->iov[v1l->ciov].iov_base = cbuf;
-			v1l->iov[v1l->ciov].iov_len = i;
-			v1l->liov += i;
+			v1l->iov[v1l->ciov].iov_len = sz;
+			v1l->liov += sz;
 
 			/* This is OK, because siov was --'ed */
-			v1l->iov[v1l->niov].iov_base = cbuf + i - 2;
+			v1l->iov[v1l->niov].iov_base = cbuf + sz - 2;
 			v1l->iov[v1l->niov++].iov_len = 2;
 			v1l->liov += 2;
 		} else if (v1l->ciov < v1l->siov) {
@@ -201,6 +221,7 @@ V1L_Flush(const struct worker *wrk)
 		}
 
 		i = 0;
+		err = 0;
 		do {
 			if (VTIM_real() > v1l->deadline) {
 				VSLb(v1l->vsl, SLT_Debug,
@@ -212,11 +233,11 @@ V1L_Flush(const struct worker *wrk)
 			}
 
 			i = writev(*v1l->wfd, v1l->iov, v1l->niov);
-			if (i > 0)
-				v1l->cnt += i;
-
-			if (i == v1l->liov)
-				break;
+			if (i > 0) {
+				v1l->cnt += (size_t)i;
+				if ((size_t)i == v1l->liov)
+					break;
+			}
 
 			/* we hit a timeout, and some data may have been sent:
 			 * Remove sent data from start of I/O vector, then retry
@@ -225,23 +246,29 @@ V1L_Flush(const struct worker *wrk)
 			 * prevent slowloris attacks
 			 */
 
-			VSLb(v1l->vsl, SLT_Debug,
-			    "Hit idle send timeout, wrote = %zd/%zd; retrying",
-			    i, v1l->liov);
+			err = errno;
+
+			if (err == EWOULDBLOCK) {
+				VSLb(v1l->vsl, SLT_Debug,
+				    "Hit idle send timeout, "
+				    "wrote = %zd/%zd; retrying",
+				    i, v1l->liov);
+			}
 
 			if (i > 0)
 				v1l_prune(v1l, i);
-		} while (i > 0 || errno == EWOULDBLOCK);
+		} while (i > 0 || err == EWOULDBLOCK);
 
 		if (i <= 0) {
 			VSLb(v1l->vsl, SLT_Debug,
 			    "Write error, retval = %zd, len = %zd, errno = %s",
-			    i, v1l->liov, vstrerror(errno));
-			AZ(v1l->werr);
-			if (errno == EPIPE)
+			    i, v1l->liov, VAS_errtxt(err));
+			assert(v1l->werr == SC_NULL);
+			if (err == EPIPE)
 				v1l->werr = SC_REM_CLOSE;
 			else
 				v1l->werr = SC_TX_ERROR;
+			errno = err;
 		}
 	}
 	v1l->liov = 0;
@@ -249,40 +276,43 @@ V1L_Flush(const struct worker *wrk)
 	v1l->niov = 0;
 	if (v1l->ciov < v1l->siov)
 		v1l->ciov = v1l->niov++;
+	CHECK_OBJ_NOTNULL(v1l->werr, STREAM_CLOSE_MAGIC);
 	return (v1l->werr);
 }
 
 size_t
-V1L_Write(const struct worker *wrk, const void *ptr, ssize_t len)
+V1L_Write(struct v1l *v1l, const void *ptr, ssize_t alen)
 {
-	struct v1l *v1l;
+	size_t len = 0;
 
-	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	v1l = wrk->v1l;
 	CHECK_OBJ_NOTNULL(v1l, V1L_MAGIC);
 	AN(v1l->wfd);
-	if (len == 0 || *v1l->wfd < 0)
+	if (alen == 0 || *v1l->wfd < 0)
 		return (0);
-	if (len == -1)
+	if (alen > 0)
+		len = (size_t)alen;
+	else if (alen == -1)
 		len = strlen(ptr);
+	else
+		WRONG("alen");
+
 	assert(v1l->niov < v1l->siov);
 	v1l->iov[v1l->niov].iov_base = TRUST_ME(ptr);
 	v1l->iov[v1l->niov].iov_len = len;
 	v1l->liov += len;
 	v1l->niov++;
 	v1l->cliov += len;
-	if (v1l->niov >= v1l->siov)
-		(void)V1L_Flush(wrk);
+	if (v1l->niov >= v1l->siov) {
+		(void)V1L_Flush(v1l);
+		VSC_C_main->http1_iovs_flush++;
+	}
 	return (len);
 }
 
 void
-V1L_Chunked(const struct worker *wrk)
+V1L_Chunked(struct v1l *v1l)
 {
-	struct v1l *v1l;
 
-	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	v1l = wrk->v1l;
 	CHECK_OBJ_NOTNULL(v1l, V1L_MAGIC);
 
 	assert(v1l->ciov == v1l->siov);
@@ -291,8 +321,10 @@ V1L_Chunked(const struct worker *wrk)
 	 * If there is no space for chunked header, a chunk of data and
 	 * a chunk tail, we might as well flush right away.
 	 */
-	if (v1l->niov + 3 >= v1l->siov)
-		(void)V1L_Flush(wrk);
+	if (v1l->niov + 3 >= v1l->siov) {
+		(void)V1L_Flush(v1l);
+		VSC_C_main->http1_iovs_flush++;
+	}
 	v1l->siov--;
 	v1l->ciov = v1l->niov++;
 	v1l->cliov = 0;
@@ -308,19 +340,129 @@ V1L_Chunked(const struct worker *wrk)
  */
 
 void
-V1L_EndChunk(const struct worker *wrk)
+V1L_EndChunk(struct v1l *v1l)
 {
-	struct v1l *v1l;
 
-	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	v1l = wrk->v1l;
 	CHECK_OBJ_NOTNULL(v1l, V1L_MAGIC);
 
 	assert(v1l->ciov < v1l->siov);
-	(void)V1L_Flush(wrk);
+	(void)V1L_Flush(v1l);
 	v1l->siov++;
 	v1l->ciov = v1l->siov;
 	v1l->niov = 0;
 	v1l->cliov = 0;
-	(void)V1L_Write(wrk, "0\r\n\r\n", -1);
+	(void)V1L_Write(v1l, "0\r\n\r\n", -1);
 }
+
+/*--------------------------------------------------------------------
+ * VDP using V1L
+ */
+
+/* remember priv pointer for V1L_Close() to clear */
+static int v_matchproto_(vdp_init_f)
+v1l_init(VRT_CTX, struct vdp_ctx *vdc, void **priv)
+{
+	struct v1l *v1l;
+
+	(void) ctx;
+	(void) vdc;
+	AN(priv);
+	CAST_OBJ_NOTNULL(v1l, *priv, V1L_MAGIC);
+
+	v1l->vdp_priv = priv;
+	return (0);
+}
+
+static int v_matchproto_(vdp_bytes_f)
+v1l_bytes(struct vdp_ctx *vdc, enum vdp_action act, void **priv,
+    const void *ptr, ssize_t len)
+{
+	size_t wl = 0;
+
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	AN(priv);
+
+	AZ(vdc->nxt);		/* always at the bottom of the pile */
+
+	if (len > 0)
+		wl = V1L_Write(*priv, ptr, len);
+	if (act > VDP_NULL && V1L_Flush(*priv) != SC_NULL)
+		return (-1);
+	if ((size_t)len != wl)
+		return (-1);
+	return (0);
+}
+
+/*--------------------------------------------------------------------
+ * VDPIO using V1L
+ *
+ * this is deliverately half-baked to reduce work in progress while heading
+ * towards VAI/VDPIO: we update the v1l with the scarab, which we
+ * return unmodified.
+ *
+ */
+
+/* remember priv pointer for V1L_Close() to clear */
+static int v_matchproto_(vpio_init_f)
+v1l_io_init(VRT_CTX, struct vdp_ctx *vdc, void **priv, int capacity)
+{
+	struct v1l *v1l;
+
+	(void) ctx;
+	(void) vdc;
+	AN(priv);
+
+	CAST_OBJ_NOTNULL(v1l, *priv, V1L_MAGIC);
+
+	v1l->vdp_priv = priv;
+	return (capacity);
+}
+
+static int v_matchproto_(vpio_init_f)
+v1l_io_upgrade(VRT_CTX, struct vdp_ctx *vdc, void **priv, int capacity)
+{
+	return (v1l_io_init(ctx, vdc, priv, capacity));
+}
+
+/*
+ * API note
+ *
+ * this VDP is special in that it does not transform data, but prepares
+ * the write. From the perspective of VDPIO, its current state is only
+ * transitional.
+ *
+ * Because the VDP prepares the actual writes, but the caller needs
+ * to return the scarab's leases, the caller in this case is
+ * required to empty the scarab after V1L_Flush()'ing.
+ */
+
+static int v_matchproto_(vdpio_lease_f)
+v1l_io_lease(struct vdp_ctx *vdc, struct vdp_entry *this, struct vscarab *scarab)
+{
+	struct v1l *v1l;
+	struct viov *v;
+	int r;
+
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(this, VDP_ENTRY_MAGIC);
+	CAST_OBJ_NOTNULL(v1l, this->priv, V1L_MAGIC);
+	VSCARAB_CHECK(scarab);
+	AZ(scarab->used);	// see note above
+	this->calls++;
+	r = vdpio_pull(vdc, this, scarab);
+	if (r < 0)
+		return (r);
+	VSCARAB_FOREACH(v, scarab)
+		this->bytes_in += V1L_Write(v1l, v->iov.iov_base, v->iov.iov_len);
+	return (r);
+}
+
+const struct vdp * const VDP_v1l = &(struct vdp){
+	.name =		"V1B",
+	.init =		v1l_init,
+	.bytes =	v1l_bytes,
+
+	.io_init =	v1l_io_init,
+	.io_upgrade =	v1l_io_upgrade,
+	.io_lease =	v1l_io_lease,
+};
