@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -32,36 +34,37 @@
 #include "config.h"
 
 #include <sys/types.h>
-#include <sys/wait.h>
 
-#include <errno.h>
-#include <fcntl.h>
 #include <poll.h>
-#include <signal.h>
+#include <stdarg.h>
 #include <stdio.h>
-#include <stdlib.h>
 #include <string.h>
 #include <syslog.h>
 #include <unistd.h>
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <sys/time.h>
+#include <sys/resource.h>
 
-#include "mgt/mgt.h"
-#include "common/heritage.h"
+#include "mgt.h"
+#include "acceptor/mgt_acceptor.h"
+
+#include "vapi/vsig.h"
 
 #include "vbm.h"
-#include "vcli.h"
-#include "vcli_priv.h"
+#include "vcli_serve.h"
 #include "vev.h"
+#include "vfil.h"
 #include "vlu.h"
 #include "vtim.h"
 
-#include "mgt_cli.h"
+#include "common/heritage.h"
 
-pid_t			child_pid = -1;
+static pid_t		child_pid = -1;
 
 static struct vbitmap	*fd_map;
 
-static int		child_cli_in = -1;
-static int		child_cli_out = -1;
+static int		child_cli_fd = -1;
 static int		child_output = -1;
 
 static enum {
@@ -85,27 +88,9 @@ static struct vev	*ev_listen;
 static struct vlu	*child_std_vlu;
 
 static struct vsb *child_panic = NULL;
-static double mgt_uptime_t0 = 0.;
-
-#ifdef HAVE_SIGALTSTACK
-#include <sys/mman.h>
-stack_t altstack;
-#endif
 
 static void mgt_reap_child(void);
-
-/*---------------------------------------------------------------------
- * A handy little function
- */
-
-static inline void
-closex(int *fd)
-{
-
-	assert(*fd >= 0);
-	AZ(close(*fd));
-	*fd = -1;
-}
+static int kill_child(void);
 
 /*=====================================================================
  * Panic string evacuation and handling
@@ -117,32 +102,27 @@ mgt_panic_record(pid_t r)
 	char time_str[30];
 
 	if (child_panic != NULL)
-		VSB_delete(child_panic);
+		VSB_destroy(&child_panic);
 	child_panic = VSB_new_auto();
 	AN(child_panic);
 	VTIM_format(VTIM_real(), time_str);
-	VSB_printf(child_panic, "Last panic at: %s\n", time_str);
+	VSB_printf(child_panic, "Child (%jd) Panic at: %s\n",
+	    (intmax_t)r, time_str);
 	VSB_quote(child_panic, heritage.panic_str,
 	    strnlen(heritage.panic_str, heritage.panic_str_len),
 	    VSB_QUOTE_NONL);
-	AZ(VSB_finish(child_panic));
-	MGT_complain(C_ERR, "Child (%jd) %s",
-	    (intmax_t)r, VSB_data(child_panic));
+	MGT_ComplainVSB(C_ERR, child_panic);
 }
 
 static void
 mgt_panic_clear(void)
 {
-	VSB_delete(child_panic);
-	child_panic = NULL;
+	VSB_destroy(&child_panic);
 }
 
-void __match_proto__(cli_func_t)
-mcf_panic_show(struct cli *cli, const char * const *av, void *priv)
+static void
+cli_panic_show(struct cli *cli, const char * const *av, int json)
 {
-	(void)av;
-	(void)priv;
-
 	if (!child_panic) {
 		VCLI_SetResult(cli, CLIS_CANT);
 		VCLI_Out(cli,
@@ -150,11 +130,33 @@ mcf_panic_show(struct cli *cli, const char * const *av, void *priv)
 		return;
 	}
 
-	VCLI_Out(cli, "%s\n", VSB_data(child_panic));
+	if (!json) {
+		VCLI_Out(cli, "%s\n", VSB_data(child_panic));
+		return;
+	}
+
+	VCLI_JSON_begin(cli, 2, av);
+	VCLI_Out(cli, ",\n");
+	VCLI_JSON_str(cli, VSB_data(child_panic));
+	VCLI_JSON_end(cli);
 }
 
-void __match_proto__(cli_func_t)
-mcf_panic_clear(struct cli *cli, const char * const *av, void *priv)
+static void v_matchproto_(cli_func_t)
+mch_cli_panic_show(struct cli *cli, const char * const *av, void *priv)
+{
+	(void)priv;
+	cli_panic_show(cli, av, 0);
+}
+
+static void v_matchproto_(cli_func_t)
+mch_cli_panic_show_json(struct cli *cli, const char * const *av, void *priv)
+{
+	(void)priv;
+	cli_panic_show(cli, av, 1);
+}
+
+static void v_matchproto_(cli_func_t)
+mch_cli_panic_clear(struct cli *cli, const char * const *av, void *priv)
 {
 	(void)priv;
 
@@ -163,7 +165,7 @@ mcf_panic_clear(struct cli *cli, const char * const *av, void *priv)
 		VCLI_Out(cli, "Unknown parameter \"%s\".", av[2]);
 		return;
 	} else if (av[2] != NULL) {
-		VSC_C_mgt->child_panic = static_VSC_C_mgt.child_panic = 0;
+		VSC_C_mgt->child_panic = 0;
 		if (child_panic == NULL)
 			return;
 	}
@@ -183,23 +185,26 @@ mcf_panic_clear(struct cli *cli, const char * const *av, void *priv)
  *
  * This is likely to a bit on the low side, as libc and other libraries
  * has a tendency to cache file descriptors (syslog, resolver, etc.)
- * so we add a margin of 100 fds.
+ * so we add a margin of 10 fds.
+ *
+ * For added safety, we check that we see no file descriptor open for
+ * another margin above the limit for which we close by design
  */
 
 static int		mgt_max_fd;
 
-#define CLOSE_FD_UP_TO	(mgt_max_fd + 100)
+#define CLOSE_FD_UP_TO	(mgt_max_fd + 10)
+#define CHECK_FD_UP_TO	(CLOSE_FD_UP_TO + 10)
 
 void
-mgt_got_fd(int fd)
+MCH_TrackHighFd(int fd)
 {
 	/*
 	 * Assert > 0, to catch bogus opens, we know where stdin goes
 	 * in the master process.
 	 */
 	assert(fd > 0);
-	if (fd > mgt_max_fd)
-		mgt_max_fd = fd;
+	mgt_max_fd = vmax(mgt_max_fd, fd);
 }
 
 /*--------------------------------------------------------------------
@@ -208,12 +213,15 @@ mgt_got_fd(int fd)
  */
 
 void
-mgt_child_inherit(int fd, const char *what)
+MCH_Fd_Inherit(int fd, const char *what)
 {
 
 	assert(fd >= 0);
+	// XXX why?
+	if (fd > 0)
+		MCH_TrackHighFd(fd);
 	if (fd_map == NULL)
-		fd_map = vbit_init(128);
+		fd_map = vbit_new(128);
 	AN(fd_map);
 	if (what != NULL)
 		vbit_set(fd_map, fd);
@@ -225,12 +233,14 @@ mgt_child_inherit(int fd, const char *what)
  * Listen to stdout+stderr from the child
  */
 
-static int
+static const char *whining_child = C_ERR;
+
+static int v_matchproto_(vlu_f)
 child_line(void *priv, const char *p)
 {
 	(void)priv;
 
-	MGT_complain(C_INFO, "Child (%jd) said %s", (intmax_t)child_pid, p);
+	MGT_Complain(whining_child, "Child (%jd) said %s", (intmax_t)child_pid, p);
 	return (0);
 }
 
@@ -238,11 +248,11 @@ child_line(void *priv, const char *p)
  * NB: Notice cleanup call from mgt_reap_child()
  */
 
-static int __match_proto__(vev_cb_f)
+static int v_matchproto_(vev_cb_f)
 child_listener(const struct vev *e, int what)
 {
 
-	if ((what & ~EV_RD) || VLU_Fd(child_output, child_std_vlu)) {
+	if ((what & ~VEV__RD) || VLU_Fd(child_std_vlu, child_output)) {
 		ev_listen = NULL;
 		if (e != NULL)
 			mgt_reap_child();
@@ -255,7 +265,7 @@ child_listener(const struct vev *e, int what)
  * Periodically poke the child, to see that it still lives
  */
 
-static int __match_proto__(vev_cb_f)
+static int v_matchproto_(vev_cb_f)
 child_poker(const struct vev *e, int what)
 {
 	char *r = NULL;
@@ -268,40 +278,27 @@ child_poker(const struct vev *e, int what)
 	if (child_pid < 0)
 		return (0);
 	if (mgt_cli_askchild(&status, &r, "ping\n") || strncmp("PONG ", r, 5)) {
-		MGT_complain(C_ERR, "Unexpected reply from ping: %u %s",
+		MGT_Complain(C_ERR, "Unexpected reply from ping: %u %s",
 		    status, r);
 		if (status != CLIS_COMMS)
-			MGT_Child_Cli_Fail();
+			MCH_Cli_Fail();
 	}
 	free(r);
-	return 0;
-}
-
-/*=====================================================================
- * signal handler for child process
- */
-
-static void __match_proto__()
-child_signal_handler(int s, siginfo_t *si, void *c)
-{
-	char buf[1024];
-
-	(void)c;
-
-	bprintf(buf, "Signal %d (%s) received at %p si_code %d",
-		s, strsignal(s), si->si_addr, si->si_code);
-
-	VAS_Fail(__func__,
-		 __FILE__,
-		 __LINE__,
-		 buf,
-		 VAS_ASSERT);
+	return (0);
 }
 
 /*=====================================================================
  * Launch the child process
  */
 
+#define mgt_launch_err(cli, status, ...) do {		\
+		MGT_Complain(C_ERR, __VA_ARGS__);	\
+		if (cli == NULL)			\
+			break;				\
+		VCLI_Out(cli, __VA_ARGS__);		\
+		VCLI_SetResult(cli, status);		\
+	} while (0)
+
 static void
 mgt_launch_child(struct cli *cli)
 {
@@ -309,36 +306,24 @@ mgt_launch_child(struct cli *cli)
 	unsigned u;
 	char *p;
 	struct vev *e;
-	int i, j, k, cp[2];
-	struct sigaction sa;
+	int i, cp[2];
+	struct rlimit rl[1];
+	vtim_dur dstart;
+	int bstart;
+	vtim_mono t0;
 
 	if (child_state != CH_STOPPED && child_state != CH_DIED)
 		return;
 
-	if (!MAC_sockets_ready(cli)) {
-		child_state = CH_STOPPED;
-		if (cli != NULL) {
-			VCLI_SetResult(cli, CLIS_CANT);
-			return;
-		}
-		MGT_complain(C_ERR,
-		    "Child start failed: could not open sockets");
-		return;
-	}
-
 	child_state = CH_STARTING;
 
-	/* Open pipe for mgr->child CLI */
-	AZ(pipe(cp));
-	heritage.cli_in = cp[0];
-	mgt_child_inherit(heritage.cli_in, "cli_in");
-	child_cli_out = cp[1];
-
-	/* Open pipe for child->mgr CLI */
-	AZ(pipe(cp));
-	heritage.cli_out = cp[1];
-	mgt_child_inherit(heritage.cli_out, "cli_out");
-	child_cli_in = cp[0];
+	/* Open pipe for mgt->child CLI */
+	AZ(socketpair(AF_UNIX, SOCK_STREAM, 0, cp));
+	heritage.cli_fd = cp[0];
+	assert(cp[0] > STDERR_FILENO);	// See #2782
+	assert(cp[1] > STDERR_FILENO);
+	MCH_Fd_Inherit(heritage.cli_fd, "cli_fd");
+	child_cli_fd = cp[1];
 
 	/*
 	 * Open pipe for child stdout/err
@@ -348,23 +333,33 @@ mgt_launch_child(struct cli *cli)
 	heritage.std_fd = cp[1];
 	child_output = cp[0];
 
-	AN(heritage.vsm);
-	mgt_SHM_Size_Adjust();
-	AN(heritage.vsm);
+	mgt_SHM_ChildNew();
+
 	AN(heritage.param);
+	AN(heritage.panic_str);
+	VJ_master(JAIL_MASTER_SYSTEM);
 	if ((pid = fork()) < 0) {
-		/* XXX */
+		VJ_master(JAIL_MASTER_LOW);
 		perror("Could not fork child");
-		exit(1);
+		exit(1);		// XXX Harsh ?
 	}
 	if (pid == 0) {
 
+		if (MGT_FEATURE(FEATURE_NO_COREDUMP)) {
+			memset(rl, 0, sizeof *rl);
+			rl->rlim_cur = 0;
+			AZ(setrlimit(RLIMIT_CORE, rl));
+		}
+
 		/* Redirect stdin/out/err */
-		AZ(close(STDIN_FILENO));
-		assert(open("/dev/null", O_RDONLY) == STDIN_FILENO);
+		VFIL_null_fd(STDIN_FILENO);
 		assert(dup2(heritage.std_fd, STDOUT_FILENO) == STDOUT_FILENO);
 		assert(dup2(heritage.std_fd, STDERR_FILENO) == STDERR_FILENO);
 
+		setbuf(stdout, NULL);
+		setbuf(stderr, NULL);
+		printf("Child starts\n");
+
 		/*
 		 * Close all FDs the child shouldn't know about
 		 *
@@ -378,101 +373,124 @@ mgt_launch_child(struct cli *cli)
 		 */
 		closelog();
 
-		for (i = STDERR_FILENO + 1; i < CLOSE_FD_UP_TO; i++) {
+		for (i = STDERR_FILENO + 1; i <= CLOSE_FD_UP_TO; i++) {
 			if (vbit_test(fd_map, i))
 				continue;
-			if (close(i) == 0) {
-				k = open("/dev/null", O_RDONLY);
-				assert(k >= 0);
-				j = dup2(k, i);
-				assert(j == i);
-				AZ(close(k));
-			}
+			if (close(i) == 0)
+				VFIL_null_fd(i);
 		}
-#ifdef HAVE_SETPROCTITLE
-		setproctitle("Varnish-Chld %s", heritage.name);
-#endif
-
-		if (mgt_param.sigsegv_handler) {
-			memset(&sa, 0, sizeof sa);
-			sa.sa_sigaction = child_signal_handler;
-			sa.sa_flags = SA_SIGINFO;
-			(void)sigaction(SIGBUS, &sa, NULL);
-			(void)sigaction(SIGABRT, &sa, NULL);
-
-#ifdef HAVE_SIGALTSTACK
-			size_t sz = SIGSTKSZ + 4096;
-			if (sz < mgt_param.wthread_stacksize)
-				sz = mgt_param.wthread_stacksize;
-			altstack.ss_sp = mmap(NULL, sz,  PROT_READ | PROT_WRITE,
-					      MAP_PRIVATE | MAP_ANONYMOUS,
-					      -1, 0);
-			AN(altstack.ss_sp != MAP_FAILED);
-			AN(altstack.ss_sp);
-			altstack.ss_size = sz;
-			altstack.ss_flags = 0;
-			AZ(sigaltstack(&altstack, NULL));
-			sa.sa_flags |= SA_ONSTACK;
-#endif
-			(void)sigaction(SIGSEGV, &sa, NULL);
-
+		for (i = CLOSE_FD_UP_TO + 1; i <= CHECK_FD_UP_TO; i++) {
+			assert(close(i) == -1);
+			assert(errno == EBADF);
 		}
-		(void)signal(SIGINT, SIG_DFL);
-		(void)signal(SIGTERM, SIG_DFL);
+
+		mgt_ProcTitle("Child");
+
+		heritage.cls = mgt_cls;
+		heritage.ident = VSB_data(vident) + 1;
+
+		vext_load();
+
+		STV_Init();
 
 		VJ_subproc(JAIL_SUBPROC_WORKER);
 
-		child_main();
+		/*
+		 * We pass these two params because child_main needs them
+		 * well before it has found its own param struct.
+		 */
+		child_main(mgt_param.sigsegv_handler,
+		    mgt_param.wthread_stacksize);
+
+		/*
+		 * It would be natural to clean VSMW up here, but it is apt
+		 * to fail in some scenarios because of the fall-back
+		 * "rm -rf" in mgt_SHM_ChildDestroy() which is there to
+		 * catch the cases were we don't get here.
+		 */
+		// VSMW_Destroy(&heritage.proc_vsmw);
 
 		exit(0);
 	}
+	VJ_master(JAIL_MASTER_LOW);
 	assert(pid > 1);
-	MGT_complain(C_DEBUG, "Child (%jd) Started", (intmax_t)pid);
-	VSC_C_mgt->child_start = ++static_VSC_C_mgt.child_start;
+	MGT_Complain(C_DEBUG, "Child (%jd) Started", (intmax_t)pid);
+	VSC_C_mgt->child_start++;
 
 	/* Close stuff the child got */
-	closex(&heritage.std_fd);
+	closefd(&heritage.std_fd);
 
-	mgt_child_inherit(heritage.cli_in, NULL);
-	closex(&heritage.cli_in);
+	MCH_Fd_Inherit(heritage.cli_fd, NULL);
+	closefd(&heritage.cli_fd);
 
-	mgt_child_inherit(heritage.cli_out, NULL);
-	closex(&heritage.cli_out);
-
-	child_std_vlu = VLU_New(NULL, child_line, 0);
+	child_std_vlu = VLU_New(child_line, NULL, 0);
 	AN(child_std_vlu);
 
+	/* Wait for cache/cache_cli.c::CLI_Run() to check in */
+	bstart = mgt_param.startup_timeout >= mgt_param.cli_timeout;
+	dstart = bstart ? mgt_param.startup_timeout : mgt_param.cli_timeout;
+	t0 = VTIM_mono();
+	u = mgt_cli_start_child(child_cli_fd, dstart);
+	if (u != CLIS_OK) {
+		assert(u == CLIS_COMMS);
+		if (VTIM_mono() - t0 < dstart)
+			mgt_launch_err(cli, u, "Child failed on launch ");
+		else
+			mgt_launch_err(cli, u, "Child failed on launch "
+			    "within %s_timeout=%.2fs%s",
+			    bstart ? "startup" : "cli", dstart,
+			    bstart ? "" : " (tip: set startup_timeout)");
+		child_pid = pid;
+		(void)kill_child();
+		mgt_reap_child();
+		child_state = CH_STOPPED;
+		return;
+	} else {
+		assert(u == CLIS_OK);
+		fprintf(stderr, "Child launched OK\n");
+	}
+	whining_child = C_INFO;
+
 	AZ(ev_listen);
-	e = vev_new();
+	e = VEV_Alloc();
 	XXXAN(e);
 	e->fd = child_output;
-	e->fd_flags = EV_RD;
+	e->fd_flags = VEV__RD;
 	e->name = "Child listener";
 	e->callback = child_listener;
-	AZ(vev_add(mgt_evb, e));
+	AZ(VEV_Start(mgt_evb, e));
 	ev_listen = e;
 	AZ(ev_poker);
 	if (mgt_param.ping_interval > 0) {
-		e = vev_new();
+		e = VEV_Alloc();
 		XXXAN(e);
 		e->timeout = mgt_param.ping_interval;
 		e->callback = child_poker;
 		e->name = "child poker";
-		AZ(vev_add(mgt_evb, e));
+		AZ(VEV_Start(mgt_evb, e));
 		ev_poker = e;
 	}
 
-	mgt_cli_start_child(child_cli_in, child_cli_out);
 	child_pid = pid;
-	if (mgt_push_vcls_and_start(cli, &u, &p)) {
-		VCLI_SetResult(cli, u);
-		MGT_complain(C_ERR, "Child (%jd) Pushing vcls failed:\n%s",
+
+	if (mgt_push_vcls(cli, &u, &p)) {
+		mgt_launch_err(cli, u, "Child (%jd) Pushing vcls failed:\n%s",
 		    (intmax_t)child_pid, p);
 		free(p);
-		child_state = CH_RUNNING;
-		mgt_stop_child();
-	} else
-		child_state = CH_RUNNING;
+		MCH_Stop_Child();
+		return;
+	}
+
+	if (mgt_cli_askchild(&u, &p, "start\n")) {
+		mgt_launch_err(cli, u, "Child (%jd) Acceptor start failed:\n%s",
+		    (intmax_t)child_pid, p);
+		free(p);
+		MCH_Stop_Child();
+		return;
+	}
+
+	free(p);
+	child_state = CH_RUNNING;
 }
 
 /*=====================================================================
@@ -480,14 +498,12 @@ mgt_launch_child(struct cli *cli)
  */
 
 static int
-kill_child(void) {
+kill_child(void)
+{
 	int i, error;
 
 	VJ_master(JAIL_MASTER_KILL);
-	if (MGT_FEATURE(FEATURE_NO_COREDUMP))
-		i = kill(child_pid, SIGKILL);
-	else
-		i = kill(child_pid, SIGQUIT);
+	i = kill(child_pid, SIGQUIT);
 	error = errno;
 	VJ_master(JAIL_MASTER_LOW);
 	errno = error;
@@ -509,21 +525,19 @@ mgt_reap_child(void)
 	 * This signals orderly shut down to child
 	 */
 	mgt_cli_stop_child();
-	if (child_cli_out >= 0)
-		closex(&child_cli_out);
-	if (child_cli_in >= 0)
-		closex(&child_cli_in);
+	if (child_cli_fd >= 0)
+		closefd(&child_cli_fd);
 
 	/* Stop the poker */
 	if (ev_poker != NULL) {
-		vev_del(mgt_evb, ev_poker);
+		VEV_Stop(mgt_evb, ev_poker);
 		free(ev_poker);
+		ev_poker = NULL;
 	}
-	ev_poker = NULL;
 
 	/* Stop the listener */
 	if (ev_listen != NULL) {
-		vev_del(mgt_evb, ev_listen);
+		VEV_Stop(mgt_evb, ev_listen);
 		free(ev_listen);
 		ev_listen = NULL;
 	}
@@ -532,74 +546,87 @@ mgt_reap_child(void)
 	vsb = VSB_new_auto();
 	XXXAN(vsb);
 
+	(void)VFIL_nonblocking(child_output);
 	/* Wait for child to die */
-	for (i = 0; i < mgt_param.cli_timeout; i++) {
+	for (i = 0; i < mgt_param.cli_timeout * 10; i++) {
+		(void)child_listener(NULL, VEV__RD);
 		r = waitpid(child_pid, &status, WNOHANG);
 		if (r == child_pid)
 			break;
-		(void)sleep(1);
+		VTIM_sleep(0.1);
 	}
 	if (r == 0) {
-		VSB_printf(vsb, "Child (%jd) not dying, killing", (intmax_t)r);
+		VSB_printf(vsb, "Child (%jd) not dying (waitpid = %jd),"
+		    " killing\n", (intmax_t)child_pid, (intmax_t)r);
 
 		/* Kick it Jim... */
 		(void)kill_child();
 		r = waitpid(child_pid, &status, 0);
 	}
 	if (r != child_pid)
-		fprintf(stderr, "WAIT 0x%jx\n", (uintmax_t)r);
+		fprintf(stderr, "WAIT 0x%jd\n", (intmax_t)r);
 	assert(r == child_pid);
 
-	MAC_reopen_sockets(NULL);
-
 	VSB_printf(vsb, "Child (%jd) %s", (intmax_t)r,
 	    status ? "died" : "ended");
 	if (WIFEXITED(status) && WEXITSTATUS(status)) {
 		VSB_printf(vsb, " status=%d", WEXITSTATUS(status));
 		exit_status |= 0x20;
 		if (WEXITSTATUS(status) == 1)
-			VSC_C_mgt->child_exit = ++static_VSC_C_mgt.child_exit;
+			VSC_C_mgt->child_exit++;
 		else
-			VSC_C_mgt->child_stop = ++static_VSC_C_mgt.child_stop;
+			VSC_C_mgt->child_stop++;
 	}
 	if (WIFSIGNALED(status)) {
 		VSB_printf(vsb, " signal=%d", WTERMSIG(status));
 		exit_status |= 0x40;
-		VSC_C_mgt->child_died = ++static_VSC_C_mgt.child_died;
+		VSC_C_mgt->child_died++;
 	}
 #ifdef WCOREDUMP
 	if (WCOREDUMP(status)) {
-		VSB_printf(vsb, " (core dumped)");
-		exit_status |= 0x80;
-		VSC_C_mgt->child_dump = ++static_VSC_C_mgt.child_dump;
+		VSB_cat(vsb, " (core dumped)");
+		if (!MGT_FEATURE(FEATURE_NO_COREDUMP))
+			exit_status |= 0x80;
+		VSC_C_mgt->child_dump++;
 	}
 #endif
-	AZ(VSB_finish(vsb));
-	MGT_complain(status ? C_ERR : C_INFO, "%s", VSB_data(vsb));
-	VSB_delete(vsb);
+	MGT_ComplainVSB(status ? C_ERR : C_INFO, vsb);
+	VSB_destroy(&vsb);
 
 	/* Dispose of shared memory but evacuate panic messages first */
 	if (heritage.panic_str[0] != '\0') {
 		mgt_panic_record(r);
-		mgt_SHM_Destroy(1);
-		VSC_C_mgt->child_panic = ++static_VSC_C_mgt.child_panic;
-	} else {
-		mgt_SHM_Destroy(MGT_DO_DEBUG(DBG_VSM_KEEP));
+		VSC_C_mgt->child_panic++;
 	}
-	mgt_SHM_Create();
-	mgt_SHM_Commit();
+
+	mgt_SHM_ChildDestroy();
 
 	if (child_state == CH_RUNNING)
 		child_state = CH_DIED;
 
 	/* Pick up any stuff lingering on stdout/stderr */
-	(void)child_listener(NULL, EV_RD);
-	closex(&child_output);
-	VLU_Destroy(child_std_vlu);
+	(void)child_listener(NULL, VEV__RD);
+	closefd(&child_output);
+	VLU_Destroy(&child_std_vlu);
 
 	child_pid = -1;
 
-	MGT_complain(C_DEBUG, "Child cleanup complete");
+	MGT_Complain(C_DEBUG, "Child cleanup complete");
+
+	/* XXX number of retries? interval? */
+	for (i = 0; i < 3; i++) {
+		if (VCA_reopen_sockets() == 0)
+			break;
+		/* error already logged */
+		(void)sleep(1);
+	}
+	if (i == 3) {
+		/* We failed to reopen our listening sockets. No choice
+		 * but to exit. */
+		MGT_Complain(C_ERR,
+		    "Could not reopen listening sockets. Exiting.");
+		exit(1);
+	}
 
 	if (child_state == CH_DIED && mgt_param.auto_restart)
 		mgt_launch_child(NULL);
@@ -620,19 +647,19 @@ mgt_reap_child(void)
  */
 
 void
-MGT_Child_Cli_Fail(void)
+MCH_Cli_Fail(void)
 {
 
-	if (child_state != CH_RUNNING)
+	if (child_state != CH_RUNNING && child_state != CH_STARTING)
 		return;
 	if (child_pid < 0)
 		return;
 	if (kill_child() == 0)
-		MGT_complain(C_ERR, "Child (%jd) not responding to CLI,"
+		MGT_Complain(C_ERR, "Child (%jd) not responding to CLI,"
 		    " killed it.", (intmax_t)child_pid);
 	else
-		MGT_complain(C_ERR, "Failed to kill child with PID %jd: %s",
-		    (intmax_t)child_pid, strerror(errno));
+		MGT_Complain(C_ERR, "Failed to kill child with PID %jd: %s",
+		    (intmax_t)child_pid, VAS_errtxt(errno));
 }
 
 /*=====================================================================
@@ -642,36 +669,85 @@ MGT_Child_Cli_Fail(void)
  */
 
 void
-mgt_stop_child(void)
+MCH_Stop_Child(void)
 {
 
-	if (child_state != CH_RUNNING)
+	if (child_state != CH_RUNNING && child_state != CH_STARTING)
 		return;
 
 	child_state = CH_STOPPING;
 
-	MGT_complain(C_DEBUG, "Stopping Child");
+	MGT_Complain(C_DEBUG, "Stopping Child");
 
 	mgt_reap_child();
 }
 
 /*=====================================================================
- * CLI command to start/stop child
  */
 
-void __match_proto__(cli_func_t)
-mcf_server_startstop(struct cli *cli, const char * const *av, void *priv)
+int
+MCH_Start_Child(void)
+{
+	mgt_launch_child(NULL);
+	if (child_state != CH_RUNNING)
+		return (2);
+	return (0);
+}
+
+/*====================================================================
+ * Query if the child is running
+ */
+
+int
+MCH_Running(void)
+{
+
+	return (child_pid > 0);
+}
+
+/*=====================================================================
+ * CLI commands
+ */
+
+static void v_matchproto_(cli_func_t)
+mch_pid(struct cli *cli, const char * const *av, void *priv)
 {
 
 	(void)av;
-	if (priv != NULL && child_state == CH_RUNNING)
-		mgt_stop_child();
-	else if (priv == NULL && child_state == CH_STOPPED) {
-		if (mgt_has_vcl()) {
+	(void)priv;
+	VCLI_Out(cli, "Master: %10jd\n", (intmax_t)getpid());
+	if (!MCH_Running())
+		return;
+	VCLI_Out(cli, "Worker: %10jd\n", (intmax_t)child_pid);
+}
+
+static void v_matchproto_(cli_func_t)
+mch_pid_json(struct cli *cli, const char * const *av, void *priv)
+{
+
+	(void)priv;
+	VCLI_JSON_begin(cli, 2, av);
+	VCLI_Out(cli, ",\n  {\"master\": %jd", (intmax_t)getpid());
+	if (MCH_Running())
+		VCLI_Out(cli, ", \"worker\": %jd", (intmax_t)child_pid);
+	VCLI_Out(cli, "}");
+	VCLI_JSON_end(cli);
+}
+
+static void v_matchproto_(cli_func_t)
+mch_cli_server_start(struct cli *cli, const char * const *av, void *priv)
+{
+	const char *err;
+
+	(void)av;
+	(void)priv;
+	if (child_state == CH_STOPPED) {
+		err = mgt_has_vcl();
+		if (err == NULL) {
 			mgt_launch_child(cli);
 		} else {
 			VCLI_SetResult(cli, CLIS_CANT);
-			VCLI_Out(cli, "No VCL available");
+			VCLI_Out(cli, "%s", err);
 		}
 	} else {
 		VCLI_SetResult(cli, CLIS_CANT);
@@ -679,46 +755,49 @@ mcf_server_startstop(struct cli *cli, const char * const *av, void *priv)
 	}
 }
 
-/*--------------------------------------------------------------------*/
+static void v_matchproto_(cli_func_t)
+mch_cli_server_stop(struct cli *cli, const char * const *av, void *priv)
+{
 
-void
-mcf_server_status(struct cli *cli, const char * const *av, void *priv)
+	(void)av;
+	(void)priv;
+	if (child_state == CH_RUNNING) {
+		MCH_Stop_Child();
+	} else {
+		VCLI_SetResult(cli, CLIS_CANT);
+		VCLI_Out(cli, "Child in state %s", ch_state[child_state]);
+	}
+}
+
+static void v_matchproto_(cli_func_t)
+mch_cli_server_status(struct cli *cli, const char * const *av, void *priv)
 {
 	(void)av;
 	(void)priv;
 	VCLI_Out(cli, "Child in state %s", ch_state[child_state]);
 }
 
-/*--------------------------------------------------------------------*/
-
-static int __match_proto__(vev_cb_f)
-mgt_sigint(const struct vev *e, int what)
+static void v_matchproto_(cli_func_t)
+mch_cli_server_status_json(struct cli *cli, const char * const *av, void *priv)
 {
-
-	(void)e;
-	(void)what;
-	MGT_complain(C_ERR, "Manager got SIGINT");
-	(void)fflush(stdout);
-	if (child_pid >= 0)
-		mgt_stop_child();
-	exit(0);
+	(void)priv;
+	VCLI_JSON_begin(cli, 2, av);
+	VCLI_Out(cli, ", ");
+	VCLI_JSON_str(cli, ch_state[child_state]);
+	VCLI_JSON_end(cli);
 }
 
-/*--------------------------------------------------------------------*/
-
-static int __match_proto__(vev_cb_f)
-mgt_uptime(const struct vev *e, int what)
-{
-
-	(void)e;
-	(void)what;
-	AN(VSC_C_mgt);
-	VSC_C_mgt->uptime = static_VSC_C_mgt.uptime =
-	    (uint64_t)(VTIM_real() - mgt_uptime_t0);
-	if (heritage.vsm != NULL)
-		VSM_common_ageupdate(heritage.vsm);
-	return (0);
-}
+static struct cli_proto cli_mch[] = {
+	{ CLICMD_SERVER_STATUS,		"", mch_cli_server_status,
+	  mch_cli_server_status_json },
+	{ CLICMD_SERVER_START,		"", mch_cli_server_start },
+	{ CLICMD_SERVER_STOP,		"", mch_cli_server_stop },
+	{ CLICMD_PANIC_SHOW,		"", mch_cli_panic_show,
+	  mch_cli_panic_show_json },
+	{ CLICMD_PANIC_CLEAR,		"", mch_cli_panic_clear },
+	{ CLICMD_PID,			"", mch_pid, mch_pid_json },
+	{ NULL }
+};
 
 /*=====================================================================
  * This thread is the master thread in the management process.
@@ -727,65 +806,8 @@ mgt_uptime(const struct vev *e, int what)
  */
 
 void
-MGT_Run(void)
+MCH_Init(void)
 {
-	struct sigaction sac;
-	struct vev *e;
-	int i;
 
-	mgt_uptime_t0 = VTIM_real();
-	e = vev_new();
-	XXXAN(e);
-	e->callback = mgt_uptime;
-	e->timeout = 1.0;
-	e->name = "mgt_uptime";
-	AZ(vev_add(mgt_evb, e));
-
-	e = vev_new();
-	XXXAN(e);
-	e->sig = SIGTERM;
-	e->callback = mgt_sigint;
-	e->name = "mgt_sigterm";
-	AZ(vev_add(mgt_evb, e));
-
-	e = vev_new();
-	XXXAN(e);
-	e->sig = SIGINT;
-	e->callback = mgt_sigint;
-	e->name = "mgt_sigint";
-	AZ(vev_add(mgt_evb, e));
-
-#ifdef HAVE_SETPROCTITLE
-	setproctitle("Varnish-Mgr %s", heritage.name);
-#endif
-
-	memset(&sac, 0, sizeof sac);
-	sac.sa_handler = SIG_IGN;
-	sac.sa_flags = SA_RESTART;
-
-	AZ(sigaction(SIGPIPE, &sac, NULL));
-	AZ(sigaction(SIGHUP, &sac, NULL));
-
-	if (!d_flag && !mgt_has_vcl())
-		MGT_complain(C_ERR, "No VCL loaded yet");
-	else if (!d_flag) {
-		mgt_launch_child(NULL);
-		if (child_state != CH_RUNNING) {
-			// XXX correct? or 0?
-			exit_status = 2;
-			return;
-		}
-	}
-
-	i = mgt_SHM_Commit();
-	if (i != 0) {
-		MGT_complain(C_ERR, "Could not commit SHM file");
-		return;
-	}
-
-	i = vev_schedule(mgt_evb);
-	if (i != 0)
-		MGT_complain(C_ERR, "vev_schedule() = %d", i);
-
-	MGT_complain(C_INFO, "manager dies");
+	VCLS_AddFunc(mgt_cls, MCF_AUTH, cli_mch);
 }
