@@ -61,6 +61,10 @@ static struct VSC_lck *lck_smp;
 
 static void smp_init(void);
 
+#ifndef WITH_PERSISTENT_STORAGE
+#error "WITH_PERSISTENT_STORAGE must be defined"
+#endif
+
 /*--------------------------------------------------------------------*/
 
 /*
@@ -363,7 +367,7 @@ smp_open(struct stevedore *st)
 		AZ(smp_open_segs(sc, &sc->seg2));
 
 	/*
-	 * Grap a reference to the tail of the ban list, until the thread
+	 * Grab a reference to the tail of the ban list, until the thread
 	 * has loaded all objects, so we can be sure that all of our
 	 * proto-bans survive until then.
 	 */
@@ -692,6 +696,7 @@ smp_init(void)
 	smp_oc_realmethods.objsetattr = SML_methods.objsetattr;
 	smp_oc_realmethods.objtouch = LRU_Touch;
 	smp_oc_realmethods.objfree = smp_oc_objfree;
+	smp_oc_realmethods.vai_init	= SML_methods.vai_init;
 }
 
 /*--------------------------------------------------------------------
