@@ -156,11 +156,10 @@ struct vcf {
 
 /* Prototypes etc ----------------------------------------------------*/
 
-/* cache_acceptor.c */
-void VCA_Init(void);
-void VCA_Shutdown(void);
-
 /* cache_backend.c */
+struct backend;
+
+/* cache_backend_cfg.c */
 void VBE_InitCfg(void);
 
 /* cache_ban.c */
@@ -176,6 +175,8 @@ vtim_real BAN_Time(const struct ban *ban);
 /* cache_busyobj.c */
 struct busyobj *VBO_GetBusyObj(const struct worker *, const struct req *);
 void VBO_ReleaseBusyObj(struct worker *wrk, struct busyobj **busyobj);
+void VBO_SetState(struct worker *wrk, struct busyobj *bo,
+    enum boc_state_e next);
 
 /* cache_director.c */
 int VDI_GetHdr(struct busyobj *);
@@ -194,11 +195,17 @@ uint64_t VDP_Close(struct vdp_ctx *, struct objcore *, struct boc *);
 void VDP_Panic(struct vsb *vsb, const struct vdp_ctx *vdc);
 int VDP_Push(VRT_CTX, struct vdp_ctx *, struct ws *, const struct vdp *,
     void *priv);
+int VDP_ObjIterate(void *priv, unsigned flush, const void *ptr, ssize_t len);
 int VDP_DeliverObj(struct vdp_ctx *vdc, struct objcore *oc);
 extern const struct vdp VDP_gunzip;
 extern const struct vdp VDP_esi;
 extern const struct vdp VDP_range;
 
+uint64_t VDPIO_Close(struct vdp_ctx *, struct objcore *, struct boc *);
+int VDPIO_Upgrade(VRT_CTX, struct vdp_ctx *vdc);
+int VDPIO_Push(VRT_CTX, struct vdp_ctx *, struct ws *, const struct vdp *,
+    void *priv);
+
 
 /* cache_exp.c */
 vtim_real EXP_Ttl(const struct req *, const struct objcore *);
@@ -300,7 +307,8 @@ htc_complete_f HTTP1_Complete;
 uint16_t HTTP1_DissectRequest(struct http_conn *, struct http *);
 uint16_t HTTP1_DissectResponse(struct http_conn *, struct http *resp,
     const struct http *req);
-unsigned HTTP1_Write(const struct worker *w, const struct http *hp, const int*);
+struct v1l;
+unsigned HTTP1_Write(struct v1l *v1l, const struct http *hp, const int*);
 
 /* cache_main.c */
 vxid_t VXID_Get(const struct worker *, uint64_t marker);
@@ -336,8 +344,8 @@ int ObjGetSpace(struct worker *, struct objcore *, ssize_t *sz, uint8_t **ptr);
 void ObjExtend(struct worker *, struct objcore *, ssize_t l, int final);
 uint64_t ObjWaitExtend(const struct worker *, const struct objcore *,
     uint64_t l, enum boc_state_e *statep);
-void ObjSetState(struct worker *, const struct objcore *,
-    enum boc_state_e next);
+void ObjSetState(struct worker *, struct objcore *, enum boc_state_e next,
+    unsigned broadcast);
 void ObjWaitState(const struct objcore *, enum boc_state_e want);
 void ObjTouch(struct worker *, struct objcore *, vtim_real now);
 void ObjFreeObj(struct worker *, struct objcore *);
@@ -347,6 +355,10 @@ void *ObjSetAttr(struct worker *, struct objcore *, enum obj_attr,
 int ObjCopyAttr(struct worker *, struct objcore *, struct objcore *,
     enum obj_attr attr);
 void ObjBocDone(struct worker *, struct objcore *, struct boc **);
+// VAI
+uint64_t ObjVAIGetExtend(struct worker *, const struct objcore *, uint64_t,
+    enum boc_state_e *, struct vai_qe *);
+void ObjVAICancel(struct worker *, struct boc *, struct vai_qe *);
 
 int ObjSetDouble(struct worker *, struct objcore *, enum obj_attr, double);
 int ObjSetU64(struct worker *, struct objcore *, enum obj_attr, uint64_t);
@@ -410,7 +422,7 @@ void pan_pool(struct vsb *);
 int VRG_CheckBo(struct busyobj *);
 
 /* cache_req.c */
-struct req *Req_New(struct sess *);
+struct req *Req_New(struct sess *, const struct req *);
 void Req_Release(struct req *);
 void Req_Rollback(VRT_CTX);
 void Req_Cleanup(struct sess *sp, struct worker *wrk, struct req *req);
@@ -471,6 +483,7 @@ void VSL_Flush(struct vsl_log *, int overflow);
 struct conn_pool;
 void VCP_Init(void);
 void VCP_Panic(struct vsb *, struct conn_pool *);
+void VCP_RelPoll(void);
 
 /* cache_backend_probe.c */
 void VBP_Init(void);
@@ -491,6 +504,7 @@ struct vsb *VCL_Rel_CliCtx(struct vrt_ctx **);
 void VCL_Panic(struct vsb *, const char *nm, const struct vcl *);
 void VCL_Poll(void);
 void VCL_Init(void);
+void VCL_Shutdown(void);
 
 #define VCL_MET_MAC(l,u,t,b) \
     void VCL_##l##_method(struct vcl *, struct worker *, struct req *, \
@@ -570,6 +584,7 @@ void V2D_Init(void);
 
 /* stevedore.c */
 void STV_open(void);
+void STV_warn(void);
 void STV_close(void);
 const struct stevedore *STV_next(void);
 int STV_BanInfoDrop(const uint8_t *ban, unsigned len);
@@ -585,7 +600,7 @@ struct stv_buffer *STV_AllocBuf(struct worker *wrk, const struct stevedore *stv,
 void STV_FreeBuf(struct worker *wrk, struct stv_buffer **pstvbuf);
 void *STV_GetBufPtr(struct stv_buffer *stvbuf, size_t *psize);
 
-#if WITH_PERSISTENT_STORAGE
+#ifdef WITH_PERSISTENT_STORAGE
 /* storage_persistent.c */
 void SMP_Ready(void);
 #endif
