@@ -25,20 +25,27 @@
 
 #define ALLOC_OBJ(to, type_magic)					\
 	do {								\
-		(to) = calloc(1, sizeof *(to));				\
+		(to) = calloc((size_t)1, sizeof *(to));			\
 		if ((to) != NULL)					\
 			(to)->magic = (type_magic);			\
 	} while (0)
 
 #define ALLOC_OBJ_EXTRA(to, extra_size, type_magic)			\
 	do {								\
-		(to) = calloc(1, sizeof(*(to)) + (extra_size));		\
+		(to) = calloc((size_t)1, sizeof(*(to)) + (extra_size));	\
 		if ((to) != NULL)					\
 			(to)->magic = (type_magic);			\
 	} while (0)
 
+#define SIZEOF_FLEX_OBJ(to, fld, len)					\
+	(offsetof(typeof(*to), fld) + sizeof *(to)->fld * (len))
+
 #define ALLOC_FLEX_OBJ(to, fld, len, type_magic)			\
-	ALLOC_OBJ_EXTRA(to, (len) * sizeof *((to)->fld), (type_magic))
+	do {								\
+		(to) = calloc((size_t)1, SIZEOF_FLEX_OBJ(to, fld, len));\
+		if ((to) != NULL)					\
+			(to)->magic = (type_magic);			\
+	} while (0)
 
 #define FREE_OBJ(to)							\
 	do {								\
