@@ -484,6 +484,21 @@ PARAM_SIMPLE(
 	/* dyn_def_reason */	"2m"
 )
 
+PARAM_SIMPLE(
+	/* name */	uncacheable_ttl,
+	/* type */	duration,
+	/* min */	"0.000",
+	/* max */	NULL,
+	/* def */	"2m",
+	/* units */	"seconds",
+	/* descr */
+	"The TTL assigned to uncacheable objects by the built-in VCL.",
+	/* flags */	OBJ_STICKY,
+	/* dyn_min_reason */	NULL,
+	/* dyn_max_reason */	NULL,
+	/* dyn_def_reason */	"2m"
+)
+
 PARAM_SIMPLE(
 	/* name */	http1_iovs,
 	/* type */	uint,
@@ -538,10 +553,12 @@ PARAM_SIMPLE(
 	/* units */	"bytes",
 	/* descr */
 	"Size of malloc buffer used for gzip processing.\n"
+	"Size of buffer used for gzip processing.\n"
 	"These buffers are used for in-transit data, for instance "
-	"gunzip'ed data being sent to a client.Making this space to small "
+	"gunzip'ed data being sent to a client. Making this space to small "
 	"results in more overhead, writes to sockets etc, making it too "
-	"big is probably just a waste of memory.",
+	"big is probably just a waste of memory.\n"
+	"Gzip buffers are allocated from Transient storage.",
 	/* flags */	EXPERIMENTAL
 )
 
@@ -653,7 +670,7 @@ PARAM_SIMPLE(
 	/* name */	http_req_overflow_status,
 	/* type */	uint_orzero,
 	/* min */	"400",
-	/* max */	"499",
+	/* max */	"500",
 	/* def */	"0",
 	/* units */	"HTTP status code or 0 to disable",
 	/* descr */
@@ -663,7 +680,9 @@ PARAM_SIMPLE(
 	"Note that there is no standard HTTP status which exactly matches "
 	"the implementation of http_req_size. 414 applies to the URL only, "
 	"while 413 applies to the request body. 400 is probably the least "
-	"incorrect alternative value to sending no response at all (0)."
+	"incorrect alternative value to sending no response at all (0). It "
+	"can also be seen as an internal error on the Varnish side due to "
+	"configured limits, so 500 is also allowed."
 )
 
 PARAM_SIMPLE(
@@ -743,7 +762,7 @@ PARAM_SIMPLE(
 	/* type */	uint,
 	/* min */	"0",
 	/* max */	NULL,
-	/* def */	"10000",
+	/* def */	"0",
 	/* units */	"checks",
 	/* descr */
 	"Maximum number of possibly non matching variants that we evaluate "
@@ -891,7 +910,7 @@ PARAM_SIMPLE(
 PARAM_SIMPLE(
 	/* name */	rush_exponent,
 	/* type */	uint,
-	/* min */	"2",
+	/* min */	"1",
 	/* max */	NULL,
 	/* def */	"3",
 	/* units */	"requests per request",
@@ -899,7 +918,9 @@ PARAM_SIMPLE(
 	"How many parked request we start for each completed request on "
 	"the object.\n"
 	"NB: Even with the implicit delay of delivery, this parameter "
-	"controls an exponential increase in number of worker threads.",
+	"controls an exponential increase in number of worker threads. "
+	"A value of 1 will instead serialize requests resumption and is "
+	"only useful for testing purposes.",
 	/* flags */	EXPERIMENTAL
 )
 
