@@ -46,7 +46,7 @@
 #  error "include vdef.h before vrt.h"
 #endif
 
-#define VRT_MAJOR_VERSION	21U
+#define VRT_MAJOR_VERSION	22U
 
 #define VRT_MINOR_VERSION	0U
 
@@ -57,6 +57,51 @@
  * Whenever something is deleted or changed in a way which is not
  * binary/load-time compatible, increment MAJOR version
  *
+ * 22.0 (2025-09-15)
+ *	VRT_r_obj_stale_age() added
+ *	VRT_r_obj_stale_can_esi() added
+ *	VRT_r_obj_stale_grace() added
+ *	VRT_r_obj_stale_hits() added
+ *	VRT_r_obj_stale_keep() added
+ *	VRT_r_obj_stale_proto() added
+ *	VRT_r_obj_stale_reason() added
+ *	VRT_r_obj_stale_status() added
+ *	VRT_r_obj_stale_storage() added
+ *	VRT_r_obj_stale_time() added
+ *	VRT_r_obj_stale_ttl() added
+ *	VRT_r_obj_stale_uncacheable() added
+ *	VRT_r_obj_stale_is_valid() added
+ *	enum gethdr_e has new value HDR_OBJ_STALE
+ *	typedef hdr_t added
+ *	struct gethdr_s.what changed to hdr_t
+ *	VRT_VSC_Alloc() renamed to VRT_VSC_Allocv()
+ *	new VRT_VSC_Alloc() added
+ *	VRT_r_param_backend_idle_timeout() added
+ *	VRT_r_param_backend_wait_limit() added
+ *	VRT_r_param_backend_wait_timeout() added
+ *	VRT_r_param_between_bytes_timeout() added
+ *	VRT_r_param_connect_timeout() added
+ *	VRT_r_param_default_grace() added
+ *	VRT_r_param_default_keep() added
+ *	VRT_r_param_default_ttl() added
+ *	VRT_r_param_first_byte_timeout() added
+ *	VRT_r_param_idle_send_timeout() added
+ *	VRT_r_param_max_esi_depth() added
+ *	VRT_r_param_max_restarts() added
+ *	VRT_r_param_max_retries() added
+ *	VRT_r_param_pipe_task_deadline() added
+ *	VRT_r_param_pipe_timeout() added
+ *	VRT_r_param_send_timeout() added
+ *	VRT_r_param_shortlived() added
+ *	VRT_r_param_timeout_idle() added
+ *	VRT_r_param_transit_buffer() added
+ *	VRT_r_param_uncacheable_ttl() added
+ *	VRT_AddVFP() removed
+ *	VRT_AddVDP() removed
+ *	VRT_RemoveVFP() removed
+ *	VRT_RemoveVDP() removed
+ *	struct vrt_blob magic added
+ *	struct strands magic added;
  * 21.0 (2025-03-17)
  *	VRT_u_req_grace() added
  *	VRT_u_req_ttl() added
@@ -343,6 +388,8 @@ typedef const struct stream_close *stream_close_t;
  */
 
 struct strands {
+	unsigned	magic;
+#define STRANDS_MAGIC	0x5d5ab196
 	int		n;
 	const char	**p;
 };
@@ -350,13 +397,15 @@ struct strands {
 /*
  * A VCL_STRANDS return value must never be NULL. Use this instead
  */
-extern const struct strands *vrt_null_strands;
+extern const struct strands *const vrt_null_strands;
 
 /*
  * Macros for VCL_STRANDS creation
  */
-#define TOSTRAND(s)(&(struct strands){.n=1,.p=(const char *[1]){(s)}})
-#define TOSTRANDS(x, ...)(&(struct strands){.n=x,.p=(const char *[x]){__VA_ARGS__}})
+#define TOSTRAND(s)(&(struct strands)				\
+	{.magic=STRANDS_MAGIC,.n=1,.p=(const char *[1]){(s)}})
+#define TOSTRANDS(x, ...)(&(struct strands)			\
+	{.magic=STRANDS_MAGIC,.n=x,.p=(const char *[x]){__VA_ARGS__}})
 
 /***********************************************************************
  * VCL_BLOB:
@@ -378,13 +427,15 @@ extern const struct strands *vrt_null_strands;
  */
 
 struct vrt_blob {
+	unsigned	magic;
+#define VRT_BLOB_MAGIC	0xe114db9e
 	unsigned	type;
 	size_t		len;
 	const void	*blob;
 };
 
 #define VRT_NULL_BLOB_TYPE 0xfade4fa0
-extern const struct vrt_blob *vrt_null_blob;
+extern const struct vrt_blob *const vrt_null_blob;
 
 /***********************************************************************
  * This is the central definition of the mapping from VCL types to
@@ -690,12 +741,40 @@ enum gethdr_e {
 	HDR_RESP,
 	HDR_OBJ,
 	HDR_BEREQ,
-	HDR_BERESP
+	HDR_BERESP,
+	HDR_OBJ_STALE
 };
 
+typedef const struct {
+	unsigned char	len;
+	const char	str[];
+} *hdr_t;
+
+//lint -esym(754, _hdr_s, _hdr_l) local struct member ... not referenced
+//lint -emacro(740, HDR) unusual cast
+#define HDR(name)							\
+	((hdr_t)&(const struct {					\
+		unsigned char _hdr_l;					\
+		char _hdr_s[sizeof(name ":")];				\
+	}){ sizeof(name), name ":" })
+
+#define CHECK_HDR(hdr)							\
+	do {								\
+		AN(hdr);						\
+		assert((hdr)->len > 0);					\
+		assert((hdr)->len == strlen((hdr)->str));		\
+		assert((hdr)->str[(hdr)->len - 1] == ':');		\
+	} while (0)
+
+#define CAST_HDR(hdr, str)						\
+	do {								\
+		hdr = (hdr_t)(str);					\
+		CHECK_HDR(hdr);						\
+	} while (0)
+
 struct gethdr_s {
 	enum gethdr_e	where;
-	const char	*what;
+	hdr_t		what;
 };
 
 VCL_HTTP VRT_selecthttp(VRT_CTX, enum gethdr_e);
@@ -822,8 +901,10 @@ struct vsmw_cluster *VRT_VSM_Cluster_New(VRT_CTX, size_t);
 void VRT_VSM_Cluster_Destroy(VRT_CTX, struct vsmw_cluster **);
 
 #ifdef va_start	// XXX: hackish
-void *VRT_VSC_Alloc(struct vsmw_cluster *, struct vsc_seg **,
+void *VRT_VSC_Allocv(struct vsmw_cluster *, struct vsc_seg **,
     const char *, size_t, const unsigned char *, size_t, const char *, va_list);
+void *VRT_VSC_Alloc(struct vsmw_cluster *, struct vsc_seg **,
+    const char *, size_t, const unsigned char *, size_t, const char *, ...);
 #endif
 void VRT_VSC_Destroy(const char *, struct vsc_seg *);
 void VRT_VSC_Hide(const struct vsc_seg *);
