@@ -130,19 +130,13 @@ init(void)
 vtim_mono
 VTIM_mono(void)
 {
-#if defined(HAVE_CLOCK_GETTIME) && !defined(USE_GETHRTIME)
+#if defined(HAVE_GETHRTIME) && defined(USE_GETHRTIME)
+	return (gethrtime() * 1e-9);
+#else
 	struct timespec ts;
 
 	AZ(clock_gettime(CLOCK_MONOTONIC, &ts));
 	return (ts.tv_sec + 1e-9 * ts.tv_nsec);
-#elif defined(HAVE_GETHRTIME)
-	return (gethrtime() * 1e-9);
-#elif defined(__MACH__)
-	uint64_t mt = mach_absolute_time() - mt_base;
-
-	return (mt * mt_scale);
-#else
-#error Varnish needs some monotonic time source
 #endif
 }
 
@@ -163,7 +157,7 @@ VTIM_real(void)
 }
 
 void
-VTIM_format(vtim_real t, char *p)
+VTIM_format(vtim_real t, char p[VTIM_FORMAT_SIZE])
 {
 	struct tm tm;
 	time_t tt;
@@ -409,22 +403,11 @@ VTIM_parse(const char *p)
 void
 VTIM_sleep(vtim_dur t)
 {
-#ifdef HAVE_NANOSLEEP
 	struct timespec ts;
 
 	ts = VTIM_timespec(t);
 
 	(void)nanosleep(&ts, NULL);
-#else
-	if (t >= 1.) {
-		(void)sleep(floor(t));
-		t -= floor(t);
-	}
-	/* XXX: usleep() is not mandated to be thread safe */
-	t *= 1e6;
-	if (t > 0)
-		(void)usleep(floor(t));
-#endif
 }
 
 /*
@@ -602,8 +585,8 @@ parse_check(time_t t, const char *s)
 	}
 }
 
-#define TTEST_MIN (sizeof(time_t) >= 8 ? -2209852800 : INT32_MIN)
-#define TTEST_MAX (sizeof(time_t) >= 8 ? 20000000000 : INT32_MAX)
+#define TTEST_MIN (sizeof(time_t) >= 8 ? -2209852800LL : INT32_MIN)
+#define TTEST_MAX (sizeof(time_t) >= 8 ? 20000000000LL : INT32_MAX)
 
 int
 main(int argc, char **argv)
