@@ -5,6 +5,8 @@
  *
  * Author: Martin Blix Grydeland <martin@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -28,12 +30,13 @@
  *
  */
 
+#include "config.h"
+
 #include <stdarg.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
 
 #include "vdef.h"
 #include "vas.h"
@@ -73,6 +76,7 @@ static const char * const vsl_r_names[VSL_r__MAX] = {
 };
 
 struct vtx;
+VTAILQ_HEAD(vtxhead, vtx);
 
 struct vslc_raw {
 	unsigned		magic;
@@ -89,7 +93,7 @@ struct synth {
 
 	VTAILQ_ENTRY(synth)	list;
 	size_t			offset;
-	uint32_t		data[2 + 64 / sizeof (uint32_t)];
+	uint32_t		data[VSL_OVERHEAD + VSL_WORDS(64)];
 };
 VTAILQ_HEAD(synthhead, synth);
 
@@ -133,10 +137,10 @@ struct vslc_vtx {
 };
 
 struct vtx_key {
-	unsigned		vxid;
-	VRB_ENTRY(vtx_key)	entry;
+	uint64_t		vxid;
+	VRBT_ENTRY(vtx_key)	entry;
 };
-VRB_HEAD(vtx_tree, vtx_key);
+VRBT_HEAD(vtx_tree, vtx_key);
 
 struct vtx {
 	struct vtx_key		key;
@@ -158,12 +162,12 @@ struct vtx {
 	enum VSL_reason_e	reason;
 
 	struct vtx		*parent;
-	VTAILQ_HEAD(,vtx)	child;
+	struct vtxhead		child;
 	unsigned		n_child;
 	unsigned		n_childready;
 	unsigned		n_descend;
 
-	VTAILQ_HEAD(,synth)	synth;
+	struct synthhead	synth;
 
 	struct chunk		shmchunks[VTX_SHMCHUNKS];
 	struct chunkhead	shmchunks_free;
@@ -186,13 +190,17 @@ struct VSLQ {
 
 	/* Structured mode */
 	struct vtx_tree		tree;
-	VTAILQ_HEAD(,vtx)	ready;
-	VTAILQ_HEAD(,vtx)	incomplete;
-	unsigned		n_outstanding;
+	struct vtxhead		ready;
+	struct vtxhead		incomplete;
+	int			n_outstanding;
 	struct chunkhead	shmrefs;
-	VTAILQ_HEAD(,vtx)	cache;
+	struct vtxhead		cache;
 	unsigned		n_cache;
 
+	/* Rate limiting */
+	double			credits;
+	vtim_mono		last_use;
+
 	/* Raw mode */
 	struct {
 		struct vslc_raw		c;
@@ -200,7 +208,7 @@ struct VSLQ {
 		struct VSL_transaction	*ptrans[2];
 		struct VSLC_ptr		start;
 		ssize_t			len;
-		size_t			offset;
+		ssize_t			offset;
 	} raw;
 };
 
@@ -221,10 +229,14 @@ vtx_keycmp(const struct vtx_key *a, const struct vtx_key *b)
 	return (0);
 }
 
-VRB_PROTOTYPE_STATIC(vtx_tree, vtx_key, entry, vtx_keycmp)
-VRB_GENERATE_STATIC(vtx_tree, vtx_key, entry, vtx_keycmp)
+VRBT_GENERATE_REMOVE_COLOR(vtx_tree, vtx_key, entry, static)
+VRBT_GENERATE_REMOVE(vtx_tree, vtx_key, entry, static)
+VRBT_GENERATE_INSERT_COLOR(vtx_tree, vtx_key, entry, static)
+VRBT_GENERATE_INSERT_FINISH(vtx_tree, vtx_key, entry, static)
+VRBT_GENERATE_INSERT(vtx_tree, vtx_key, entry, vtx_keycmp, static)
+VRBT_GENERATE_FIND(vtx_tree, vtx_key, entry, vtx_keycmp, static)
 
-static int
+static enum vsl_status v_matchproto_(vslc_next_f)
 vslc_raw_next(const struct VSL_cursor *cursor)
 {
 	struct vslc_raw *c;
@@ -235,14 +247,14 @@ vslc_raw_next(const struct VSL_cursor *cursor)
 	AN(c->ptr);
 	if (c->cursor.rec.ptr == NULL) {
 		c->cursor.rec.ptr = c->ptr;
-		return (1);
+		return (vsl_more);
 	} else {
 		c->cursor.rec.ptr = NULL;
-		return (0);
+		return (vsl_end);
 	}
 }
 
-static int
+static enum vsl_status v_matchproto_(vslc_reset_f)
 vslc_raw_reset(const struct VSL_cursor *cursor)
 {
 	struct vslc_raw *c;
@@ -253,7 +265,7 @@ vslc_raw_reset(const struct VSL_cursor *cursor)
 	AN(c->ptr);
 	c->cursor.rec.ptr = NULL;
 
-	return (0);
+	return (vsl_end);
 }
 
 static const struct vslc_tbl vslc_raw_tbl = {
@@ -264,11 +276,12 @@ static const struct vslc_tbl vslc_raw_tbl = {
 	.check	= NULL,
 };
 
-static int
+static enum vsl_status v_matchproto_(vslc_next_f)
 vslc_vtx_next(const struct VSL_cursor *cursor)
 {
 	struct vslc_vtx *c;
 	const uint32_t *ptr;
+	unsigned overrun;
 
 	CAST_OBJ_NOTNULL(c, cursor->priv_data, VSLC_VTX_MAGIC);
 	assert(&c->cursor == cursor);
@@ -282,10 +295,10 @@ vslc_vtx_next(const struct VSL_cursor *cursor)
 			c->cursor.rec.ptr = c->synth->data;
 			c->synth = VTAILQ_NEXT(c->synth, list);
 		} else {
-			assert(c->offset <= c->vtx->len);
+			overrun = c->offset > c->vtx->len;
+			AZ(overrun);
 			if (c->offset == c->vtx->len)
-				/* End of cursor */
-				return (0);
+				return (vsl_end);
 
 			/* Advance chunk pointer */
 			if (c->chunk == NULL) {
@@ -312,10 +325,10 @@ vslc_vtx_next(const struct VSL_cursor *cursor)
 		}
 	} while (VSL_TAG(c->cursor.rec.ptr) == SLT__Batch);
 
-	return (1);
+	return (vsl_more);
 }
 
-static int
+static enum vsl_status v_matchproto_(vslc_reset_f)
 vslc_vtx_reset(const struct VSL_cursor *cursor)
 {
 	struct vslc_vtx *c;
@@ -329,7 +342,7 @@ vslc_vtx_reset(const struct VSL_cursor *cursor)
 	c->offset = 0;
 	c->cursor.rec.ptr = NULL;
 
-	return (0);
+	return (vsl_end);
 }
 
 static const struct vslc_tbl vslc_vtx_tbl = {
@@ -364,12 +377,12 @@ chunk_newbuf(struct vtx *vtx, const uint32_t *ptr, size_t len)
 static void
 chunk_freebuf(struct chunk **pchunk)
 {
+	struct chunk *chunk;
 
-	CHECK_OBJ_NOTNULL(*pchunk, CHUNK_MAGIC);
-	assert((*pchunk)->type == chunk_t_buf);
-	free((*pchunk)->buf.data);
-	FREE_OBJ(*pchunk);
-	*pchunk = NULL;
+	TAKE_OBJ_NOTNULL(chunk, pchunk, CHUNK_MAGIC);
+	assert(chunk->type == chunk_t_buf);
+	free(chunk->buf.data);
+	FREE_OBJ(chunk);
 }
 
 /* Append a set of records to a chunk */
@@ -423,19 +436,22 @@ chunk_shm_to_buf(struct VSLQ *vslq, struct chunk *chunk)
 }
 
 /* Append a set of records to a vtx structure */
-static void
+static enum vsl_status
 vtx_append(struct VSLQ *vslq, struct vtx *vtx, const struct VSLC_ptr *start,
     size_t len)
 {
 	struct chunk *chunk;
+	enum vsl_check i;
 
 	AN(vtx);
-	if (len == 0)
-		return;
+	AN(len);
 	AN(start);
 
-	if (VSL_Check(vslq->c, start) == 2 &&
-	    !VTAILQ_EMPTY(&vtx->shmchunks_free)) {
+	i = VSL_Check(vslq->c, start);
+	if (i == vsl_check_e_inval)
+		return (vsl_e_overrun);
+
+	if (i == vsl_check_valid && !VTAILQ_EMPTY(&vtx->shmchunks_free)) {
 		/* Shmref it */
 		chunk = VTAILQ_FIRST(&vtx->shmchunks_free);
 		CHECK_OBJ_NOTNULL(chunk, CHUNK_MAGIC);
@@ -463,6 +479,7 @@ vtx_append(struct VSLQ *vslq, struct vtx *vtx, const struct VSLC_ptr *start,
 		}
 	}
 	vtx->len += len;
+	return (vsl_more);
 }
 
 /* Allocate a new vtx structure */
@@ -510,7 +527,7 @@ vtx_new(struct VSLQ *vslq)
 	vtx->n_childready = 0;
 	vtx->n_descend = 0;
 	vtx->len = 0;
-	(void)vslc_vtx_reset(&vtx->c.cursor);
+	AN(vslc_vtx_reset(&vtx->c.cursor) == vsl_end);
 
 	return (vtx);
 }
@@ -526,11 +543,7 @@ vtx_retire(struct VSLQ *vslq, struct vtx **pvtx)
 	struct chunk *chunk;
 
 	AN(vslq);
-	AN(pvtx);
-
-	vtx = *pvtx;
-	*pvtx = NULL;
-	CHECK_OBJ_NOTNULL(vtx, VTX_MAGIC);
+	TAKE_OBJ_NOTNULL(vtx, pvtx, VTX_MAGIC);
 
 	AN(vtx->flags & VTX_F_COMPLETE);
 	AN(vtx->flags & VTX_F_READY);
@@ -551,7 +564,8 @@ vtx_retire(struct VSLQ *vslq, struct vtx **pvtx)
 	AZ(vtx->n_child);
 	AZ(vtx->n_descend);
 	vtx->n_childready = 0;
-	AN(VRB_REMOVE(vtx_tree, &vslq->tree, &vtx->key));
+	// remove rval is no way to check if element was present
+	(void)VRBT_REMOVE(vtx_tree, &vslq->tree, &vtx->key);
 	vtx->key.vxid = 0;
 	vtx->flags = 0;
 
@@ -582,22 +596,21 @@ vtx_retire(struct VSLQ *vslq, struct vtx **pvtx)
 	if (vslq->n_cache < VTX_CACHE) {
 		VTAILQ_INSERT_HEAD(&vslq->cache, vtx, list_child);
 		vslq->n_cache++;
-	} else {
+	} else
 		FREE_OBJ(vtx);
-		vtx = NULL;
-	}
+
 }
 
 /* Lookup a vtx by vxid from the managed list */
 static struct vtx *
-vtx_lookup(const struct VSLQ *vslq, unsigned vxid)
+vtx_lookup(const struct VSLQ *vslq, uint64_t vxid)
 {
 	struct vtx_key lkey, *key;
 	struct vtx *vtx;
 
 	AN(vslq);
 	lkey.vxid = vxid;
-	key = VRB_FIND(vtx_tree, &vslq->tree, &lkey);
+	key = VRBT_FIND(vtx_tree, &vslq->tree, &lkey);
 	if (key == NULL)
 		return (NULL);
 	CAST_OBJ_NOTNULL(vtx, (void *)key, VTX_MAGIC);
@@ -606,7 +619,7 @@ vtx_lookup(const struct VSLQ *vslq, unsigned vxid)
 
 /* Insert a new vtx into the managed list */
 static struct vtx *
-vtx_add(struct VSLQ *vslq, unsigned vxid)
+vtx_add(struct VSLQ *vslq, uint64_t vxid)
 {
 	struct vtx *vtx;
 
@@ -614,7 +627,7 @@ vtx_add(struct VSLQ *vslq, unsigned vxid)
 	vtx = vtx_new(vslq);
 	AN(vtx);
 	vtx->key.vxid = vxid;
-	AZ(VRB_INSERT(vtx_tree, &vslq->tree, &vtx->key));
+	AZ(VRBT_INSERT(vtx_tree, &vslq->tree, &vtx->key));
 	VTAILQ_INSERT_TAIL(&vslq->incomplete, vtx, list_vtx);
 	vslq->n_outstanding++;
 	return (vtx);
@@ -677,10 +690,10 @@ vtx_set_parent(struct vtx *parent, struct vtx *child)
    successfully parsed. */
 static int
 vtx_parse_link(const char *str, enum VSL_transaction_e *ptype,
-    unsigned *pvxid, enum VSL_reason_e *preason)
+    uint64_t *pvxid, enum VSL_reason_e *preason, uint64_t *psub)
 {
 	char type[16], reason[16];
-	unsigned vxid;
+	uintmax_t vxid, sub;
 	int i;
 	enum VSL_transaction_e et;
 	enum VSL_reason_e er;
@@ -690,12 +703,12 @@ vtx_parse_link(const char *str, enum VSL_transaction_e *ptype,
 	AN(pvxid);
 	AN(preason);
 
-	i = sscanf(str, "%15s %u %15s", type, &vxid, reason);
+	i = sscanf(str, "%15s %ju %15s %ju", type, &vxid, reason, &sub);
 	if (i < 1)
 		return (0);
 
 	/* transaction type */
-	for (et = 0; et < VSL_t__MAX; et++)
+	for (et = VSL_t_unknown; et < VSL_t__MAX; et++)
 		if (!strcmp(type, vsl_t_names[et]))
 			break;
 	if (et >= VSL_t__MAX)
@@ -711,13 +724,19 @@ vtx_parse_link(const char *str, enum VSL_transaction_e *ptype,
 		return (2);
 
 	/* transaction reason */
-	for (er = 0; er < VSL_r__MAX; er++)
+	for (er = VSL_r_unknown; er < VSL_r__MAX; er++)
 		if (!strcmp(reason, vsl_r_names[er]))
 			break;
 	if (er >= VSL_r__MAX)
 		er = VSL_r_unknown;
 	*preason = er;
-	return (3);
+	if (i == 3)
+		return (3);
+
+	/* request sub-level */
+	if (psub != NULL)
+		*psub = sub;
+	return (4);
 }
 
 /* Parse and process a begin record */
@@ -727,15 +746,15 @@ vtx_scan_begin(struct VSLQ *vslq, struct vtx *vtx, const uint32_t *ptr)
 	int i;
 	enum VSL_transaction_e type;
 	enum VSL_reason_e reason;
-	unsigned p_vxid;
+	uint64_t p_vxid;
 	struct vtx *p_vtx;
 
 	assert(VSL_TAG(ptr) == SLT_Begin);
 
 	AZ(vtx->flags & VTX_F_READY);
 
-	i = vtx_parse_link(VSL_CDATA(ptr), &type, &p_vxid, &reason);
-	if (i != 3)
+	i = vtx_parse_link(VSL_CDATA(ptr), &type, &p_vxid, &reason, NULL);
+	if (i < 3)
 		return (vtx_diag_tag(vtx, ptr, "parse error"));
 	if (type == VSL_t_unknown)
 		(void)vtx_diag_tag(vtx, ptr, "unknown vxid type");
@@ -794,15 +813,15 @@ vtx_scan_link(struct VSLQ *vslq, struct vtx *vtx, const uint32_t *ptr)
 	int i;
 	enum VSL_transaction_e c_type;
 	enum VSL_reason_e c_reason;
-	unsigned c_vxid;
+	uint64_t c_vxid;
 	struct vtx *c_vtx;
 
 	assert(VSL_TAG(ptr) == SLT_Link);
 
 	AZ(vtx->flags & VTX_F_READY);
 
-	i = vtx_parse_link(VSL_CDATA(ptr), &c_type, &c_vxid, &c_reason);
-	if (i != 3)
+	i = vtx_parse_link(VSL_CDATA(ptr), &c_type, &c_vxid, &c_reason, NULL);
+	if (i < 3)
 		return (vtx_diag_tag(vtx, ptr, "parse error"));
 	if (c_type == VSL_t_unknown)
 		(void)vtx_diag_tag(vtx, ptr, "unknown vxid type");
@@ -911,10 +930,32 @@ vtx_force(struct VSLQ *vslq, struct vtx *vtx, const char *reason)
 	AN(vtx->flags & VTX_F_COMPLETE);
 }
 
+static int
+vslq_ratelimit(struct VSLQ *vslq)
+{
+	vtim_mono now;
+	vtim_dur delta;
+
+	CHECK_OBJ_NOTNULL(vslq, VSLQ_MAGIC);
+	CHECK_OBJ_NOTNULL(vslq->vsl, VSL_MAGIC);
+
+	now = VTIM_mono();
+	delta = now - vslq->last_use;
+	vslq->credits += (delta / vslq->vsl->R_opt_p) * vslq->vsl->R_opt_l;
+	vslq->credits = vmin_t(double, vslq->credits, vslq->vsl->R_opt_l);
+	vslq->last_use = now;
+
+	if (vslq->credits < 1.0)
+		return (0);
+
+	vslq->credits -= 1.0;
+	return (1);
+}
+
 /* Build transaction array, do the query and callback. Returns 0 or the
    return value from func */
 static int
-vslq_callback(const struct VSLQ *vslq, struct vtx *vtx, VSLQ_dispatch_f *func,
+vslq_callback(struct VSLQ *vslq, struct vtx *vtx, VSLQ_dispatch_f *func,
     void *priv)
 {
 	unsigned n = vtx->n_descend + 1;
@@ -936,7 +977,7 @@ vslq_callback(const struct VSLQ *vslq, struct vtx *vtx, VSLQ_dispatch_f *func,
 		return (0);
 
 	/* Build transaction array */
-	(void)vslc_vtx_reset(&vtx->c.cursor);
+	AN(vslc_vtx_reset(&vtx->c.cursor) == vsl_end);
 	vtxs[0] = vtx;
 	trans[0].level = 1;
 	trans[0].vxid = vtx->key.vxid;
@@ -949,7 +990,7 @@ vslq_callback(const struct VSLQ *vslq, struct vtx *vtx, VSLQ_dispatch_f *func,
 	while (j < i) {
 		VTAILQ_FOREACH(vtx, &vtxs[j]->child, list_child) {
 			assert(i < n);
-			(void)vslc_vtx_reset(&vtx->c.cursor);
+			AN(vslc_vtx_reset(&vtx->c.cursor) == vsl_end);
 			vtxs[i] = vtx;
 			if (vtx->reason == VSL_r_restart)
 				/* Restarts stay at the same level as parent */
@@ -976,6 +1017,9 @@ vslq_callback(const struct VSLQ *vslq, struct vtx *vtx, VSLQ_dispatch_f *func,
 	if (vslq->query != NULL && !vslq_runquery(vslq->query, ptrans))
 		return (0);
 
+	if (vslq->vsl->R_opt_l != 0 && !vslq_ratelimit(vslq))
+		return (0);
+
 	/* Callback */
 	return ((func)(vslq->vsl, ptrans, priv));
 }
@@ -989,12 +1033,13 @@ vtx_synth_rec(struct vtx *vtx, unsigned tag, const char *fmt, ...)
 	va_list ap;
 	char *buf;
 	int l, buflen;
+	uint64_t vxid;
 
 	ALLOC_OBJ(synth, SYNTH_MAGIC);
 	AN(synth);
 
-	buf = (char *)&synth->data[2];
-	buflen = sizeof (synth->data) - 2 * sizeof (uint32_t);
+	buf = VSL_DATA(synth->data);
+	buflen = sizeof(synth->data) - VSL_BYTES(VSL_OVERHEAD);
 	va_start(ap, fmt);
 	l = vsnprintf(buf, buflen, fmt, ap);
 	assert(l >= 0);
@@ -1002,18 +1047,21 @@ vtx_synth_rec(struct vtx *vtx, unsigned tag, const char *fmt, ...)
 	if (l > buflen - 1)
 		l = buflen - 1;
 	buf[l++] = '\0';	/* NUL-terminated */
-	synth->data[1] = vtx->key.vxid;
+	vxid = vtx->key.vxid;
 	switch (vtx->type) {
 	case VSL_t_req:
-		synth->data[1] |= VSL_CLIENTMARKER;
+		vxid |= VSL_CLIENTMARKER;
 		break;
 	case VSL_t_bereq:
-		synth->data[1] |= VSL_BACKENDMARKER;
+		vxid |= VSL_BACKENDMARKER;
 		break;
 	default:
 		break;
 	}
-	synth->data[0] = (((tag & 0xff) << 24) | l);
+	synth->data[2] = vxid >> 32;
+	synth->data[1] = vxid;
+	synth->data[0] = (((tag & VSL_IDMASK) << VSL_IDSHIFT) |
+	    (VSL_VERSION_3 << VSL_VERSHIFT) | l);
 	synth->offset = vtx->c.offset;
 
 	VTAILQ_FOREACH_REVERSE(it, &vtx->synth, synthhead, list) {
@@ -1048,7 +1096,7 @@ static int
 vtx_diag_tag(struct vtx *vtx, const uint32_t *ptr, const char *reason)
 {
 
-	vtx_synth_rec(vtx, SLT_VSL, "%s (%u:%s \"%.*s\")", reason, VSL_ID(ptr),
+	vtx_synth_rec(vtx, SLT_VSL, "%s (%ju:%s \"%.*s\")", reason, VSL_ID(ptr),
 	    VSL_tags[VSL_TAG(ptr)], (int)VSL_LEN(ptr), VSL_CDATA(ptr));
 	return (-1);
 }
@@ -1061,7 +1109,7 @@ VSLQ_New(struct VSL_data *vsl, struct VSL_cursor **cp,
 	struct VSLQ *vslq;
 
 	CHECK_OBJ_NOTNULL(vsl, VSL_MAGIC);
-	if (grouping > VSL_g_session) {
+	if (grouping >= VSL_g__MAX) {
 		(void)vsl_diag(vsl, "Illegal query grouping");
 		return (NULL);
 	}
@@ -1081,9 +1129,13 @@ VSLQ_New(struct VSL_data *vsl, struct VSL_cursor **cp,
 	}
 	vslq->grouping = grouping;
 	vslq->query = query;
+	if (vslq->vsl->R_opt_l != 0) {
+		vslq->last_use = VTIM_mono();
+		vslq->credits = 1;
+	}
 
 	/* Setup normal mode */
-	VRB_INIT(&vslq->tree);
+	VRBT_INIT(&vslq->tree);
 	VTAILQ_INIT(&vslq->ready);
 	VTAILQ_INIT(&vslq->incomplete);
 	VTAILQ_INIT(&vslq->shmrefs);
@@ -1109,10 +1161,7 @@ VSLQ_Delete(struct VSLQ **pvslq)
 	struct VSLQ *vslq;
 	struct vtx *vtx;
 
-	AN(pvslq);
-	vslq = *pvslq;
-	*pvslq = NULL;
-	CHECK_OBJ_NOTNULL(vslq, VSLQ_MAGIC);
+	TAKE_OBJ_NOTNULL(vslq, pvslq, VSLQ_MAGIC);
 
 	(void)VSLQ_Flush(vslq, NULL, NULL);
 	AZ(vslq->n_outstanding);
@@ -1129,6 +1178,7 @@ VSLQ_Delete(struct VSLQ **pvslq)
 	while (!VTAILQ_EMPTY(&vslq->cache)) {
 		AN(vslq->n_cache);
 		vtx = VTAILQ_FIRST(&vslq->cache);
+		CHECK_OBJ_NOTNULL(vtx, VTX_MAGIC);
 		VTAILQ_REMOVE(&vslq->cache, vtx, list_child);
 		vslq->n_cache--;
 		FREE_OBJ(vtx);
@@ -1162,17 +1212,17 @@ VSLQ_SetCursor(struct VSLQ *vslq, struct VSL_cursor **cp)
 static int
 vslq_raw(struct VSLQ *vslq, VSLQ_dispatch_f *func, void *priv)
 {
-	int i = 1;
-	int r;
+	enum vsl_status r = vsl_more;
+	int i;
 
 	assert(vslq->grouping == VSL_g_raw);
 
 	assert(vslq->raw.offset <= vslq->raw.len);
 	do {
 		if (vslq->raw.offset == vslq->raw.len) {
-			i = VSL_Next(vslq->c);
-			if (i <= 0)
-				return (i);
+			r = VSL_Next(vslq->c);
+			if (r != vsl_more)
+				return (r);
 			AN(vslq->c->rec.ptr);
 			vslq->raw.start = vslq->c->rec;
 			if (VSL_TAG(vslq->c->rec.ptr) == SLT__Batch)
@@ -1192,96 +1242,153 @@ vslq_raw(struct VSLQ *vslq, VSLQ_dispatch_f *func, void *priv)
 		vslq->raw.offset += VSL_NEXT(vslq->raw.c.ptr) - vslq->raw.c.ptr;
 	} while (VSL_TAG(vslq->raw.c.ptr) == SLT__Batch);
 
+	assert (r == vsl_more);
+
 	if (func == NULL)
-		return (i);
+		return (r);
 
 	if (vslq->query != NULL &&
 	    !vslq_runquery(vslq->query, vslq->raw.ptrans))
-		return (i);
-
-	r = (func)(vslq->vsl, vslq->raw.ptrans, priv);
-	if (r)
 		return (r);
 
-	return (i);
+	if (vslq->vsl->R_opt_l != 0 && !vslq_ratelimit(vslq))
+		return (r);
+
+	i = (func)(vslq->vsl, vslq->raw.ptrans, priv);
+	if (i)
+		return (i);
+
+	return (r);
 }
 
 /* Check the beginning of the shmref list, and buffer refs that are at
  * warning level.
- *
- * Returns:
- *    0:	OK
- *   -3:	Failure
  */
-static int
+static enum vsl_status
 vslq_shmref_check(struct VSLQ *vslq)
 {
 	struct chunk *chunk;
-	int i;
+	enum vsl_check i;
 
 	while ((chunk = VTAILQ_FIRST(&vslq->shmrefs)) != NULL) {
 		CHECK_OBJ_NOTNULL(chunk, CHUNK_MAGIC);
 		assert(chunk->type == chunk_t_shm);
 		i = VSL_Check(vslq->c, &chunk->shm.start);
-		if (i == 2)
+		switch (i) {
+		case vsl_check_valid:
 			/* First on list is OK, refs behind it must also
 			   be OK */
-			return (0);
-		else if (i == 1)
-			/* Warning level. Buffer this chunk */
+			return (vsl_more);
+		case vsl_check_warn:
+			/* Buffer this chunk */
 			chunk_shm_to_buf(vslq, chunk);
-		else
+			break;
+		default:
 			/* Too late to buffer */
-			return (-3);
+			return (vsl_e_overrun);
+		}
 	}
 
-	return (0);
+	return (vsl_more);
+}
+
+static unsigned
+vslq_candidate(struct VSLQ *vslq, const uint32_t *ptr)
+{
+	enum VSL_transaction_e type;
+	enum VSL_reason_e reason;
+	struct VSL_data *vsl;
+	enum VSL_tag_e tag;
+	uint64_t p_vxid, sub;
+	int i;
+
+	CHECK_OBJ_NOTNULL(vslq, VSLQ_MAGIC);
+	AN(ptr);
+
+	assert(vslq->grouping != VSL_g_raw);
+	if (vslq->grouping == VSL_g_session)
+		return (1); /* All are needed */
+
+	vsl = vslq->vsl;
+	CHECK_OBJ_NOTNULL(vsl, VSL_MAGIC);
+	if (vslq->grouping == VSL_g_vxid) {
+		if (!vsl->c_opt && !vsl->b_opt)
+			AZ(vsl->E_opt);
+		else if (!vsl->b_opt && !VSL_CLIENT(ptr))
+			return (0);
+		else if (!vsl->c_opt && !VSL_BACKEND(ptr))
+			return (0);
+		/* Need to parse the Begin tag - fallthrough to below */
+	}
+
+	tag = VSL_TAG(ptr);
+	assert(tag == SLT_Begin);
+	i = vtx_parse_link(VSL_CDATA(ptr), &type, &p_vxid, &reason, &sub);
+	if (i < 3 || type == VSL_t_unknown)
+		return (0);
+
+	if (vslq->grouping == VSL_g_request && type == VSL_t_sess)
+		return (0);
+
+	if (vslq->grouping == VSL_g_vxid && i > 3 && sub > 0 && !vsl->E_opt)
+		return (0);
+
+	return (1);
 }
 
 /* Process next input record */
-static int
+static enum vsl_status
 vslq_next(struct VSLQ *vslq)
 {
+	const uint32_t *ptr;
 	struct VSL_cursor *c;
-	int i;
+	enum vsl_status r;
 	enum VSL_tag_e tag;
 	ssize_t len;
-	unsigned vxid;
+	uint64_t vxid;
+	unsigned keep;
 	struct vtx *vtx;
 
 	c = vslq->c;
-	i = VSL_Next(c);
-	if (i != 1)
-		return (i);
+	r = VSL_Next(c);
+	if (r != vsl_more)
+		return (r);
 
-	tag = VSL_TAG(c->rec.ptr);
+	assert (r == vsl_more);
+
+	tag = (enum VSL_tag_e)VSL_TAG(c->rec.ptr);
 	if (tag == SLT__Batch) {
 		vxid = VSL_BATCHID(c->rec.ptr);
 		len = VSL_END(c->rec.ptr, VSL_BATCHLEN(c->rec.ptr)) -
 		    c->rec.ptr;
 		if (len == 0)
-			return (i);
-		tag = VSL_TAG(VSL_NEXT(c->rec.ptr));
+			return (r);
+		ptr = VSL_NEXT(c->rec.ptr);
+		tag = (enum VSL_tag_e)VSL_TAG(ptr);
 	} else {
 		vxid = VSL_ID(c->rec.ptr);
 		len = VSL_NEXT(c->rec.ptr) - c->rec.ptr;
+		ptr = c->rec.ptr;
 	}
 	assert(len > 0);
 	if (vxid == 0)
 		/* Skip non-transactional records */
-		return (i);
+		return (r);
 
 	vtx = vtx_lookup(vslq, vxid);
-	if (vtx == NULL && tag == SLT_Begin) {
+	keep = tag != SLT_Begin || vslq_candidate(vslq, ptr);
+	if (vtx == NULL && tag == SLT_Begin && keep) {
 		vtx = vtx_add(vslq, vxid);
 		AN(vtx);
 	}
 	if (vtx != NULL) {
-		vtx_append(vslq, vtx, &c->rec, len);
-		vtx_scan(vslq, vtx);
+		AN(keep);
+		r = vtx_append(vslq, vtx, &c->rec, len);
+		if (r == vsl_more)
+			vtx_scan(vslq, vtx);
 	}
 
-	return (i);
+	return (r);
 }
 
 /* Test query and report any ready transactions */
@@ -1314,7 +1421,8 @@ vslq_process_ready(struct VSLQ *vslq, VSLQ_dispatch_f *func, void *priv)
 int
 VSLQ_Dispatch(struct VSLQ *vslq, VSLQ_dispatch_f *func, void *priv)
 {
-	int i, r;
+	enum vsl_status r;
+	int i;
 	double now;
 	struct vtx *vtx;
 
@@ -1322,23 +1430,25 @@ VSLQ_Dispatch(struct VSLQ *vslq, VSLQ_dispatch_f *func, void *priv)
 
 	/* Check that we have a cursor */
 	if (vslq->c == NULL)
-		return (-2);
+		return (vsl_e_abandon);
 
 	if (vslq->grouping == VSL_g_raw)
 		return (vslq_raw(vslq, func, priv));
 
 	/* Process next cursor input */
-	i = vslq_next(vslq);
-	if (i <= 0)
+	r = vslq_next(vslq);
+	if (r != vsl_more)
 		/* At end of log or cursor reports error condition */
-		return (i);
+		return (r);
 
 	/* Check shmref list and buffer if necessary */
 	r = vslq_shmref_check(vslq);
-	if (r)
+	if (r != vsl_more)
 		/* Buffering of shm ref failed */
 		return (r);
 
+	assert (r == vsl_more);
+
 	/* Check vtx timeout */
 	now = VTIM_mono();
 	while (!VTAILQ_EMPTY(&vslq->incomplete)) {
@@ -1357,22 +1467,21 @@ VSLQ_Dispatch(struct VSLQ *vslq, VSLQ_dispatch_f *func, void *priv)
 		CHECK_OBJ_NOTNULL(vtx, VTX_MAGIC);
 		vtx_force(vslq, vtx, "store overflow");
 		AN(vtx->flags & VTX_F_COMPLETE);
-		r = vslq_process_ready(vslq, func, priv);
-		if (r)
+		i = vslq_process_ready(vslq, func, priv);
+		if (i)
 			/* User return code */
-			return (r);
+			return (i);
 	}
 
 	/* Check ready list */
 	if (!VTAILQ_EMPTY(&vslq->ready)) {
-		r = vslq_process_ready(vslq, func, priv);
-		if (r)
+		i = vslq_process_ready(vslq, func, priv);
+		if (i)
 			/* User return code */
-			return (r);
+			return (i);
 	}
 
-	/* Return cursor return value */
-	return (i);
+	return (vsl_more);
 }
 
 /* Flush any incomplete vtx held on to. Do callbacks if func != NULL */
