@@ -40,10 +40,14 @@
 #include "cache/cache_filter.h"
 
 #include "vsa.h"
+#include "vss.h"
+#include "vtcp.h"
 #include "vtim.h"
 #include "vcc_debug_if.h"
 #include "VSC_debug.h"
 
+#include "vmod_debug.h"
+
 struct priv_vcl {
 	unsigned		magic;
 #define PRIV_VCL_MAGIC		0x8E62FA9D
@@ -55,6 +59,7 @@ struct priv_vcl {
 	VCL_BACKEND		be;
 	unsigned		cold_be;
 	unsigned		cooling_be;
+	int			tmpf;
 };
 
 
@@ -62,228 +67,13 @@ static pthread_mutex_t vsc_mtx = PTHREAD_MUTEX_INITIALIZER;
 static struct vsc_seg *vsc_seg = NULL;
 static struct VSC_debug *vsc = NULL;
 static int loads;
-static const int store_ip_token;
-static const int fail_task_fini_token;
+static const int store_ip_token = 0;
+static const int fail_task_fini_token = 0;
 extern void mylog(struct vsl_log *vsl, enum VSL_tag_e tag,
     const char *fmt, ...) v_printflike_(3,4);
 
 /**********************************************************************/
 
-static enum vfp_status v_matchproto_(vfp_pull_f)
-xyzzy_vfp_rot13_pull(struct vfp_ctx *vc, struct vfp_entry *vfe, void *p,
-    ssize_t *lp)
-{
-	enum vfp_status vp;
-	char *q;
-	ssize_t l;
-
-	(void)vfe;
-	vp = VFP_Suck(vc, p, lp);
-	if (vp == VFP_ERROR)
-		return (vp);
-	q = p;
-	for (l = 0; l < *lp; l++, q++) {
-		if (*q >= 'A' && *q <= 'Z')
-			*q = (((*q - 'A') + 13) % 26) + 'A';
-		if (*q >= 'a' && *q <= 'z')
-			*q = (((*q - 'a') + 13) % 26) + 'a';
-	}
-	return (vp);
-}
-
-static const struct vfp xyzzy_vfp_rot13 = {
-	.name = "rot13",
-	.pull = xyzzy_vfp_rot13_pull,
-};
-
-/**********************************************************************/
-
-// deliberately fragmenting the stream to make testing more interesting
-#define ROT13_BUFSZ 8
-
-static int v_matchproto_(vdp_init_f)
-xyzzy_vfp_rot13_init(VRT_CTX, struct vdp_ctx *vdc, void **priv, struct objcore *oc)
-{
-	(void)vdc;
-	(void)oc;
-
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	AN(priv);
-	*priv = malloc(ROT13_BUFSZ);
-	if (*priv == NULL)
-		return (-1);
-	return (0);
-}
-
-static int v_matchproto_(vdp_bytes_f)
-xyzzy_vfp_rot13_bytes(struct vdp_ctx *vdc, enum vdp_action act, void **priv,
-    const void *ptr, ssize_t len)
-{
-	char *q;
-	const char *pp;
-	int i, j, retval = 0;
-
-	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
-	AN(priv);
-	AN(*priv);
-	if (len <= 0)
-		return (VDP_bytes(vdc, act, ptr, len));
-	AN(ptr);
-	if (act != VDP_END)
-		act = VDP_FLUSH;
-	q = *priv;
-	pp = ptr;
-
-	for (i = 0, j = 0; j < len; i++, j++) {
-		if (pp[j] >= 'A' && pp[j] <= 'Z')
-			q[i] = (((pp[j] - 'A') + 13) % 26) + 'A';
-		else if (pp[j] >= 'a' && pp[j] <= 'z')
-			q[i] = (((pp[j] - 'a') + 13) % 26) + 'a';
-		else
-			q[i] = pp[j];
-		if (i == ROT13_BUFSZ - 1 && j < len - 1) {
-			retval = VDP_bytes(vdc, VDP_FLUSH, q, ROT13_BUFSZ);
-			if (retval != 0)
-				return (retval);
-			i = -1;
-		}
-	}
-	if (i >= 0)
-		retval = VDP_bytes(vdc, act, q, i);
-	return (retval);
-}
-
-static int v_matchproto_(vdp_fini_f)
-xyzzy_vfp_rot13_fini(struct vdp_ctx *vdc, void **priv)
-{
-	(void)vdc;
-	AN(priv);
-	free(*priv);
-	*priv = NULL;
-	return (0);
-}
-
-static const struct vdp xyzzy_vdp_rot13 = {
-	.name  = "rot13",
-	.init  = xyzzy_vfp_rot13_init,
-	.bytes = xyzzy_vfp_rot13_bytes,
-	.fini  = xyzzy_vfp_rot13_fini,
-};
-
-/**********************************************************************
- * pendantic tests of the VDP API:
- * - assert that we see a VDP_END
- * - assert that _fini gets called before the task ends
- *
- * note:
- * we could lookup our own vdpe in _fini and check for vdpe->end == VDP_END
- * yet that would cross the API
- */
-
-enum vdp_state_e {
-	VDPS_NULL = 0,
-	VDPS_INIT,	// _init called
-	VDPS_BYTES,	// _bytes called act != VDP_END
-	VDPS_END,	// _bytes called act == VDP_END
-	VDPS_FINI	// _fini called
-};
-
-struct vdp_state_s {
-	unsigned		magic;
-#define VDP_STATE_MAGIC	0x57c8d309
-	enum vdp_state_e	state;
-};
-
-static void v_matchproto_(vmod_priv_fini_f)
-priv_pedantic_fini(VRT_CTX, void *priv)
-{
-	struct vdp_state_s *vdps;
-
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CAST_OBJ_NOTNULL(vdps, priv, VDP_STATE_MAGIC);
-
-	assert(vdps->state == VDPS_FINI);
-}
-
-static const struct vmod_priv_methods priv_pedantic_methods[1] = {{
-	.magic = VMOD_PRIV_METHODS_MAGIC,
-	.type = "debug_vdp_pedantic",
-	.fini = priv_pedantic_fini
-}};
-
-static int v_matchproto_(vdp_init_f)
-xyzzy_pedantic_init(VRT_CTX, struct vdp_ctx *vdc, void **priv,
-    struct objcore *oc)
-{
-	struct vdp_state_s *vdps;
-	struct vmod_priv *p;
-
-	(void)oc;
-
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	WS_TASK_ALLOC_OBJ(ctx, vdps, VDP_STATE_MAGIC);
-	if (vdps == NULL)
-		return (-1);
-	assert(vdps->state == VDPS_NULL);
-
-	p = VRT_priv_task(ctx, (void *)vdc);
-	if (p == NULL)
-		return (-1);
-	p->priv = vdps;
-	p->methods = priv_pedantic_methods;
-
-	AN(priv);
-	*priv = vdps;
-
-	vdps->state = VDPS_INIT;
-
-	return (0);
-}
-
-static int v_matchproto_(vdp_bytes_f)
-xyzzy_pedantic_bytes(struct vdp_ctx *vdc, enum vdp_action act, void **priv,
-    const void *ptr, ssize_t len)
-{
-	struct vdp_state_s *vdps;
-
-	CAST_OBJ_NOTNULL(vdps, *priv, VDP_STATE_MAGIC);
-	assert(vdps->state >= VDPS_INIT);
-	assert(vdps->state < VDPS_END);
-
-	if (act == VDP_END)
-		vdps->state = VDPS_END;
-	else
-		vdps->state = VDPS_BYTES;
-
-	return (VDP_bytes(vdc, act, ptr, len));
-}
-
-static int v_matchproto_(vdp_fini_f)
-xyzzy_pedantic_fini(struct vdp_ctx *vdc, void **priv)
-{
-	struct vdp_state_s *vdps;
-
-	(void) vdc;
-	AN(priv);
-	if (*priv == NULL)
-		return (0);
-	CAST_OBJ_NOTNULL(vdps, *priv, VDP_STATE_MAGIC);
-	assert(vdps->state == VDPS_INIT || vdps->state == VDPS_END);
-	vdps->state = VDPS_FINI;
-
-	*priv = NULL;
-	return (0);
-}
-
-static const struct vdp xyzzy_vdp_pedantic = {
-	.name  = "debug.pedantic",
-	.init  = xyzzy_pedantic_init,
-	.bytes = xyzzy_pedantic_bytes,
-	.fini  = xyzzy_pedantic_fini,
-};
-
-/**********************************************************************/
-
 VCL_STRING v_matchproto_(td_debug_author)
 xyzzy_author(VRT_CTX, VCL_ENUM person, VCL_ENUM someone)
 {
@@ -401,21 +191,18 @@ VCL_VOID v_matchproto_(td_debug_test_priv_vcl)
 xyzzy_test_priv_vcl(VRT_CTX, struct vmod_priv *priv)
 {
 	struct priv_vcl *priv_vcl;
+	char t[PATH_MAX];
+	ssize_t l;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	AN(priv);
 	CAST_OBJ_NOTNULL(priv_vcl, priv->priv, PRIV_VCL_MAGIC);
+
+	l = pread(priv_vcl->tmpf, t, sizeof t, 0);
+	assert(l > 0);
+
 	AN(priv_vcl->foo);
-	assert(!strcmp(priv_vcl->foo, "FOO"));
-}
-
-VCL_VOID v_matchproto_(td_debug_rot104)
-xyzzy_rot104(VRT_CTX)
-{
-
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	// This should fail
-	AN(VRT_AddFilter(ctx, &xyzzy_vfp_rot13, &xyzzy_vdp_rot13));
+	assert(!strncmp(priv_vcl->foo, t, l));
 }
 
 VCL_VOID v_matchproto_(td_debug_rot52)
@@ -436,7 +223,7 @@ xyzzy_argtest(VRT_CTX, struct VARGS(argtest) *arg)
 	AN(arg);
 	bprintf(buf, "%s %g %s %s %jd %d %s",
 	    arg->one, arg->two, arg->three, arg->comma, (intmax_t)arg->four,
-	    arg->valid_opt, arg->valid_opt ? arg->opt : "<undef>");
+	    arg->valid_string, arg->valid_string ? arg->string : "<undef>");
 	return (WS_Copy(ctx->ws, buf, -1));
 }
 
@@ -501,7 +288,9 @@ priv_vcl_fini(VRT_CTX, void *priv)
 	struct priv_vcl *priv_vcl;
 
 	CAST_OBJ_NOTNULL(priv_vcl, priv, PRIV_VCL_MAGIC);
+	AZ(close(priv_vcl->tmpf));
 	AN(priv_vcl->foo);
+	AZ(unlink(priv_vcl->foo));
 	free(priv_vcl->foo);
 	if (priv_vcl->obj_cb != 0) {
 		ObjUnsubscribeEvents(&priv_vcl->obj_cb);
@@ -534,13 +323,17 @@ event_load(VRT_CTX, struct vmod_priv *priv)
 
 	ALLOC_OBJ(priv_vcl, PRIV_VCL_MAGIC);
 	AN(priv_vcl);
-	priv_vcl->foo = strdup("FOO");
+	priv_vcl->foo = strdup("worker_tmpdir/vmod_debug.XXXXXX");
 	AN(priv_vcl->foo);
+	priv_vcl->tmpf = mkstemp(priv_vcl->foo);
+	assert(priv_vcl->tmpf >= 0);
+	AN(write(priv_vcl->tmpf, priv_vcl->foo, strlen(priv_vcl->foo)));
 	priv->priv = priv_vcl;
 	priv->methods = priv_vcl_methods;
 
-	AZ(VRT_AddFilter(ctx, &xyzzy_vfp_rot13, &xyzzy_vdp_rot13));
-	AZ(VRT_AddFilter(ctx, NULL, &xyzzy_vdp_pedantic));
+	debug_add_filters(ctx);
+	debug_transport_reembarking_http1_init();
+	debug_transport_vai_init();
 	return (0);
 }
 
@@ -702,8 +495,7 @@ event_discard(VRT_CTX, void *priv)
 
 	AZ(ctx->msg);
 
-	VRT_RemoveFilter(ctx, &xyzzy_vfp_rot13, &xyzzy_vdp_rot13);
-	VRT_RemoveFilter(ctx, NULL, &xyzzy_vdp_pedantic);
+	debug_remove_filters(ctx);
 
 	if (--loads)
 		return (0);
@@ -886,18 +678,19 @@ xyzzy_sethdr(VRT_CTX, VCL_HEADER hdr, VCL_STRANDS s)
 	hp = VRT_selecthttp(ctx, hdr->where);
 	if (hp == NULL) {
 		VRT_fail(ctx, "debug.sethdr(): header argument "
-		    "can not be used here");
+		    "cannot be used here");
 		return;
 	}
 	AN(hdr->what);
 	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
+	CHECK_OBJ_NOTNULL(s, STRANDS_MAGIC);
 	if (s->n == 0) {
 		http_Unset(hp, hdr->what);
 	} else {
-		b = VRT_StrandsWS(hp->ws, hdr->what + 1, s);
+		b = VRT_StrandsWS(hp->ws, hdr->what->str, s);
 		if (b == NULL) {
 			VSLbs(ctx->vsl, SLT_LostHeader,
-			    TOSTRAND(hdr->what + 1));
+			    TOSTRAND(hdr->what->str));
 		} else {
 			if (*b != '\0')
 				AN(WS_Allocated(hp->ws, b, strlen(b) + 1));
@@ -965,10 +758,27 @@ xyzzy_return_strands(VRT_CTX, VCL_STRANDS strand)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	VSLbs(ctx->vsl, SLT_Debug, strand);
+	if (ctx->vsl)
+		VSLbs(ctx->vsl, SLT_Debug, strand);
+	else
+		VSLs(SLT_Debug, NO_VXID, strand);
 	return (strand);
 }
 
+VCL_STRANDS
+xyzzy_return_null_strands(VRT_CTX)
+{
+	(void)ctx;
+	return (vrt_null_strands);
+}
+
+VCL_BOOL
+xyzzy_is_null_strands(VRT_CTX, VCL_STRANDS s)
+{
+	(void)ctx;
+	return (s == vrt_null_strands);
+}
+
 VCL_VOID
 xyzzy_vsl_flush(VRT_CTX)
 {
@@ -1429,3 +1239,137 @@ xyzzy_caller_xsub(VRT_CTX, struct VPFX(debug_caller) *caller)
 
 	return (caller->sub);
 }
+
+struct resolve_priv {
+	struct vsb vsb[1];
+	const char *fail_port;
+	const char *errp[1];
+};
+
+static int v_matchproto_(vus_resolved_f)
+resolve_cb(void *priv, const struct suckaddr *sa)
+{
+	struct resolve_priv *p;
+	char abuf[VTCP_ADDRBUFSIZE], pbuf[VTCP_PORTBUFSIZE];
+
+	p = (struct resolve_priv *)priv;
+	CHECK_OBJ_NOTNULL(p->vsb, VSB_MAGIC);
+	AN(sa);
+	VTCP_name(sa, abuf, sizeof abuf, pbuf, sizeof pbuf);
+	if (p->fail_port != NULL && !strcmp(p->fail_port, pbuf)) {
+		*(p->errp) = "bad port";
+		return (-1);
+	}
+	VSB_printf(p->vsb, "%s%s:%s", VSB_len(p->vsb) ? ", " : "", abuf, pbuf);
+	return (0);
+}
+
+VCL_STRING
+xyzzy_resolve_range(VRT_CTX, struct VARGS(resolve_range) *args)
+{
+	struct resolve_priv p;
+	const char *def_port;
+	int ret;
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	if (args->addr == NULL)
+		return ("vmod-debug: s was NULL");
+	memset(&p, 0, sizeof p);
+	WS_VSB_new(p.vsb, ctx->ws);
+	p.fail_port = args->valid_fail_port ? args->fail_port : NULL;
+	def_port = args->valid_def_port ? args->def_port : NULL;
+	ret = VSS_resolver_range(args->addr, def_port, resolve_cb, &p, p.errp);
+	if (ret)
+		VSB_printf(p.vsb, "%s%s", VSB_len(p.vsb) ? ", " : "Failed: ",
+		    *(p.errp));
+	return (WS_VSB_finish(p.vsb, ctx->ws, NULL));
+}
+
+VCL_VOID
+xyzzy_use_reembarking_http1(VRT_CTX)
+{
+	debug_transport_reembarking_http1_use(ctx);
+}
+
+VCL_VOID
+xyzzy_use_vai_http1(VRT_CTX)
+{
+	debug_transport_vai_use(ctx);
+}
+
+static int
+in_oc(struct worker *wrk, struct objcore *oc, const char *p)
+{
+	const char *hdrs;
+	ssize_t len = 0;
+
+	if (oc == NULL)
+		return (0);
+	hdrs = ObjGetAttr(wrk, oc, OA_HEADERS, &len);
+	if (hdrs == NULL)
+		return (0);
+	if (p < hdrs)
+		return (0);
+	if (p > hdrs + len)
+		return (0);
+	return (1);
+}
+
+static const char *
+ptr_where(VRT_CTX, const char *p)
+{
+	struct ws *ws = ctx->ws;
+	struct ws *aws;
+	struct worker *wrk;
+	struct objcore *oc, *stale_oc;
+
+	if (ctx->req != NULL) {
+		wrk = ctx->req->wrk;
+		oc = ctx->req->objcore;
+		stale_oc = ctx->req->stale_oc;
+	}
+	else if (ctx->bo != NULL) {
+		wrk = ctx->bo->wrk;
+		oc = ctx->bo->fetch_objcore;
+		stale_oc = ctx->bo->stale_oc;
+	}
+	else
+		WRONG("ctx");
+
+	AN(wrk);
+	aws = wrk->aws;
+
+	if (WS_Allocated(ws, p, -1))
+		return ("ws");
+	if (WS_Allocated(aws, p, -1))
+		return ("aws");
+	if (in_oc(wrk, oc, p))
+		return ("oc");
+	if (in_oc(wrk, stale_oc, p))
+		return ("stale_oc");
+	return ("?");
+}
+
+VCL_VOID
+xyzzy_log_strands(VRT_CTX, VCL_STRING prefix, VCL_STRANDS subject, VCL_INT nn)
+{
+	int i, n;
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	if (prefix == NULL)
+		prefix = "";
+	AN(subject);
+	if (nn > INT_MAX)
+		n = INT_MAX;
+	else if (nn < 0)
+		n = 0;
+	else
+		n = nn;
+
+	CHECK_OBJ_NOTNULL(subject, STRANDS_MAGIC);
+	for (i = 0; i < subject->n; i++) {
+		const char *p = subject->p[i];
+		mylog(ctx->vsl, SLT_Debug, "%s[%d]: (%s) %p %.*s%s", prefix, i,
+		    ptr_where(ctx, p), p, n, p, strlen(p) > (unsigned)n ? "..." : "");
+	}
+}
